
#ifndef CMFIELD_H
#define CMFIELD_H
 
#include <iostream>
#include <giac/giac.h>

#include <string>
#include <map>
#include <algorithm>
//#include "Util.h"

using namespace std;
using namespace giac;

extern context ct;

class CMfield
{

 public:

  CMfield(string s);
  // so far this only works with a few strings (corresponding to sporadic groups)

  CMfield(gen M); 
  // matrix of a Hermitian form, will constuct the field generated by the coeffs

  //  gen eps, precfactor;
  unsigned nd, ndmax, ndrecord, pf; // nd = current number of digits used in realroot, never allowed to go beyond ndmax
                                    // just for fun, we keep track of the largest number of digits needed (ndrecord)!
                                    // pf = precision factor when increasing precision (replaces nd by nd*pf)
  gen prec;

  gen zeronumber, onenumber, twonumber, ten, zerovector;

  gen t1, t2;
  vecteur vars1d;
  vecteur vars2d;
  
  // number field data;
  gen minpol, 
    genconj, 
    genasrootof, realgenasrootof,
    realminpol, realgen, 
    xv, sv, // approx values of generator of number field (resp. of totally real field)
    xvi, svi, // box/interval for previous values
    othersvals, // isolating intervals for wrong roots of realminpol
    dc, dr, // degrees of complex/real fields 
    genericreal,
    realgencoeffsinverse;

  vecteur realminpolcoeffs, realgenquadnum;

  gen genasrootofpowers;

  string tag, tautag;
  gen tau, v;
  vecteur vars;
  gen Id, Id2;

  map<string,gen> * data();

  /*
   * linear algebra/number field methods
   */
  void computeXval();
  void increasePrecision();
  void lowerPrecision();

  void red(gen &z); // beware this modifies the argument!
  void vectred(gen &v);
  void matred(gen &M);

  gen myconj(gen g);
  gen vectconj(gen g);
  gen matconj(gen g);

  gen inverse(gen g);
  gen matinverse(gen M);
  
  bool testDependent(gen V, gen W);

  gen evali(gen g, gen prec);
  gen evali(gen g);
  gen vectevali(gen g);
  gen matevali(gen g);
  gen evaln(gen g);
  gen vectevaln(gen g);
  gen matevaln(gen g);
  bool isReal(gen g);

  gen realPart(gen g);

  gen convertToReal(gen g);

  bool checkSign(gen g, int &res);
  bool checkSignAttempt(gen g, int &res);

  void findgenasrootof();
  gen findInField(gen pol, gen approx);
  /*
   * pol is the minimal polynomial of an algebraic number,
   *  approx is a float approximation of the root
   *   of course if the approx is too crude, this won't work!
   */

  gen e1,e2,e3; // standard basis vectors
  gen createMatrix(vecteur v);
  gen createMatrix(unsigned m, unsigned n, vecteur v);
  bool isScalar(gen M);
  gen mypower(gen M, unsigned k);
  int order(gen M, unsigned k);
  int order(gen M);
  int linearorder(gen M, unsigned k);
  int linearorder(gen M);

  gen transpose(gen M);

  gen det(gen M);
  gen trace(gen M);
  gen discr(gen z);
  gen discrAdjusted(gen z);

  bool areDep(gen v1, gen v2);
  bool isMultiple(gen A, gen B);
  int braidOrder(gen M, gen N, unsigned k);
  int braidOrder(gen M, gen N);

  bool siftAttempt(gen &rur, vecteur &res);
  bool safeSift(gen &rur, vecteur &res);

  bool checkSignPol(gen par, gen mpol, gen q, int &res);
  bool checkSignPolAttempt(gen par, gen mpol, gen q, int &res);

  bool identifyQuadraticNumber(gen &te, vecteur &res);
  bool convertRootOf(gen &te, gen &res);
  gen getCoeffs(gen &te);

};

extern map<string,gen> * minpols;

#endif
