
#ifndef POLYTOPESMALL_H
#define POLYTOPESMALL_H
 
#include <iostream>
#include <giac/giac.h>

#include <string>
#include <map>
#include <algorithm>

using namespace std;
using namespace giac;

#include "CMfield.h"
#include "Prism.h"
#include "Util.h"
#include "edge.h"
#include "ridge.h"
#include "sgraph.h"
#include "FiniteGroup.h"
#include "PolytopeSmall.h"

extern context ct;
class PolytopeSmall: public CMfield
{
  public:

  PolytopeSmall(string s);

  gen p0, p0c;
 
  gen H; // Matrix of the ambient Hermitian form
  gen R1, R2, R3, R1i, R2i, R3i;

  gen mult, multconj;

  WVector e1w, e2w, e3w;

  vector<int> word_p1, word_m1;
  vector<Prism> faces;

  vector<WVector> vertices;
  vector<edge> edges;
  vector<ridge> ridges;

  vector<vector<int> > relations;

  gen Proj(gen v, gen w);  
  gen Inn(gen v, gen w);
  gen SqNorm(gen v);
  gen BoxProd(gen v, gen w);

  gen refl(gen x, gen v, gen z);
  gen reflMat(gen v, gen z);

  bool testSame(WVector V, WVector W);
  bool isInList(WVector V, vector<WVector> li);
  int findInList(WVector V, vector<WVector> li);
  int findInList(gen V, vector<WVector> li);

  void importSporadicGroup();

  Prism createPrism(WVector v0, WVector v1, WVector v2);

  bool isPrismNew(Prism A);
  bool isVertexNew(WVector V);
  bool comparePrisms(Prism A, Prism B);
  bool samePolygon(vector<WVector> V, vector<WVector> W);  
  bool samePolygon(vector<gen> V, vector<gen> W);  

  bool computeVertices(Prism &A);

  bool isOnPrism(gen X, Prism A);
  bool isOnMirror(gen X, Prism A);
  bool isOnTopFace(gen X, Prism A);
  gen reflectAcrossSpine(gen X, Prism A);

  Prism applyReflection(Prism A);
  Prism applyInverseReflection(Prism A);

  Prism applyMatrix(gen M, Prism A);

  bool isVertexSafe(WVector V, Prism A);

  vector<int> findFaces(WVector v);
  vector<int> findFaces(gen V);
  vector<int> findBisectors(WVector v);
  vector<int> findBisectors(gen V);
  vector<WVector> findMirrors(WVector v);
  vector<WVector> findMirrors(gen V);

  void useRelations(vector<int> &w);
  void useRelations(Prism &A);

  void addEdge(edge e);
  bool isSame(edge e1, edge e2);
  bool isFlipped(edge e1, edge e2);
  bool isEdgeNew(edge e);

  void addRidge(ridge e);
  bool isSame(ridge e1, ridge e2);
  bool isRotated(ridge e1, ridge e2);
  bool isRidgeNew(ridge e);

  void addEdgeOld(edge e);

  void createBraidRelation(vector<int> w1, vector<int> w2);

  gen evalWord(vector<int> w);
  void evalWord(vector<int> w, gen &X);

  int applyPairing(ridge e);
  void checkCycles();

  int applyPairingToVertex(int jface, int j);
  int applyPairingToEdge(int jface, int j);
  int applyPairingToRidge(int jface, int j);
  
  void drawFace(unsigned j); 
  void drawProjection(unsigned axis); 

  bool checkSide();
  void checkEdge(unsigned j);
  void checkRidge(int j, int k);

  void checkPairings();
  void computeAllVertices();
  void computeAllEdges();
  void computeAllRidges();

  bool checkt1val(gen rur, gen par, gen t1v, int &res);
  bool checkt1valAttempt(gen rur, gen par, gen t1v, int &res);
  bool checkEqualt1val(gen rur, gen par, gen t1v);

  bool checkt2val(gen rur, gen par, gen t1v, int &res);
  bool checkt2valAttempt(gen rur, gen par, gen t1v, int &res);
  bool checkEqualt2val(gen rur, gen par, gen t1v);

  bool checkt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res);
  bool checkt3valAttempt(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res);
  bool checkEqualt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt);

  bool checkEqSign(gen rur, gen par, gen eq, int &res);
  bool checkEqSignAttempt(gen rur, gen par, gen eq, int &res);
  bool checkEqZero(gen rur, gen par, gen eq);

  void drawRidgePic(unsigned i);

  bool isRidgeComplex(int j, gen &mi);

  void importFaceData();

  gen identifyMirror(vector<int> w);
};

#endif
