#include "Polytope.h"
#include <cmath>
#include <iostream>
#include <fstream>
#include <string>

using namespace std;
using namespace giac;

Polytope::Polytope(string s): CMfield(s) {
  initSporadicGroup();
};

void Polytope::initSporadicGroup() {

  emb_dim1 = false;
  emb_dim2 = false;
  check_cycles = true;
  check_euler = true;
  check_sphere = false;

  perform_algo = true;
  draw_pics = true;

  /*  emb_dim1 = false;
  emb_dim2 = false;
  check_cycles = false;
  check_euler = false;
  check_sphere = false;*/

  /*  string string_emb_dim1;
  while (string_emb_dim1!="y" && string_emb_dim1!="n") {
    string_emb_dim1="";
    //    cout << "Would you like me to check that the 1-skeleton is embedded? (there are " << edgePOrbits.size() << " P-orbits of edges)" << endl;
    cout << "Would you like me to check that the 1-skeleton is embedded?" << endl;
    cout << "[y/n] ";
    cin >> string_emb_dim1;
  }
  emb_dim1 = string_emb_dim1=="y";
  
  string string_emb_dim2;
  while (string_emb_dim2!="y" && string_emb_dim2!="n") {
    string_emb_dim2="";
    //    cout << "Would you like me to check that the 2-skeleton is embedded? (there are " << ridgePOrbits.size() << " P-orbits of ridges)" << endl;
    cout << "Would you like me to check that the 2-skeleton is embedded?" << endl;
    cout << "[y/n] ";
    cin >> string_emb_dim2;
  }
  emb_dim2 = string_emb_dim2=="y";

  string string_check_cycles;
  while (string_check_cycles!="y" && string_check_cycles!="n") {
    cout << "Would you like me to check the ridge cycles?" << endl;
    cout << "[y/n] ";
    string_check_cycles="";
    cin >> string_check_cycles;
  }
  check_cycles = string_check_cycles=="y";

  string string_check_euler;
  while (string_check_euler!="y" && string_check_euler!="n") {
    cout << "Would you like me to compute the orbifold Euler characteristic?" << endl;
    cout << "[y/n] ";
    string_check_euler="";
    cin >> string_check_euler;
  }
  check_euler=string_check_euler=="y";

  string string_check_sphere;
  while (string_check_sphere!="y" && string_check_sphere!="n") {
    cout << "Would you like me to check that the boundary is a 3-sphere?" << endl;
    cout << "[y/n] ";
    string_check_sphere = "";
    cin >> string_check_sphere;
  }
  check_sphere = string_check_sphere=="y";*/

  cout << "Initializing sporadic group..." << endl;

  word_p1.push_back(1);
  word_m1.push_back(-1);

  vector<int> w1;
  w1.push_back(1);
  vector<int> w2;
  w2.push_back(2);
  vector<int> w3;
  w3.push_back(3);

  mult = v;
  multconj = myconj(v);

  gen tauc = myconj(tau);
  gen a = 2-mult-multconj;
  gen b12 = (multconj-1)*tau;
  gen b13 = (1-multconj)*tauc;
  red(a);
  red(b12);
  red(b13);

  vecteur coh(9);
  coh[0] = a;           coh[1] = b12;         coh[2] = b13;
  coh[3] = myconj(b12); coh[4] = a;           coh[5] = b12;
  coh[6] = myconj(b13); coh[7] = myconj(b12); coh[8] = a;
  H = createMatrix(coh);

  vecteur cor1(9);
  cor1[0] = mult;  cor1[1] = tau;  cor1[2] = -tauc;
  cor1[3] = 0;  cor1[4] = 1;    cor1[5] = 0;
  cor1[6] = 0;  cor1[7] = 0;    cor1[8] = 1;
  R1 = createMatrix(cor1);
  matred(R1);

  vecteur cor2(9);
  cor2[0] = 1;  cor2[1] = 0;  cor2[2] = 0;
  cor2[3] = -tauc*mult;  cor2[4] = mult;    cor2[5] = tau;
  cor2[6] = 0;  cor2[7] = 0;    cor2[8] = 1;
  R2 = createMatrix(cor2);
  matred(R2);

  vecteur cor3(9);
  cor3[0] = 1;  cor3[1] = 0;    cor3[2] = 0;
  cor3[3] = 0;  cor3[4] = 1;    cor3[5] = 0;
  cor3[6] = tau*mult;  cor3[7] = -tauc*mult;    cor3[8] = mult;
  R3 = createMatrix(cor3);
  matred(R3);

  vecteur cor1i(9);
  cor1i[0] = multconj;  cor1i[1] = -tau*multconj;  cor1i[2] = tauc*multconj;
  cor1i[3] = 0;  cor1i[4] = 1;    cor1i[5] = 0;
  cor1i[6] = 0;  cor1i[7] = 0;    cor1i[8] = 1;
  R1i = createMatrix(cor1i);
  matred(R1i);

  vecteur cor2i(9);
  cor2i[0] = 1;  cor2i[1] = 0;  cor2i[2] = 0;
  cor2i[3] = tauc;  cor2i[4] = multconj;    cor2i[5] = -tau*multconj;
  cor2i[6] = 0;  cor2i[7] = 0;    cor2i[8] = 1;
  R2i = createMatrix(cor2i);
  matred(R2i);

  vecteur cor3i(9);
  cor3i[0] = 1;  cor3i[1] = 0;    cor3i[2] = 0;
  cor3i[3] = 0;  cor3i[4] = 1;    cor3i[5] = 0;
  cor3i[6] = -tau;  cor3i[7] = tauc;    cor3i[8] = multconj;
  R3i = createMatrix(cor3i);
  matred(R3i);

  vecteur coj(9);
  coj[0] = 0;  coj[1] = 0;    coj[2] = multconj;
  coj[3] = 1;  coj[4] = 0;    coj[5] = 0;
  coj[6] = 0;  coj[7] = 1;    coj[8] = 0;
  J = createMatrix(coj);

  vecteur coji(9);
  coji[0] = 0;  coji[1] = 1;    coji[2] = 0;
  coji[3] = 0;  coji[4] = 0;    coji[5] = 1;
  coji[6] = mult;  coji[7] = 0;    coji[8] = 0;
  Ji = createMatrix(coji);

  P = R1*J; // or R1*R2*R3?
  matred(P);

  Pi = Ji*R1i;
  matred(Pi);
  
  //  cout << "Signature of Hermitian form: " << checkSignature(H) << endl;

  /*  vector<int> wt;
  wt.push_back(2);
  wt.push_back(-3);
  wt.push_back(-2);
  wt.push_back(-1);
  wt.push_back(2);
  wt.push_back(3);
  wt.push_back(-2);
  wt.push_back(1);
  wt.push_back(2);
  wt.push_back(3);
  gen A = evalWord(wt);
  gen A3 = A*A;
  matred(A3);
  A3 = A3*A;
  matred(A3);
  gen M = A3-Id;
  gen zt = det(M);
  red(zt);
  cout << "det should be zero: " << zt << endl;
  gen te1 = _row(makesequence(M,0),&ct);
  gen te2 = _row(makesequence(M,1),&ct);
  vecteur te;
  te.push_back(te1[1]*te2[2]-te1[2]*te2[1]);
  te.push_back(te1[2]*te2[0]-te1[0]*te2[2]);
  te.push_back(te1[0]*te2[1]-te1[1]*te2[0]);
  gen X = gen(te);
  vectred(X);
  cout << "Mirror: " << X << endl;
  cout << " <X,X> = " << evali(SqNorm(X)) << endl;
  
  vector<int> wtt;
  wtt.push_back(-4);
  wtt.push_back(-4);
  wtt.push_back(-4);
  wtt.push_back(2);
  wtt.push_back(-3);
  wtt.push_back(-2);
  wtt.push_back(1);
  wtt.push_back(2);
  wtt.push_back(3);
  wtt.push_back(-2);
  gen B = evalWord(wtt);
  
  cout << "Order of B: " << order(B) << endl;

  gen Y = B*X;
  vectred(Y);
  cout << "Is in stab? " << areDep(X,Y) << endl;*/

  orderofp = order(P);
  cout << "R1*J = P has order " << orderofp << endl;
  
  if (tautag=="s5") {
    Ppower = Id;
    bool ir = false;
    orderofppower = 0;
    while (!ir && orderofppower<orderofp-1) {
      orderofppower++;
      Ppower = Ppower*P;
      matred(Ppower);
      gen dt = discrAdjusted(Ppower);
      ir = operator_equal(dt,zeronumber,&ct);
    }
    if (ir) {
      cout << "Power " << orderofppower << " of P is a complex reflection!" << endl;
      // not quite right! this does not work for s1 for instance...
      
      ///////////////
      // the following only works for s5!!!
      ///////////////
      gen te = gen("z^9-1",&ct);
      gen tenum = _evalf(makesequence(gen("exp(2*pi*i/9)",&ct),20),&ct);
      gen aa = findInField(te,tenum);
      red(aa);
      vecteur em;
      em.push_back(aa);
      em.push_back(onenumber);
      em.push_back(myconj(aa));
      extramirror = gen(em);
      cout << "This should give the polar vector to the mirror of P^" << orderofppower << endl;
      cout << extramirror << endl;
      cout << evali(extramirror) << endl;
      
    }
    else {
      cout << "No power of P is a complex reflection!" << endl;  
    }
  }

  vecteur coq;
  // this is done in a strange place, should define a global variable...
  if (tautag=="s4c") {
    coq.push_back(gen("exp(-2*pi*i/7)",&ct));
    coq.push_back(gen("exp(-8*pi*i/7)",&ct));
    coq.push_back(gen("exp(10*pi*i/7)",&ct));
    coq.push_back(onenumber);
    coq.push_back(onenumber);
    coq.push_back(onenumber);
    coq.push_back(gen("exp(2*pi*i/7)",&ct));
    coq.push_back(gen("exp(8*pi*i/7)",&ct));
    coq.push_back(gen("exp(-10*pi*i/7)",&ct));
  }
  else {
    if (tautag=="s1") {
      coq.push_back(gen("exp(pi*i/3)",&ct));
      coq.push_back(gen("exp(pi*i/12)",&ct));
      coq.push_back(gen("exp(-5*pi*i/12)",&ct));
      coq.push_back(onenumber);
      coq.push_back(onenumber);
      coq.push_back(onenumber);
      coq.push_back(gen("exp(-pi*i/3)",&ct));
      coq.push_back(gen("exp(-pi*i/12)",&ct));
      coq.push_back(gen("exp(5*pi*i/12)",&ct));
    }
    else {
      if (tautag=="s5") {
	coq.push_back(gen("exp(13*pi*i/45)",&ct));
	coq.push_back(gen("exp(2*pi*i/9)",&ct));
	coq.push_back(gen("exp(-23*pi*i/45)",&ct));
	coq.push_back(onenumber);
	coq.push_back(onenumber);
	coq.push_back(onenumber);
	coq.push_back(gen("exp(-13*pi*i/45)",&ct));
	coq.push_back(gen("exp(-2*pi*i/9)",&ct));
	coq.push_back(gen("exp(23*pi*i/45)",&ct));
      }
      else {
	cerr << "Unknown tau value.." << endl;
	exit(1);
      }
    }
  }
  Q = createMatrix(coq);
  //  cout << "Q=" << Q << endl;

  vecteur p0temp;
  cout << "Creating p0" << endl;
  for (unsigned k=0; k<3; k++) {
    p0temp.push_back(convert_interval(Q[k][0],20,&ct));
  }
  p0 = gen(p0temp);
  p0c = conj(p0,&ct);
  
  cout << "p0=" << p0 << endl;

  gen R123 = R1*R2*R3;
  matred(R123);
  
  int o123 = order(R123);
  vector<int> w123;
  for (unsigned k=0; k<o123; k++) {
    w123.push_back(1);
    w123.push_back(2);
    w123.push_back(3);
  }

  relations.push_back(w123);
  relations.push_back(flipword(w123));

  vector<int> w232b;
  w232b.push_back(2);
  w232b.push_back(3);
  w232b.push_back(-2);

  vector<int> w3b23;
  w3b23.push_back(-3);
  w3b23.push_back(2);
  w3b23.push_back(3);

  /*  vector<int> w1232b1b;
  w1232b1b.push_back(1);
  w1232b1b.push_back(2);
  w1232b1b.push_back(3);
  w1232b1b.push_back(-2);
  w1232b1b.push_back(-1);

  vector<int> wlong;
  wlong.push_back(1);
  wlong.push_back(2);
  wlong.push_back(3);
  wlong.push_back(-2);
  wlong.push_back(1);
  wlong.push_back(2);
  wlong.push_back(-3);
  wlong.push_back(-2);
  wlong.push_back(-1);*/

  createBraidRelation(word_p1,w2);
  createBraidRelation(word_p1,w232b);
  createBraidRelation(word_p1,w3b23);

  cout << "Relations:" << endl;
  for (unsigned k=0; k<relations.size(); k++) {
    cout << "  " << convertword(relations[k]) << endl;
  }

  /*  cout << "Check matrices..." << endl;
  cout << "R1 check: " << endl;
  gen te = transpose(matconj(R1))*H*R1-H;
  matred(te);
  cout << te << endl;
  cout << "R2 check: " << endl;
  te = transpose(matconj(R2))*H*R2-H;
  matred(te);
  cout << te << endl;
  cout << "R3 check: " << endl;
  te = transpose(matconj(R3))*H*R3-H;
  matred(te);
  cout << te << endl;
  cout << "J check: " << endl;
  te = transpose(matconj(J))*H*J-H;
  matred(te);
  cout << te << endl;*/

  e1w = WVector(e1,w1);
  e2w = WVector(e2,w2);
  e3w = WVector(e3,w3);

  /*
  bool is_arith = isArithmetic();
  if (is_arith) {
    cout << "Sorry, this group is arithmetic." << endl;
  }
  else {
    cout << "The group is NOT arithmetic!" << endl;
    }*/

  if (perform_algo) {
    cout << "Creating initial prism..." << endl;
    Prism A0 = createPrism(e1w,e2w,e3w);  
    // THIS STARTS THE ALGO
    computeVertices(A0);
    addFaceOrbit(A0);
    cout << "Expanding family of prisms..." << endl;
    expand();
  }

}

gen Polytope::Proj(gen v, gen w) {
  gen fa = inverse(Inn(w,w));
  if (operator_equal(fa,zeronumber,&ct)) {
    cerr << "Do you really mean to divide by zero?" << endl;
    exit(1);
  }
  else {
    gen te = _simplify(v-Inn(v,w)*fa*w,&ct);
    //    cout << "te=" << te << endl;
    vectred(te);  
    //    cout << "   reduced... " << te << endl;
    return te;
  }
}

gen Polytope::Inn(gen v, gen w) {
  gen res = vectconj(w)*H*v;
  red(res);
  return res;
}

gen Polytope::SqNorm(gen v) {
  gen res = vectconj(v)*H*v;
  red(res);
  return convertToReal(res);
}

gen Polytope::BoxProd(gen v, gen w) {
  gen X = vectconj(v)*H;
  gen Y = vectconj(w)*H;
  vecteur res(3);
  res[0] = X[1]*Y[2] - X[2]*Y[1];
  res[1] = X[2]*Y[0] - X[0]*Y[2];
  res[2] = X[0]*Y[1] - X[1]*Y[0];
  gen bp = gen(res);
  vectred(bp);
  return bp;
}

gen Polytope::refl(gen X, gen V, gen z) {
  gen te = Inn(V,V);
  if (te==zeronumber) {
    cerr << "Trying to reflect across a null vector!" << endl;
    exit(1);
  }
  else {
    gen yo = _simplify(X + (z-1)*Inn(X,V)*inverse(te)*V,&ct);
    vectred(yo);
    return yo;
  }
}

gen Polytope::reflMat(gen u, gen z) {
  vecteur te;
  gen te1 = refl(e1,u,z);
  gen te2 = refl(e2,u,z);
  gen te3 = refl(e3,u,z);
  for (unsigned k=0; k<3; k++) {
    te.push_back(te1[k]);
    te.push_back(te2[k]);
    te.push_back(te3[k]);
  }
  return createMatrix(te);
}

Prism Polytope::createPrism(WVector v0, WVector v1, WVector v2){  

  Prism A;
  
  A.jstart = 0;

  A.mirror = v0;

  A.sides.push_back(v1); 
  A.sides.push_back(v2); 

  vector<int> w; // provisoire, on trouvera w plus tard
  A.apex = WVector(BoxProd(v1.coords,v2.coords),w);

  A.apexProjection = Proj(A.apex.coords,A.mirror.coords);
  //  vectred(A.apexProjection);
  A.cspine = BoxProd(A.apex.coords,A.apexProjection);

  /*  cout << "Apex: " << A.apex << " sq norm: " << SqNorm(A.apex) << endl;
  cout << "Apex projection: " << A.apexProjection <<  " sq norm: " << evali(SqNorm(A.apexProjection)) << endl;
  cout << "Polar to complex spine: " << A.cspine <<  " sq norm: " << evali(SqNorm(A.cspine)) << endl;*/
 
  bool done = false;
  unsigned n = 2;
  while (!done && n<100) {
    //    cout << "TEST n=" << n << endl;
    gen M1 = reflMat(A.sides[0].coords,mult);
    vector<int> wl = conjugateword(A.sides[0].word,A.sides[1].word); 
    //   cout << "wl: " << wl << endl;
    gen vl = M1*A.sides[1].coords; 
    vectred(vl);
    WVector Vl(vl,wl);
    if (!testSame(Vl,A.sides[n-1])) {
      A.sides.insert(A.sides.begin(),Vl);
      n++;
      A.jstart++;
    }
    else {
      done = true;
    }
    if (!done) {
      gen M2 = reflMat(A.sides[n-1].coords,multconj);
      vector<int> wr = conjugateword(invertword(A.sides[n-1].word),A.sides[n-2].word);
      gen vr = M2*A.sides[n-2].coords;
      vectred(vr);
      WVector Vr(vr,wr);
      //      cout << "TEST vr=" << vr << endl;
      if (!testSame(Vr,A.sides[0])) {
	A.sides.push_back(Vr);
	n++;
      }
      else {
	done = true;
      }
    }
  }
  if (!done) {
    cerr << "Braiding cycle did not close up!" << endl;
    exit(1);
  }
  else {

    unsigned k1 = 0;
    unsigned l1 = A.sides[0].word.size();
    for (unsigned k=1; k<A.sides.size(); k++) {
      unsigned l = A.sides[k].word.size();
      if (l<l1) {
	l1 = l;
	k1 = k;
      }
    }
    unsigned k2;
    if (k1==0) {
      if (A.sides[A.sides.size()-1].word.size()<A.sides[1].word.size()) {
	k2 = A.sides.size()-1;
      }
      else {
	k2 = 1;
      }
    }
    else {
      k2 = (k1+1)%A.sides.size();
    }

    gen M1 = reflMat(A.sides[k1].coords,mult)*reflMat(A.sides[k2].coords,mult);
    matred(M1);
    int om1 = order(M1);
    cout << "Product of order " << om1 << endl;
    int bo = braidOrder(reflMat(A.sides[k1].coords,mult),reflMat(A.sides[k2].coords,mult));
    cout << "Braiding order " << bo << endl;

    vector<int> w0 = concat(A.sides[k1].word,A.sides[k2].word);
    vector<int> w = concat(A.sides[k1].word,A.sides[k2].word);

    bool isrefl = false;
    unsigned k=1;
    gen M = M1;
    while (k<om1 && !isrefl) {
      gen dt = discrAdjusted(M);
      isrefl = operator_equal(dt,zeronumber,&ct); // THIS IT NOT QUITE THE RIGHT TEST (COULD BE PARABOLIC)
      if (isrefl) {
	int sg;
	if (!checkSign(SqNorm(A.apex.coords),sg)) {
	  cerr << "Problem checking sign" << endl;
	  exit(1);
	}
	else {
	  if (sg>0) {
	    gen X1 = BoxProd(A.cspine,A.apex.coords);
	    gen X2 = M*X1;
	    vectred(X2);
	    isrefl = areDep(X1,X2);
	  }
	  else {
	    if (sg<0) {
	      gen X1 = A.apex.coords;
	      gen X2 = M*X1;
	      vectred(X2);
	      isrefl = areDep(X1,X2);
	    }
	    else {
	      // don't know what to do...
	    }
	  }
	}
      }
      if (isrefl) {
	cout << "Found reflection, power " << k << endl;
      }
      else {
	M = M*M1;
	w = concat(w,w0);
	matred(M);
      }
      k++;
    }
    if (!isrefl) {
      cout << "Did not find any reflection power..." << endl;
      //      exit(1);
    }

    useRelations(w);
    A.apex.word = w;
    cout << "word for apex=" << convertword(w) << endl;

    unsigned ns = A.sides.size();
    cout << "Constructed a pyramid with " << ns << " sides!" << endl;
    A.isSurrounded = false;
    for (unsigned k=0; k<ns; k++) {
      A.freeRidges.push_back(k);
      //      cout << A.sides[k].word << endl;
    }
    //    cout << "jstart: " << A.jstart << endl;
    return A;
  }

}

bool Polytope::isPrismNew(Prism A){
  unsigned k=0;
  bool res = true;
  while (res && k<faces.size()) {
    res = !comparePrisms(A,faces[k]);
    k++;
  }
  return res;
}

bool Polytope::isVertexNew(WVector V) {
  return !isInList(V,vertices);
}

bool Polytope::comparePrisms(Prism A, Prism B) {
  unsigned ns = A.sides.size();
  if (!testSame(A.mirror,B.mirror) || ns!=B.sides.size()) {
    //  if (!areDep(A.mirror.coords,B.mirror.coords) || ns!=B.sides.size()) {
    return false;
  }
  else {
    return samePolygon(A.sides,B.sides);
  }
}

bool Polytope::samePolygon(vector<WVector> V, vector<WVector> W){
  unsigned vs = V.size();
  if (W.size()!=vs) {
    return false;
  }
  else {
    bool foundsame = false;
    unsigned k=0;
    while (!foundsame && k<vs) {
      foundsame = testSame(V[0],W[k]);
      k++;
    }
    if (!foundsame) {
      return false;
    }
    else {
      //      cout << "Found common vector..." << endl;
      unsigned l=1;
      bool same = true;
      while (same && l<V.size()) {
	k = k%vs;
	same = testSame(V[l],W[k]);
	k++;
	l++;
      }
      return same;
    }
  }
}

void Polytope::updateFreeRidges(){
  unsigned nf = faces.size();
  unsigned ns = faces[nf-1].sides.size();
  for (unsigned k=0; k<ns; k++) {    
    // test if side #k is free...
    vector<WVector> ridge1;
    ridge1.push_back(faces[nf-1].sides[k]);
    ridge1.push_back(faces[nf-1].sides[(k+1)%ns]);
    ridge1.push_back(faces[nf-1].mirror);
    bool isfree = true;
    unsigned l = 0;
    while (isfree && l<nf-1) {
      if (!faces[l].isSurrounded) {
	unsigned m=0;
	bool foundNeighbor = false;
	while (!foundNeighbor && m<faces[l].freeRidges.size()) {
	  vector<WVector> ridge2;
	  unsigned mm = faces[l].freeRidges[m];
	  ridge2.push_back(faces[l].sides[mm]);
	  ridge2.push_back(faces[l].sides[(mm+1)%faces[l].sides.size()]);
	  ridge2.push_back(faces[l].mirror);
	  if (samePolygon(ridge1,ridge2)) {
	    //	    cout << "Same ridge..." << endl;
	    //	    cout << "Filling face #" << (nf-1) << endl;
	    faces[nf-1].fillRidge(k);
	    //	    cout << "Filling face #" << (l) << endl;
	    faces[l].fillRidge(mm);
	    foundNeighbor = true;
	    isfree = false;
	  }
	  m++;
	}
      }
      l++;
      //      cout << "l=" << l << endl;
    }
  }
  /*  for (unsigned k=0; k<faces.size(); k++) {
    if (faces[k].isSurrounded) {
      cout << "Face #" << k << " is surrounded!" << endl;      
    }
    }*/
}

Prism Polytope::applyReflection(Prism A) {

  gen M = reflMat(A.mirror.coords,mult);
  vector<int> w = A.mirror.word;

  Prism A1;
 
  gen m = M*A.mirror.coords;
  vectred(m);
  vector<int> w1 = conjugateword(w,A.mirror.word);
  A1.mirror = WVector(m,w1);

  A1.apex = WVector(M*A.apex.coords,conjugateword(w,A.apex.word));
  vectred(A1.apex.coords);

  A1.apexProjection = M*A.apexProjection;
  vectred(A1.apexProjection);

  A1.cspine = M*A.cspine;
  vectred(A1.cspine);

  // I think we want to switch, here, right?
  /*  A1.X0 = M*A.X1;
  A1.X1 = M*A.X0;
  vectred(A1.X0);
  vectred(A1.X1);*/

  for (unsigned k=0; k<A.sides.size(); k++) {
      vector<int> wk = conjugateword(w,A.sides[k].word);
      gen sk = M*A.sides[k].coords;
      vectred(sk);
      A1.sides.push_back(WVector(sk,wk));
  }

  for (unsigned k=0; k<A.bottomface.size(); k++) {
      vector<int> wk = conjugateword(w,A.bottomface[k].word);
      gen sk = M*A.bottomface[k].coords;
      vectred(sk);
      A1.bottomface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.topface.size(); k++) {
      vector<int> wk = conjugateword(w,A.topface[k].word);
      gen sk = M*A.topface[k].coords;
      vectred(sk);
      A1.topface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
      vector<int> wk = conjugateword(w,A.midvertices[k].word);
      gen sk = M*A.midvertices[k].coords;
      vectred(sk);
      A1.midvertices.push_back(WVector(sk,wk));
      A1.midvertindices.push_back(A.midvertindices[k]);
  }

  A1.isSurrounded = false;
  for (unsigned k=0; k<A1.sides.size(); k++) {
    A1.freeRidges.push_back(k);
  }
  A1.jstart = A.jstart;
 
  return A1;
  
}

Prism Polytope::applyInverseReflection(Prism A) {

  gen M = reflMat(A.mirror.coords,multconj);
  vector<int> w = A.mirror.word;

  Prism A1;
 
  gen m = M*A.mirror.coords;
  vectred(m);
  vector<int> w1 = conjugateword(w,A.mirror.word);
  A1.mirror = WVector(m,w1);

  A1.apex = WVector(M*A.apex.coords,conjugateword(w,A.apex.word));
  vectred(A1.apex.coords);
  A1.apexProjection = M*A.apexProjection;
  vectred(A1.apexProjection);
  A1.cspine = M*A.cspine;
  vectred(A1.cspine);
  
  // I think we want to switch, here, right?
  /*  A1.X0 = M*A.X1;
  A1.X1 = M*A.X0;
  vectred(A1.X0);
  vectred(A1.X1);*/

  for (unsigned k=0; k<A.sides.size(); k++) {
      vector<int> wk = conjugateword(w,A.sides[k].word);
      gen sk = M*A.sides[k].coords;
      vectred(sk);
      A1.sides.push_back(WVector(sk,wk));
  }

  for (unsigned k=0; k<A.bottomface.size(); k++) {
      vector<int> wk = conjugateword(w,A.bottomface[k].word);
      gen sk = M*A.bottomface[k].coords;
      vectred(sk);
      A1.bottomface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.topface.size(); k++) {
      vector<int> wk = conjugateword(w,A.topface[k].word);
      gen sk = M*A.topface[k].coords;
      vectred(sk);
      A1.topface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
      vector<int> wk = conjugateword(w,A.midvertices[k].word);
      gen sk = M*A.midvertices[k].coords;
      vectred(sk);
      A1.midvertices.push_back(WVector(sk,wk));
      A1.midvertindices.push_back(A.midvertindices[k]);
  }

  A1.isSurrounded = false;
  for (unsigned k=0; k<A1.sides.size(); k++) {
    A1.freeRidges.push_back(k);
  }
  A1.jstart = A.jstart;

  return A1;
  
}

/*
Prism Polytope::applyMatrix(gen M, Prism A) {
  // This will not keep track of words...

  Prism A1 = A.mycopy();
  A1.mirror.coords = M*A1.mirror.coords;
  vectred(A1.mirror.coords);

  vector<int> wvoid;
  A1.apex = WVector(M*A.apex.coords,wvoid); // best we can do?
  vectred(A1.apex.coords);
  A1.apexProjection = M*A1.apexProjection;
  vectred(A1.apexProjection);
  A1.cspine = M*A1.cspine;
  vectred(A1.cspine);

  // this is not good, but oh well..
  A1.X0 = M*A.X0;
  A1.X1 = M*A.X1;
  vectred(A1.X0);
  vectred(A1.X1);
  
  for (unsigned k=0; k<A1.sides.size(); k++) {
      gen sk = M*A1.sides[k].coords;
      vectred(sk);
      A1.sides[k].coords = sk;
  }

  for (unsigned k=0; k<A1.bottomface.size(); k++) {
      gen sk = M*A1.bottomface[k].coords;
      vectred(sk);
      A1.bottomface[k].coords = sk;
  }
  for (unsigned k=0; k<A1.topface.size(); k++) {
      gen sk = M*A1.topface[k].coords;
      vectred(sk);
      A1.topface[k].coords = sk;
  }
  for (unsigned k=0; k<A1.midvertices.size(); k++) {
      gen sk = M*A1.midvertices[k].coords;
      vectred(sk);
      A1.midvertices[k].coords = sk;
  }

  A1.isSurrounded = false;
  for (unsigned k=0; k<A1.sides.size(); k++) {
    A1.freeRidges.push_back(k);
  }
  A1.jstart = A.jstart;

  return A1;
  
  }*/


void Polytope::expand() {
  cout << "There are now " << faces.size() << " faces." << endl;
  if (faces.size()>300) {
    cout << "Found over 300 faces, are you sure you want to continue?!" << endl;
  }
  else {
    //  updateFreeRidges();
    bool closed = true;
    unsigned nf = faces.size();
    unsigned n = 0;
    while (closed && n < nf) {
      closed = faces[n].isSurrounded;
      n++;
    } 
    if (closed) {
      cout << "The polytope has no free ridge!" << endl;

      exportFaceData();

      // This will also find points of coequidistance (by intersecting complex spines)
      bool test = checkSide();
      if (test) {
	cout << "The fixed point of P is on the correct side of all faces" << endl;
      }
      else {
	cout << "The fixed point of P does not seem to be on the correct side (increasing precision may help)..." << endl;
      }

      checkPairings();

      computeAllVertices();      
      computeAllEdges();
      computeAllRidges();

      cout << endl;

      if (emb_dim1) {
	cout << "Will now check that the 1-skeleton is embedded (there are " << edgePOrbits.size() << " orbits of edges)" << endl;
	for (unsigned k=0; k<edgePOrbits.size(); k++) {
	  checkEdge(edgePOrbits[k][0]);
	}
	cout << "If you read this, the 1-skeleton is embedded!" << endl;
	cout << endl;
      }

      if (check_cycles) {
	checkCycles();
	cout << endl;
      }

      if (check_euler) {
	computeEulerCharacteristic();
	cout << endl;
      }

      if (check_sphere) {
	checkLinks();
	checkSphere();
	cout << endl;
      }

      if (emb_dim2) {
	cout << "Will now check that G-polygons defining ridges are embedded" << endl;
	cout << "   be patient, there are " << ridgePOrbits.size() << " P-orbits of ridges..." << endl;

	if (tag=="4s5") {
	  cout << "For this group, the number field is large (degree 24), I will switch to the smallest possible field for critical points (degree 8)" << endl;
	  cout << "This requires re-constructing the polytope without using P, this may take some time." << endl;
	  PolytopeSmall F = PolytopeSmall("4s5b");
	  for (unsigned k=0; k<ridgePOrbits.size(); k++) {
	    cout << "Ridge P-orbit #" << k << ", on faces ";
	    int jf =  ridges[ridgePOrbits[k][0]].faceindices[0];
	    int kf =  ridges[ridgePOrbits[k][0]].faceindices[1];
	    cout << jf << " and " << kf << endl;
	    F.checkRidge(jf,kf);
	  }
	}
	else {
	  for (unsigned k=0; k<ridgePOrbits.size(); k++) {
	    cout << "Ridge P-orbit #" << k << "..." << endl;
	    checkRidge(ridgePOrbits[k][0]);
	  }
	}

	cout << "     " << endl;
	cout << " * * " << endl;
	cout << " \\_/ " << endl;
	cout << "     " << endl;
	cout << "If you read this, the program checked that the polytope for " << tag << " is embedded!" << endl;
	cout << " Used at most " << ndrecord << " digits to perform the computations." << endl;
	cout << endl;

      }

      /*   string s;
      cout << "Would you like me to check that the 1-skeleton is embedded? (there are " << edgePOrbits.size() << " P-orbits of edges)" << endl;
      cout << "[y/n] ";
      cin >> s;
      if (s=="y") {
      for (unsigned k=0; k<edgePOrbits.size(); k++) {
	  checkEdge(edgePOrbits[k][0]);
      }
      cout << "If you read this, the 1-skeleton is embedded!" << endl;
      }*/
      
      /*      cout << "Would you like me to check that the 2-skeleton is embedded? (there are " << ridgePOrbits.size() << " P-orbits of ridges)" << endl;
      cout << "[y/n] ";
      string s1;
      cin >> s1;
      if (s1=="y") {
	cout << "Will now check that G-polygons defining ridges are embedded" << endl;
	cout << "   be patient, there are " << ridgePOrbits.size() << " P-orbits of ridges..." << endl;

	if (tag=="4s5") {
	  cout << "For this group, the number field is large (degree 24), I will switch to the smallest possible field for critical points (degree 8)" << endl;
	  cout << "This requires re-constructing the polytope without using P, this may take some time." << endl;
	  PolytopeSmall F = PolytopeSmall("4s5b");
	  for (unsigned k=0; k<ridgePOrbits.size(); k++) {
	    cout << "Ridge P-orbit #" << k << ", on faces ";
	    int jf =  ridges[ridgePOrbits[k][0]].faceindices[0];
	    int kf =  ridges[ridgePOrbits[k][0]].faceindices[1];
	    cout << jf << " and " << kf << endl;
	    F.checkRidge(jf,kf);
	  }
	}
	else {
	  for (unsigned k=0; k<ridgePOrbits.size(); k++) {
	    cout << "Ridge P-orbit #" << k << "..." << endl;
	    checkRidge(ridgePOrbits[k][0]);
	  }
	}

	cout << "     " << endl;
	cout << " * * " << endl;
	cout << " \\_/ " << endl;
	cout << "     " << endl;
	cout << "If you read this, the program checked that the polytope for " << tag << " is embedded!" << endl;
	cout << "Had to use " << ndrecord << " digits to perform some computations." << endl;
	
	}*/

	/*      cout << "Would you like me to check the ridge cycles?" << endl;
      cout << "[y/n] ";
      string s2;
      cin >> s2;
      if (s2=="y") {
	checkCycles();
    }*/

	/*      cout << "Would you like me to compute the orbifold Euler characteristic?" << endl;
      cout << "[y/n] ";
      string s3;
      cin >> s3;
      if (s3=="y") {
	computeEulerCharacteristic();
	}*/

      //      drawRidgePic(ridgePOrbits[4][0]);
      
      cout << endl;

      bool is_arith = isArithmetic();
      if (is_arith) {
	cout << "Sorry, this group is arithmetic." << endl;
      }
      else {
	cout << "The group is NOT arithmetic!" << endl;
      }
      
      cout << endl;

      cout << "Done studying group " << tag << endl;
      cout << "If you read this message, all requested checks were successful:" << endl;
      if (emb_dim1) {
	cout << "    1-skeleton is embedded" << endl;
      }
      if (emb_dim2) {
	cout << "    2-skeleton is embedded" << endl;
      }
      if (check_sphere) {
	cout << "    The boundary of the polytope is a sphere" << endl;
      }
      if (check_cycles) {
	cout << "    Hypotheses of the Poincaré polyhedron theorem hold" << endl;
      }
      if (check_euler) {
	cout << "    I computed the Euler characteristic, it is " << euler_char << endl;
	cout << "     (the least common multiple of orders of stabilizers is " << lcm << ")" << endl;
      }

      cout << endl;
      
      // do this only to produce a picture of the polyhedron
      if (draw_pics) {
	cout << "I will now draw pictures of the polyhedron, in projections onto the axes of P, one for each axis" << endl;
	cout << "The pictures will be stored in in the pics directory, they need to be compiled with asymptote." << endl;

	drawProjection(1);
	drawProjection(2);

	for (unsigned k=0; k<facePOrbits.size(); k++) {
	  drawFace(facePOrbits[k][0]);
	}
      }

    }
    else {

      //      cout << "faces[" << (n-1) << "].freeRidges: " << faces[n-1].freeRidges << endl;
      unsigned k = faces[n-1].freeRidges[0];
      cout << "Creating two new prisms: " << endl;

      Prism A1 = createPrism(faces[n-1].sides[(k+1)%faces[n-1].sides.size()], faces[n-1].mirror, faces[n-1].sides[k]);
      cout << "A1: " << A1 << endl;

      Prism A2 = createPrism(faces[n-1].sides[k], faces[n-1].sides[(k+1)%faces[n-1].sides.size()], faces[n-1].mirror);
      cout << "A2: " << A2 << endl;

      gen X1 = BoxProd(faces[n-1].cspine,A1.cspine);
      gen X2 = BoxProd(faces[n-1].cspine,A2.cspine);
      // this will fail if two complex spines coincide?
      if (operator_equal(X1,zerovector,&ct) || operator_equal(X2,zerovector,&ct)) {
	cerr << "Two complex spines coincide, did not find point of coequidistance..." << endl;
	exit(1);
      }

      int sx1;
      if (checkSign(SqNorm(X1),sx1)) {
	if (isOnPrism(X1,faces[n-1]) || isOnPrism(X1,A1)) {
	  cerr << "The complex spines intersect inside CH^2, but on one of the real spines!" << endl;
	  exit(1);
	}
	if (sx1>=0) {
	  cout << "Sign of box product: " << sx1 << endl;
	  cerr << "Not coequidistant, just wanted to warn you ;)" << endl;
	  //	  exit(1);
	}
      }
      else {
	cerr << "Problem checking sign" << endl;
	exit(1);
      }

      cout << "Coequidistant" << endl;

      gen Y1 = reflectAcrossSpine(X1,faces[n-1]);
      gen Z1 = reflectAcrossSpine(X1,A1);
      //      A1.X0 = X1;
      //      A1.X1 = Z1;
      
      gen Y2 = reflectAcrossSpine(X2,faces[n-1]);
      gen Z2 = reflectAcrossSpine(X2,A2);

      gen t1v = abs(p0c*evali(H)*evali(X1),&ct);
      gen t2v = abs(p0c*evali(H)*evali(Y1),&ct);
      gen t3v = abs(p0c*evali(H)*evali(Z1),&ct);
      
      gen t4v = abs(p0c*evali(H)*evali(X2),&ct);
      gen t5v = abs(p0c*evali(H)*evali(Y2),&ct);
      gen t6v = abs(p0c*evali(H)*evali(Z2),&ct);

      cout << t1v << endl;
      cout << t2v << endl;
      cout << t3v << endl;
	  
      cout << t4v << endl;
      cout << t5v << endl;
      cout << t6v << endl;

      //      if (is_greater(_left(t2,&ct),_right(t1,&ct),&ct)) {
      if (is_greater(_left(t2v,&ct),_right(t1v,&ct),&ct) && is_greater(_left(t3v,&ct),_right(t1v,&ct),&ct)) {
	cout << "Choosing A1" << endl;
	if (isPrismNew(A1)) {
	  useRelations(A1);
	  cout << "Adding new face.." << endl;
	  computeVertices(A1);
	  addFaceOrbit(A1);
	}
      }
      else {
	//	if (is_greater(_left(t5,&ct),_right(t4,&ct),&ct)) {
	if (is_greater(_left(t5v,&ct),_right(t4v,&ct),&ct) && is_greater(_left(t6v,&ct),_right(t4v,&ct),&ct)) {
	  cout << "Choosing A2" << endl;
	  if (isPrismNew(A2)) {
	    useRelations(A2);
	    cout << "Adding new face.." << endl;
	    computeVertices(A2);
	    addFaceOrbit(A2);
	  }
	}
	else {	  
	  cerr << "Undecided between A1 and A2, will try with relaxed condition..." << endl;
	  //	  exit(1);

	  bool test1 = is_greater(_left(t2v,&ct),_right(t1v,&ct),&ct) || is_greater(_left(t3v,&ct),_right(t1v,&ct),&ct);
	  bool test2 = is_greater(_left(t5v,&ct),_right(t4v,&ct),&ct) || is_greater(_left(t6v,&ct),_right(t4v,&ct),&ct);

	  if (test1 && test2) {
	    cerr << "Really undecided, don't know what to do..." << endl;
	  }
	  else {
	    if (test1) {
	      cout << "Choosing A2" << endl;
	      if (isPrismNew(A2)) {
		cout << "Adding new face.." << endl;
		computeVertices(A2);
		addFaceOrbit(A2);
	      }
	    }
	    else {
	      if (test2) {
		cout << "Choosing A1" << endl;
		if (isPrismNew(A1)) {
		  cout << "Adding new face.." << endl;
		  computeVertices(A1);
		  addFaceOrbit(A1);
		}
	      }
	      else {	  
		cerr << "Both prisms are bad, don't know what to do..." << endl;
	      }
	    }
	  }

	}
      }

      expand();
    }    
  }
}

gen Polytope::eulerdim0() {

  gen ed0 = zeronumber;
  vector<vector<int> > vertexOrbits;
  for (unsigned k=0; k<vertexPOrbits.size(); k++) {
    vector<int> ok;
    ok.push_back(k);
    vertexOrbits.push_back(ok);
  }

  for (unsigned k=0; k<vertexPOrbitPairings.size(); k++) {
    cout << vertexPOrbitPairings[k][0] << " -> " << vertexPOrbitPairings[k][1] << endl;
    cout << "   via " << convertword(vertexPOrbitPairingMaps[k]) << endl;

    int u1 = -1;
    bool fd1 = false;
    while (!fd1 && u1!=vertexOrbits.size()-1) {
      u1++;
      int v1 = -1;
      while (!fd1 && v1!=vertexOrbits[u1].size()-1) {
	v1++;
	fd1 = vertexPOrbitPairings[k][0]==vertexOrbits[u1][v1];
      }
    }
    int u2 = -1;
    bool fd2 = false;
    while (!fd2 && u2!=vertexOrbits.size()-1) {
      u2++;
      int v2 = -1;
      while (!fd2 && v2!=vertexOrbits[u2].size()-1) {
	v2++;
	fd2 = vertexPOrbitPairings[k][1]==vertexOrbits[u2][v2];
      }
    }
    if (!fd1 || !fd1) {
      cerr << "Problem identifiying indices in vertexPOrbits..." << endl;
      exit(1);
    }
    else {
      if (u1!=u2) {
	for (unsigned z=0; z<vertexOrbits[u2].size(); z++) {
	  vertexOrbits[u1].push_back(vertexOrbits[u2][z]);
	}
	vertexOrbits.erase(vertexOrbits.begin()+u2);
      }
    }
    // probably better to construct the graph right away, to compute its connected components, then the fund group of each component...
    //   (I will not do that for now, this is not the time-consuming part of the program)

  }
  cout << "Vertex orbits (under side pairings, * denotes an ideal vertex): " << endl;
  for (unsigned z=0; z<vertexOrbits.size(); z++) {
    cout << "[";
    cout << convertword(vertexOrbits[z]) << "]";
    if (vertices[vertexPOrbits[vertexOrbits[z][0]][0]].faceindices[0]==-1) {
      cout << "*";
    }
    cout << ",  ";
  }
  cout << endl;
  cout << "There are " << vertexOrbits.size() << " orbits of vertices" << endl;

  cout << "Still need to compute the order of stabilizer (only when finite vertex!)" << endl;
  for (unsigned z=0; z<vertexOrbits.size(); z++) {
    vector<int> vo = vertexOrbits[z];

    bool needpowerofp;
    int baseptindex=0;
    if (tautag=="s5") {
      needpowerofp = false;
      int bpi = -1;
      while (!needpowerofp && bpi!=vo.size()-1) {
	bpi++;
	gen X = vertices[vertexPOrbits[vo[bpi]][0]].coords;
	needpowerofp = operator_equal(Inn(extramirror,X),zeronumber,&ct);
      }
      if (needpowerofp) {
	baseptindex = bpi;
	cout << "Some element in this orbit (index " << bpi << ") is on mirror of power of P!" << endl;
      }
    } 

    cout << "Computing stabilizer of vertex P-Orbit #" << vo[baseptindex] << endl;
    vector<vector<int> > stab;
    if (vo.size()>1) {
      vector<vector<int> > graph_edges;
      vector<vector<int> > maps;
      for (unsigned uu=0; uu<vertexPOrbitPairings.size(); uu++) {
	if (find(vo.begin(),vo.end(),vertexPOrbitPairings[uu][0])!=vo.end()) {
	  graph_edges.push_back(vertexPOrbitPairings[uu]);
	  maps.push_back(vertexPOrbitPairingMaps[uu]);
	}
      }
      sgraph G(graph_edges);
      
      vector<vector<int> > stabgens;
      vector<vector<bool> > orientations;
      G.fundamental_group(vo[baseptindex],stabgens,orientations);
      for (unsigned kk=0; kk<stabgens.size(); kk++) {
	vector<int> yo;
	for (unsigned ll=0; ll<stabgens[kk].size(); ll++) {
	  if (orientations[kk][ll]) {
	    yo = concat(maps[stabgens[kk][ll]],yo);
	  }
	  else {
	    yo = concat(invertword(maps[stabgens[kk][ll]]),yo);
	  }
	}
	stab.push_back(yo);
      }
    }
    else {
      // there is a single vertex, won't use graph routine for this (just take all pairings as generators)
      for (unsigned uu=0; uu<vertexPOrbitPairings.size(); uu++) {
	if (find(vo.begin(),vo.end(),vertexPOrbitPairings[uu][0])!=vo.end()) {
	  stab.push_back(vertexPOrbitPairingMaps[uu]);
	}
      }
    }
    cout << "Generators of stab:" << endl;
    vector<gen> stabgens;
    vector<int> trivial_elts;
    for (unsigned y=0; y<stab.size(); y++) {
      if (stab[y].size()>0) {
	cout << "  " << convertword(stab[y]) << endl;
	gen M = evalWord(stab[y]);
	if (!isScalar(M)) {
	  stabgens.push_back(M);
	  gen Z1 = vertices[vertexPOrbits[vo[baseptindex]][0]].coords;
	  gen Z2 = M*Z1;
	  vectred(Z2);
	  if (areDep(Z1,Z2)) {
	    cout << "Check!" << endl;
	  }
	  else {
	    cout << "Problem, matrix is actually not in stab!!" << endl;	    
	  }
	}
	else {
	  cout << convertword(stab[y]) << " is identity" << endl;
	  trivial_elts.push_back(y);
	}
      }
    }
    //    cout << "stab.size() before erasure: " << stab.size() << endl;
    for (int z=trivial_elts.size()-1; z>=0; z--) {
      stab.erase(stab.begin()+trivial_elts[z]);
    }
    cout << endl;
    //    cout << "stab.size() after erasure: " << stab.size() << endl;

    /*    cout << "Double check stab elts" << endl;
    for (unsigned r=0; r<stab.size(); r++) {
      cout << convertword(stab[r]) << endl;
    }
    cout << "done.." << endl;*/
        
    if (vertices[vertexPOrbits[vo[baseptindex]][0]].faceindices[0]!=-1) {
      // the commented out method runs slow, to be used only for small groups
      /*	cout << "Creating finite group.." << endl;
		FiniteGroup L(tag);
		L.addElements(stabgens);
		L.expandToGroup();*/
       
      gen og;
      gen ogmax = zeronumber;
      bool needpowerofp;
      if (tautag=="s5") { // shouldn't we do this also for s1?
	needpowerofp = operator_equal(Inn(extramirror,vertices[vertexPOrbits[vo[0]][0]].coords),zeronumber,&ct);
      }
      else {
	needpowerofp = false;
      }

      if (needpowerofp) {
      	cout << "THIS POINT IS ON THE MIRROR OF P^" << orderofppower << endl;
      }

      /*
       * SHOULD WORK WITH 2x2 MATRICES HERE!
       *   may also want to use the fact that we have a candidate for the group, where we know the center?
       *
       */

      if (tag!="5s4c" && tag!="3s5") {
	
	vector<gen> sg2d;
	// would like to find a mirror of reflection first
	gen mii;
	bool foundrefl = false;
	int zz=-1;      
	while (!foundrefl && zz!=stab.size()-1) {
	  zz++;
	  vector<int> w = stab[zz];
	  if (w.size()%2==1) {
	    int n = (w.size()-1)/2;
	    bool is_pal = true;
	    int uu=0;
	    while (is_pal && uu<n) {
	      is_pal = (w[uu]==-w[w.size()-1-uu]);
	      uu++;
	    }
	    foundrefl = is_pal;
	  }
	}
	if (foundrefl) {
	  cout << "Found a reflection: " << convertword(stab[zz]) << endl;
	  int n = (stab[zz].size()-1)/2;
	  mii = _col(makesequence(Id,abs(stab[zz][n])-1),&ct);
	  if (n>0) {
	    vector<int> w;
	    for (unsigned jj=0; jj<n; jj++) {
	      w.push_back(stab[zz][jj]);
	    }
	    mii = evalWord(w)*mii;
	    vectred(mii);
	  }
	  cout << "mirror=" << mii << endl;
	}
	else {
	  // use any vector and do Gramm-Schmidt?
	  mii = e1; // (this will not always work?!)
	}

      //      cout << "Test mirror: " << Inn(vertices[vertexPOrbits[vo[baseptindex]][0]].coords,mi) << endl;

	vecteur basis;
	basis.push_back(vertices[vertexPOrbits[vo[baseptindex]][0]].coords);
	basis.push_back( mii );
	basis.push_back(BoxProd(basis[0],basis[1]));
	
	gen A = _tran(gen(basis),&ct);
	gen Ai = matinverse(A);
	for (unsigned tt=0; tt<stabgens.size(); tt++) {
	  gen M = Ai*stabgens[tt]*A;
	  matred(M);
	  //	cout << "M=" << M << endl;
	  gen coe = inverse(M[0][0]);
	  vecteur r1;
	  r1.push_back(M[1][1]*coe);
	  r1.push_back(M[1][2]*coe);
	  vecteur r2;
	  r2.push_back(M[2][1]*coe);
	  r2.push_back(M[2][2]*coe);
	  vecteur rows;
	  rows.push_back(r1);
	  rows.push_back(r2);
	  gen R = gen(rows);
	  matred(R);
	  sg2d.push_back(R);
	}
      // sg2d should provide 2x2 version of the stab generators!

      //      cout << "sg2d:" << endl;
      //      for (unsigned mm=0; mm<sg2d.size(); mm++) {
      //	cout << sg2d[mm] << endl;
      //     }

	cout << "I'm not quite sure how to guess the order of stabilizer, will use brute force from incidence graph (this may take some time!)" << endl;
	cout << "Creating finite group (will first conjugate to linear group of 2x2 matrices).." << endl;
      //      FiniteGroup L(tag);
      //      L.addElements(stabgens);
      //      L.expandToGroup();

	FiniteLinearGroup L(tag);
	L.addElements(sg2d);
	L.expandToGroup();
	
	cout << "Stab has order " << L.elements.size() << endl;
	ogmax = L.elements.size();    

      }
      else {

	cout << "This group has a vertex stabilizer of very high order!" << endl;
	cout << "I will use the order of the known stabilizer, this is not valid in all generality (but it is in this case)" << endl;

	vector<WVector> mi = findMirrors(vertices[vertexPOrbits[vo[baseptindex]][0]]);
	vector<gen> stabgens_guess;
	for (unsigned kk=0; kk<mi.size(); kk++) {
	  cout << "Mirror " << convertword(mi[kk].word) << " contains vertex" << endl;
	  stabgens_guess.push_back(evalWord(mi[kk].word));
	}
	for (unsigned kk=0; kk<mi.size(); kk++) {
	  stabgens_guess.push_back(evalWord(invertword(mi[kk].word)));
	}
	int nsg = stabgens_guess.size()/2;

	unsigned kk_max=0;
	unsigned ll_max=0;
	for (unsigned kk=0; kk<nsg; kk++) {
	  int ok = order(stabgens_guess[kk]);
	  //	  cout << "ok=" << ok << endl;
	  for (unsigned ll=kk+1; ll<nsg; ll++) {
	    int ol = order(stabgens_guess[ll]);
	    //	    cout << "ol=" << ol << endl;
	    int te = braidOrder(stabgens_guess[kk],stabgens_guess[ll],20); // 10 is small, but it works for sporadic groups...
	    if (te>-1) {
	      cout << "Stab gens #" << kk << " (order " << ok << ") and " << ll << " (order " << ol << ") braid with order " << te << endl;
	      if (te>2) {
		og = gen("8",&ct);
		gen den = onenumber/ok + onenumber/ol + twonumber/te - onenumber; // formula from Mostow, say
		og = og/(te*den*den);
		cout << "  these generate a group of order " << og << endl;
	      }
	      else {
		if (te==2) {
		  og = ok*ol;
		}
		else {
		  cerr << "Braiding of order 0 or 1??" << endl;
		  exit(1);
		}
	      }
	      if (is_greater(og,ogmax,&ct)) {
		ogmax = og;
		kk_max = kk;
		ll_max = ll;
	      } 
	    }
	  }
	}
	cout << "Order of stab should be " << ogmax << endl;
	cout << "  (should check that the generators from incidence graph are all in the guessed stabilizer)" << endl;
	
      }
    
      cout << endl;
      lcm = _lcm(makesequence(lcm,ogmax),&ct);
      ed0 = ed0+onenumber/ogmax;
    }
  }
  return ed0;

}

bool Polytope::expressAsWord(gen M, vecteur list, gen fixed_pt, gen mirror, vector<int> &inds) {

  unsigned ndsave = nd;
  increasePrecision();

  inds.clear();
  gen other_vect = BoxProd(fixed_pt,mirror);
  vecteur ov;
  ov.push_back(fixed_pt);
  ov.push_back(mirror);
  ov.push_back(other_vect);
  gen O = _tran(gen(ov),&ct);
  // then need its inverse...
  gen detO = det(O);
  red(detO);
  if (operator_equal(detO,zeronumber,&ct)) {
    cerr << "Problem in express as word, I expected a basis of C^3" << endl;
    exit(1);
  }
  else {
    vecteur coi;
    coi.push_back(O[1][1]*O[2][2]-O[1][2]*O[2][1]);
    coi.push_back(-O[1][0]*O[2][2]+O[1][2]*O[2][0]);
    coi.push_back(O[1][0]*O[2][1]-O[1][1]*O[2][0]);
    coi.push_back(-O[0][1]*O[2][2]+O[0][2]*O[2][1]);
    coi.push_back(O[0][0]*O[2][2]-O[0][2]*O[2][0]);
    coi.push_back(-O[0][0]*O[2][1]+O[0][1]*O[2][0]);
    coi.push_back(O[0][1]*O[1][2]-O[0][2]*O[1][1]);
    coi.push_back(-O[0][0]*O[1][2]+O[0][2]*O[1][0]);
    coi.push_back(O[0][0]*O[1][1]-O[0][1]*O[1][0]);
    gen Oi = _tran(createMatrix(coi),&ct)*inverse(detO);
    matred(Oi);
    gen te = O*Oi;
    matred(te);
    cout << "Check, should be Id: " << te << endl;
  
    gen On = matevali(O);
    gen Oin = matevali(Oi);

    cout << "Oin: " << Oin << endl;

    bool found = false;
    int count = 0;
    gen base_pt = p0;
    gen goal = conj(Oin*(matevali(M)*p0),&ct);
    goal = goal/goal[0];
    cout << "goal: " << goal << endl;
    gen den = sqrt(re(goal[1]*conj(goal[1],&ct)+goal[2]*conj(goal[2],&ct),&ct),&ct);

    //    cout << "den: " << den << endl;

    gen C = Id;
    while (!found && count<100) {
      gen max = convert_interval(-1,nd,&ct);
      //     cout << "max: " << max << endl;
      int lval;
      for (unsigned l=0; l<list.size(); l++) {
	gen U = Oin*(matevali(list[l])*base_pt);
	U = U/U[0];
	gen den2 = sqrt(re(U[1]*conj(U[1],&ct)+U[2]*conj(U[2],&ct),&ct),&ct);
	gen val = re(U[1]*goal[1]+U[2]*goal[2],&ct)/(den*den2);
	cout << "val #" << l << ":" << val << endl;
	if (is_greater(val,max+prec,&ct)) {
	  max = val;
	  lval = l;
	}
      }
      inds.insert(inds.begin(),lval);
      cout << "inds=" << convertword(inds) << endl;
      base_pt = matevali(list[lval])*base_pt;
      C = list[lval]*C;
      matred(C);
      found = isMultiple(C,M);
      count++;
    }
    return found;
  }
  nd = ndsave;
  lowerPrecision();
}

gen Polytope::eulerdim1() {

  gen res = zeronumber;

  vector<vector<int> > edgePOrbitPairings;
  vector<vector<int> > edgePOrbitPairingMaps;
  
  vector<vector<int> > edgeOrbits;
  for (unsigned k=0; k<edgePOrbits.size(); k++) {
    vector<int> ok;
    ok.push_back(k);
    edgeOrbits.push_back(ok);
    vector<int> inds = edges[edgePOrbits[k][0]].faceindices;
    for (unsigned l=0; l<inds.size(); l++) {
      int image = applyPairingToEdge(inds[l],edgePOrbits[k][0]);
      cout << "Pairing for face #" << inds[l] << " maps edge #" <<  edgePOrbits[k][0] << " to edge #" << image << endl;
      int m=-1;
      int n;
      bool fd = false;
      while (!fd && m!=edgePOrbits.size()-1) {
	m++;
	n = -1;
	while (!fd && n!=edgePOrbits[m].size()-1) {
	  n++;
	  fd = (image==edgePOrbits[m][n]);
	}
      }
      if (!fd) {
	cerr << "Problem identifying image edge in a P-orbit" << endl;
	exit(1);
      }
      else {
	cout << "This gives a map from P-orbit #" << k << " to P-orbit #" << m << ", at index " << n << endl;
	vector<int> pa;
	pa.push_back(k);
	pa.push_back(m);
	edgePOrbitPairings.push_back(pa);
	vector<int> w;
	if (faces[inds[l]].invertpairing) {
	  w = invertword(faces[inds[l]].mirror.word);
	}
	else {
	  w = concat(w,faces[inds[l]].mirror.word);
	}
	if (n>orderofp/2) {
	  for (unsigned zz=0; zz<orderofp-n; zz++) {
	    w.insert(w.begin(),4);
	  }
	}
	else {
	  for (unsigned zz=0; zz<n; zz++) {
	    w.insert(w.begin(),-4);
	  }
	}	 
	cout << "Map: " << convertword(w) << endl;
	edgePOrbitPairingMaps.push_back(w);
      }
    }
  }

  for (unsigned k=0; k<edgePOrbitPairings.size(); k++) {
    int u1 = -1;
    bool fd1 = false;
    while (!fd1 && u1!=edgeOrbits.size()-1) {
      u1++;
      int v1 = -1;
      while (!fd1 && v1!=edgeOrbits[u1].size()-1) {
	v1++;
	fd1 = edgePOrbitPairings[k][0]==edgeOrbits[u1][v1];
      }
    }
    int u2 = -1;
    bool fd2 = false;
    while (!fd2 && u2!=edgeOrbits.size()-1) {
      u2++;
      int v2 = -1;
      while (!fd2 && v2!=edgeOrbits[u2].size()-1) {
	v2++;
	fd2 = edgePOrbitPairings[k][1]==edgeOrbits[u2][v2];
      }
    }
    if (!fd1 || !fd1) {
      cerr << "Problem identifiying indices in vertexPOrbits..." << endl;
      exit(1);
    }
    else {
      if (u1!=u2) {
	for (unsigned z=0; z<edgeOrbits[u2].size(); z++) {
	  edgeOrbits[u1].push_back(edgeOrbits[u2][z]);
	}
	edgeOrbits.erase(edgeOrbits.begin()+u2);
      }
    }
  }

  cout << "Edge orbits (under side pairings): " << endl;
  for (unsigned z=0; z<edgeOrbits.size(); z++) {
    cout << "[";
    cout << convertword(edgeOrbits[z]) << "]";
    cout << ",  ";
  }
  cout << endl;
  cout << "There are " << edgeOrbits.size() << " orbits of edges" << endl;


  cout << "Still need to compute the order of stabilizer..." << endl;
  for (unsigned z=0; z<edgeOrbits.size(); z++) {

    vector<int> vo = edgeOrbits[z];

    bool needpowerofp = false;
    gen Xo, Xe;
    int baseptindex = 0;
    if (tautag=="s5") {
      needpowerofp = false;
      int bpi = -1;
      while (!needpowerofp && bpi!=vo.size()-1) {
	bpi++;
	Xo = vertices[edges[edgePOrbits[vo[bpi]][0]].origin].coords;
	Xe = vertices[edges[edgePOrbits[vo[bpi]][0]].end].coords;
	needpowerofp = operator_equal(Inn(extramirror,Xo),zeronumber,&ct) && operator_equal(Inn(extramirror,Xe),zeronumber,&ct);
      }
      if (needpowerofp) {
	baseptindex = bpi;
	cout << "Some element in this orbit (index " << bpi << ") is on mirror of P^" << orderofppower << endl;
	cout << "   (it joins v#" << edges[edgePOrbits[vo[bpi]][0]].origin << " and v#" << edges[edgePOrbits[vo[bpi]][0]].end << ")" << endl;
      }
      else {
	cout << "No element in this orbit is on mirror of power of P!" << endl;
	Xo = vertices[edges[edgePOrbits[vo[baseptindex]][0]].origin].coords;
	Xe = vertices[edges[edgePOrbits[vo[baseptindex]][0]].end].coords;
      }
    }
    else {
      // not s5
      Xo = vertices[edges[edgePOrbits[vo[baseptindex]][0]].origin].coords;
      Xe = vertices[edges[edgePOrbits[vo[baseptindex]][0]].end].coords;
    }

    cout << "Computing stabilizer of edge P-Orbit #" << vo[baseptindex] << endl;
    vector<vector<int> > stab;
    if (vo.size()>1) {
      vector<vector<int> > graph_edges;
      vector<vector<int> > maps;
      for (unsigned uu=0; uu<edgePOrbitPairings.size(); uu++) {
	if (find(vo.begin(),vo.end(),edgePOrbitPairings[uu][0])!=vo.end()) {
	  graph_edges.push_back(edgePOrbitPairings[uu]);
	  maps.push_back(edgePOrbitPairingMaps[uu]);
	}
      }
      sgraph G(graph_edges);      
      vector<vector<int> > stabgens;
      vector<vector<bool> > orientations;
      //      cout << "G=" << G << endl;
      //      cout << "basepoint? " << vo[baseptindex] << endl;
      G.fundamental_group(vo[baseptindex],stabgens,orientations);
      //      cout << "Done computing fund gp " << endl;

      for (unsigned kk=0; kk<stabgens.size(); kk++) {
	vector<int> yo;
	//	cout << "Constructing element of stab" << endl;
	for (unsigned ll=0; ll<stabgens[kk].size(); ll++) {
	  //	  cout << "stabgens elt: " << stabgens[kk][ll] << ",  maps: " << maps[stabgens[kk][ll]] << endl;
	  //	  cout << "orientations: " << orientations[kk][ll] << endl;
	  if (orientations[kk][ll]) {
	    yo = concat(maps[stabgens[kk][ll]],yo);
	  }
	  else {
	    yo = concat(invertword(maps[stabgens[kk][ll]]),yo);
	  }
	}
	if (yo.size()>0) {
	  cout << convertword(yo) << endl;
	}
	else {
	  cout << "Trivial elt!" << endl;
	}
	stab.push_back(yo);
      }
      //      cout << "done" << endl;
    }
    else {
      // there is a single vertex, won't use graph routine for this (just take all pairings as generators)
      for (unsigned uu=0; uu<edgePOrbitPairings.size(); uu++) {
	if (find(vo.begin(),vo.end(),edgePOrbitPairings[uu][0])!=vo.end()) {
	  stab.push_back(edgePOrbitPairingMaps[uu]);
	}
      }
    }

    
    // Perhaps the edge is on the mirror of P^5 (s5 cases)
    if (needpowerofp) {
      cout << "THIS EDGE IS ON THE MIRROR OF P^" << orderofppower << endl;      
      vector<int> we;
      for (unsigned oo=0; oo<orderofppower; oo++) {
	we.push_back(4);
      }
      stab.push_back(we);
    }
    cout << "Done computing stab" << endl;

    vector<gen> sgmatrices;
    cout << "Generators of stab: (there are " << stab.size() << ")" << endl;
    for (unsigned y=0; y<stab.size(); y++) {
      if (stab[y].size()>0) {
	cout << "  " << convertword(stab[y]) << endl;
	gen M = evalWord(stab[y]);
	//	cout << M << endl;
	if (!isScalar(M)) {
	  sgmatrices.push_back(M);
	  gen W1 = M*Xo;
	  gen W2 = M*Xe;
	  vectred(W1);
	  vectred(W2);
	  
	  if (areDep(W1,Xo) && areDep(W2,Xe)) {
	    cout << "Check (edge is fixed pointwise)" << endl;
	  }
	  else {
	    if (areDep(W1,Xe) && areDep(W2,Xo)) {
	      cout << "Check (edge is flipped)" << endl;
	    }
	    else {
	      cerr << "Trouble, edge is not stabilized!" << endl;
	      exit(1);
	    }
	  }
	}
      }
    }

    FiniteGroup L(tag);
    L.addElements(sgmatrices);
    L.expandToGroup();
    
    cout << "Stab has order " << L.elements.size() << endl;
    res = res + onenumber/L.elements.size();
    lcm = _lcm(makesequence(lcm,L.elements.size()),&ct);

  }

  return res;  

}

gen Polytope::computeAngle(int j, gen X) {
  //  cout << "In compute angles, X=" << X << endl;
  gen mi;
  if (isRidgeComplex(j,mi)) {
    gen te = SqNorm(X);
    //    cout << "<X,X> = " << te << endl;
    int sg;
    if (checkSign(te,sg)) {
      if (sg<0) {
	gen Y = BoxProd(X,mi);
	gen Z = BoxProd(X,Y);
	// X,Z spans the complex line orthog to common slice
	//   (beware it's orthogonal but not orthonormal)

	//	cout << "Will compute equations of faces in complex line" << endl;
	Prism P1 = faces[ridges[j].faceindices[0]];
	Prism P2 = faces[ridges[j].faceindices[1]];
	gen a1 = Inn(X,P1.X0)*Inn(P1.X0,X) - Inn(X,P1.X1)*Inn(P1.X1,X);
	gen a2 = Inn(X,P2.X0)*Inn(P2.X0,X) - Inn(X,P2.X1)*Inn(P2.X1,X);
	red(a1);
	red(a2);
       
	if (!operator_equal(a1,zeronumber,&ct) || !operator_equal(a2,zeronumber,&ct)) {
	  cerr << "Trouble, X is not on bisectors?" << endl;
	  exit(1);
	}	
	gen b1 = Inn(Z,P1.X0)*Inn(P1.X0,X) - Inn(Z,P1.X1)*Inn(P1.X1,X);
	gen b2 = Inn(Z,P2.X0)*Inn(P2.X0,X) - Inn(Z,P2.X1)*Inn(P2.X1,X);
	
	gen mu1 = evali(b1);
	gen mu2 = evali(b2);
	gen te = -re(mu1*conj(mu2,&ct)/(abs(mu1,&ct)*abs(mu2,&ct)),&ct);
	return acos(te,&ct)/(2*gen("pi",&ct));
      }
      else {
	cerr << "Can't compute the intersection angle at a point outside complex hyperbolic space" << endl;
	exit(1);
      }
    }
    else {
      cerr << "Trouble checking sign of X" << endl;
      exit(1);
    }
  }
  else {
    cerr << "Angle computation is implemented only for cotranchal intersections!" << endl;
    return zeronumber;
  }
}

gen Polytope::computeAngleNum(int j, gen X) {
  // same as previous method, but X has interval coordinates

  //  cout << "In compute angles, X=" << X << endl;
  gen mi;
  if (isRidgeComplex(j,mi)) {
    gen Hn = matevali(H);

    gen te1 = conj(X,&ct)*Hn;
    gen te2 = conj(vectevali(mi),&ct)*Hn;
    vecteur Yv;
    Yv.push_back(te1[1]*te2[2]-te1[2]*te2[1]);
    Yv.push_back(te1[2]*te2[0]-te1[0]*te2[2]);
    Yv.push_back(te1[0]*te2[1]-te1[1]*te2[0]);
    gen Y = gen(Yv);

    te2 = conj(Y,&ct)*Hn;
    vecteur Zv;
    Zv.push_back(te1[1]*te2[2]-te1[2]*te2[1]);
    Zv.push_back(te1[2]*te2[0]-te1[0]*te2[2]);
    Zv.push_back(te1[0]*te2[1]-te1[1]*te2[0]);
    gen Z = gen(Zv);

    Prism P1 = faces[ridges[j].faceindices[0]];
    Prism P2 = faces[ridges[j].faceindices[1]];

    gen p1x0 = vectevali(P1.X0);
    gen p1x1 = vectevali(P1.X1);
    gen p2x0 = vectevali(P2.X0);
    gen p2x1 = vectevali(P2.X1);

    gen mu1 = (conj(p1x0,&ct)*Hn*Z)*(conj(X,&ct)*Hn*p1x0) - (conj(p1x1,&ct)*Hn*Z)*(conj(X,&ct)*Hn*p1x1);
    gen mu2 = (conj(p2x0,&ct)*Hn*Z)*(conj(X,&ct)*Hn*p2x0) - (conj(p2x1,&ct)*Hn*Z)*(conj(X,&ct)*Hn*p2x1);

    //    cout << "mu1=" << mu1 << endl;
    //    cout << "mu2=" << mu2 << endl;

    gen te = -re(mu1*conj(mu2,&ct)/(abs(mu1,&ct)*abs(mu2,&ct)),&ct);
    //    cout << "te=" << te << endl;

    return acos(te,&ct)/(2*gen("pi",&ct));
  }
}

bool Polytope::isRidgeComplex(int j, gen &mi) {
  ridge r = ridges[j];
  unsigned jface = r.faceindices[0];
  
  bool iscomplex = false;
  bool isbottom = true;
  int uu=0;
  while (isbottom && uu<r.verts.size()) {
    gen X = vertices[r.verts[uu]].coords;
    isbottom = operator_equal(Inn(X,faces[jface].mirror.coords),zeronumber,&ct);
    uu++;
  }
  if (isbottom) {
    //    cout << "This is a complex ridge!" << endl;
    mi = faces[jface].mirror.coords;
    return true;
  }
  else {
    bool istop = true;
    uu=0;
    while (istop && uu<r.verts.size()) {
      gen X = vertices[r.verts[uu]].coords;
      istop = operator_equal(Inn(X,faces[jface].apex.coords),zeronumber,&ct);
      uu++;
    }
    if (istop) {
      //      cout << "This is a complex ridge!" << endl;
      mi = faces[jface].apex.coords;
      return true;
    }
  }
  return false;
}

gen Polytope::eulerdim2() {

  gen res = zeronumber;

  vector<int> ridgeorbitstocheck;
  for (unsigned k=0; k<ridgePOrbits.size(); k++) {
    ridgeorbitstocheck.push_back(k);
  }

  while (ridgeorbitstocheck.size()>0) {
    cout << "Ridge orbits to check: " << convertword(ridgeorbitstocheck) << endl;
    cout << "Studying orbit " << ridgePOrbits[ridgeorbitstocheck[0]] << endl;

    int init_orbit_index = ridgeorbitstocheck[0]; 
    cout << "Computing cycle for ridge on faces " << convertword(ridges[ridgePOrbits[init_orbit_index][0]].faceindices) << endl;

    gen mi;
    bool iscomplex = isRidgeComplex(ridgePOrbits[init_orbit_index][0],mi);

    unsigned jstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[0];
    unsigned kstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[1];

    //    int init_ridge = ridgePOrbits[init_orbit_index][0];
    int current_ridge = ridgePOrbits[init_orbit_index][0];
    int j = jstart;

    bool back = false;

    int ridge_stab;
    bool found_stab = false;
    if (tautag=="s5" && iscomplex) {
      while (!back && !found_stab) {
	gen Y = Ppower*mi;
	found_stab = areDep(mi,Y);
	//	found_stab = isStabilizedByPpower(current_ridge);
	if (found_stab) {
	  cout << "Element in ridge orbit is stabilized by P^5!" << endl;
	  ridge_stab = current_ridge;
	}
	current_ridge = applyPairingToRidge(j,current_ridge);
	// need to change mi!!
	isRidgeComplex(current_ridge,mi);
	bool fd = false;
	int u = -1;
	int v;
	while (!fd && u!=ridgePOrbits.size()-1) {
	  u++;
	  v = -1;
	  while (!fd && v!=ridgePOrbits[u].size()-1) {
	    v++;
	    fd = current_ridge == ridgePOrbits[u][v]; 
	  }
	}
	if (!fd) {
	  cerr << "Trouble finding image ridge in a ridge P-orbit" << endl;
	  exit(1);
	}
	else {
	  if (ridges[current_ridge].faceindices[0]==facePairings[j]) {
	    j = ridges[current_ridge].faceindices[1];
	  }
	  else {
	    if (ridges[current_ridge].faceindices[1]==facePairings[j]) {
	      j = ridges[current_ridge].faceindices[0];
	    }
	    else {
	      cerr << "Trouble finding next pairing" << endl;
	      exit(1);
	    }
	  }
	  back = (u==init_orbit_index);
	}
      }

      if (found_stab) {
	cout << "Found ridge in orbit stabilized by P power!" << endl;
	int jj=-1;
	int kk;
	bool fdd = false;
	while (!fdd && jj!=ridgePOrbits.size()-1) {
	  jj++;
	  kk=-1;
	  while (!fdd && kk!=ridgePOrbits[jj].size()-1) {
	    kk++;;
	    fdd = (ridge_stab==ridgePOrbits[jj][kk]);
	  }
	}
	if (!fdd) {
	  cout << "Trouble locating ridge fixed by P power" << endl;
	  exit(1);
	}
	else {
	  init_orbit_index=jj;
	}
      }
    }    

    gen C;
    vector<int> cw;

    if (found_stab) {
      cout << "Will actually start with orbit of " << ridges[ridgePOrbits[init_orbit_index][0]] << ", because it is fixed py power of P" << endl;
    }

    jstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[0];
    kstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[1];

    if (faces[jstart].invertpairing) {
      cw = invertword(faces[jstart].mirror.word);
    }
    else  {
      cw = faces[jstart].mirror.word;      
    }
    C = faces[jstart].pairing;

    current_ridge = ridgePOrbits[init_orbit_index][0];
    j = jstart;    

    back = false;
    while (!back) {
      current_ridge = applyPairingToRidge(j,current_ridge);
      bool fd = false;
      int u = -1;
      int v;
      while (!fd && u!=ridgePOrbits.size()-1) {
	u++;
	v = -1;
	while (!fd && v!=ridgePOrbits[u].size()-1) {
	  v++;
	  fd = current_ridge == ridgePOrbits[u][v]; 
	}
      }
      if (!fd) {
	cerr << "Trouble finding image in some ridge P-orbit" << endl;
	exit(1);
      }
      else {
	//	cout << "u=" << u << endl;
	ridgeorbitstocheck.erase(remove(ridgeorbitstocheck.begin(),ridgeorbitstocheck.end(),u),ridgeorbitstocheck.end());
	if (ridges[current_ridge].faceindices[0]==facePairings[j]) {
	  j = ridges[current_ridge].faceindices[1];
	}
	else {
	  if (ridges[current_ridge].faceindices[1]==facePairings[j]) {
	    j = ridges[current_ridge].faceindices[0];
	  }
	  else {
	    cerr << "Trouble finding next pairing" << endl;
	    exit(1);
	  }
	}
	back = (u==init_orbit_index);
	//	cout << "back? " << back << endl;	
	//	cout << "j=" << j << endl;	
	if (!back) {
	  if (faces[j].invertpairing) {
	    cw = concat(invertword(faces[j].mirror.word),cw);
	  }
	  else  {
	    cw = concat(faces[j].mirror.word,cw);      
	  }
	  C = faces[j].pairing*C;
	  matred(C);
	}
	else {
	  vector<int> power_of_p;
	  for (int k=0; k<v; k++) {
	    power_of_p.push_back(-4);
	    C = Pi*C;
	    matred(C);
	  }
	  cw = concat(power_of_p,cw);	  
	}
      }
    }
    if (!back) {
      cerr << "Trouble with some ridge cycle" << endl;
      exit(1);
    }
    else {
      // just add 1/order(C)? (this is OK if ridge is not stabilized by power of P!
      if (found_stab) {
	vector<gen> sgmatrices;
	sgmatrices.push_back(C);
	cout << "Elements in stab: " << endl;
	cout << convertword(cw) << endl;
	cout << "P^" << orderofppower << endl;
	sgmatrices.push_back(Ppower);
	
	// check??
	isRidgeComplex(ridgePOrbits[init_orbit_index][0],mi);
	for (unsigned yy=0; yy<sgmatrices.size(); yy++) {
	  gen Y = sgmatrices[yy]*mi;
	  vectred(Y);
	  if (!areDep(mi,Y)) {
	    cout << "PROBLEM, MATRIX IN SGMATRICES DOES NOT PRESERVE RIDGE!" << endl;
	  }
	}					     

	FiniteGroup L(tag);
	L.addElements(sgmatrices);
	L.expandToGroup();
      
	cout << "Stab has order " << L.elements.size() << endl;
	res = res + onenumber/L.elements.size();
	lcm = _lcm(makesequence(lcm,L.elements.size()),&ct);
      }
      else {
	int oc = order(C);
	if (oc<0) {
	  cerr << "Ridge with large order of stab (>1000), I will consider it as infinite, and STOP" << endl;
	  exit(1);
	}
	else {
	  cout << "Found a ridge with stabilizer of order " << oc << endl;
	  res = res + onenumber/oc;
	  lcm = _lcm(makesequence(lcm,oc),&ct);
	}
      }
    }    
  }
  return res;
}

gen Polytope::eulerdim3() {
  //  cout << "Computing 3d contrib" << endl;
  if (tautag=="s5") {
    gen res = zeronumber;
    for (unsigned k=0; k<facePOrbits.size(); k++) {
      gen te1 = Inn(faces[facePOrbits[k][0]].apex.coords,extramirror);
      if (operator_equal(te1,zeronumber,&ct)) {
	gen te2 = Inn(faces[facePOrbits[k][0]].apexProjection,extramirror);
	if (operator_equal(te2,zeronumber,&ct)) {
	  //	  cout << "Adding " << orderofppower/orderofp << endl;
	  int te = orderofp/orderofppower;
	  res = res+onenumber/te;
	  lcm = _lcm(makesequence(lcm,te),&ct);
	}
	else {
	  //	  cout << "Adding 1" << endl;
	  res = res + onenumber;
	}
      }
      else {
	//	cout << "Adding 1" << endl;
	res = res + onenumber;
      }
    }
    return res/twonumber;
  }
  else {
    return(facePOrbits.size()/2);
  }
}

void Polytope::computeEulerCharacteristic() {

  // VERTEX ORBITS
  //  gen ed0 = zeronumber;

  lcm = orderofp*onenumber; // will keep track of lcm of orders of stabilizers

  gen ed0 = eulerdim0();

  // EDGE ORBITS
  gen ed1 = eulerdim1();

  // RIDGE ORBITS
  gen ed2 = eulerdim2();

  // FACE ORBITS
  gen ed3 = eulerdim3();
  // Can there ever be faces that are paired with something in the same P-orbit?
  //   (if so this may not be the right formula)
  
  // 4-dim contribution is just (order of P)^-1:
  gen ed4 = onenumber/orderofp;

  cout << "Euler 0-dim contribution: " << ed0 << endl;
  cout << "Euler 1-dim contribution: " << ed1 << endl;
  cout << "Euler 2-dim contribution: " << ed2 << endl;
  cout << "Euler 3-dim contribution: " << ed3 << endl;
  cout << "Euler 4-dim contribution: " << ed4 << endl;

  cout << endl;
  euler_char = ed0-ed1+ed2-ed3+ed4;
  cout << "Euler characteristic: " << euler_char << endl;
  cout << endl;

}

void Polytope::checkPairings() {
  for (unsigned k=0; k<faces.size(); k++) {
    
    cout << "#" << k << ": ";
    // << faces[k] << endl;
    //	useRelations(faces[k]);
    cout << faces[k] << endl;
    //	cout << "" << endl;
    
    Prism B = applyReflection(faces[k]);
    //	cout << "  maps to " << B << endl;
    bool found = false;
    unsigned l=0;
    while (!found && l<faces.size()) {
      found = comparePrisms(B,faces[l]);
      l++;
    }
    if (found) {
      cout << "  paired with " << faces[l-1] << endl;
      faces[k].pairing = reflMat(faces[k].mirror.coords,mult);
      faces[k].invertpairing = false;
      facePairings.push_back(l-1);
    }
    else {
      Prism Bi = applyInverseReflection(faces[k]);
      found = false;
      l=0;
      while (!found && l<faces.size()) {
	found = comparePrisms(Bi,faces[l]);
	l++;
      }
      if (found) {
	cout << "  [inverse] paired with " << faces[l-1] << endl;
	faces[k].invertpairing = true;
	faces[k].pairing = reflMat(faces[k].mirror.coords,multconj);
	facePairings.push_back(l-1);
      }
      else {
	cerr << "Problem with pairings!" << endl;
	exit(1);
      }
    }
  }
  cout << "List of pairings: " << endl;
  for (unsigned k=0; k<faces.size() ; k++) {
    cout << "Face #" << k << " --> " << facePairings[k] << endl;
  }
}

void Polytope::computeAllVertices() {
  vector<WVector> rawverts;
  for (unsigned k=0; k<faces.size(); k++) {
    unsigned nb = faces[k].bottomface.size();
    for (unsigned l=0; l<nb; l++) {
      if (!isInList(faces[k].bottomface[l],rawverts)) {
	rawverts.push_back(faces[k].bottomface[l]);
      }
    }
    nb = faces[k].topface.size();
    for (unsigned l=0; l<nb; l++) {
      if (!isInList(faces[k].topface[l],rawverts)) {
	rawverts.push_back(faces[k].topface[l]);
      }
    }
    nb = faces[k].midvertices.size();
    for (unsigned l=0; l<nb; l++) {
      if (!isInList(faces[k].midvertices[l],rawverts)) {
	rawverts.push_back(faces[k].midvertices[l]);
      }
    }
  }
  cout << "Collected " << rawverts.size() << " vertices..." << endl;
  
  while (rawverts.size()>0) {
    WVector v0 = rawverts[0];
    WVector vk = rawverts[0];
    vector<int> orb;
    orb.push_back(vertices.size());
    vertices.push_back(v0);
    vertReps.push_back(v0);
    rawverts.erase(rawverts.begin());
    bool back = false;
    while (!back) {
      vk.coords = P*vk.coords;
      vk.word = conjugateword(word_p1,shiftIndicesUp(vk.word));
      vectred(vk.coords);
      back = testSame(vk,v0);
      if (!back) {
	orb.push_back(vertices.size());
	vertices.push_back(vk);
	int ind = findInList(vk,rawverts);
	if (ind>-1) {
	  rawverts.erase(rawverts.begin()+ind);
	}
	else {
	  cerr << "Trouble computing orbit!" << endl;
	  exit(1);
	}
      }
      else {
	vertexPOrbits.push_back(orb);
      }
    }
  }
  cout << "The vertices come in " << vertReps.size() << " P-orbits." << endl;
  
  for (unsigned k=0; k<vertexPOrbits.size(); k++) {
    cout << "Orbit #" << k << ": " << vertexPOrbits[k] << endl;
    int no = vertexPOrbits[k].size();
    int k0 = vertexPOrbits[k][0];
    vector<int> fi = findFaces(vertices[k0]);
    cout << "Vert #" << k0 << " is on " << convertword(fi) << endl;
    if (fi[0]==-1) {
      fi.erase(fi.begin());
      for (unsigned m=0; m<no; m++) {
	vertices[k0+m].faceindices.push_back(-1);
      }
    }
    for (unsigned l=0; l<fi.size(); l++) {
      unsigned fil = fi[l];
      // identify it in some POrbit of faces!
      bool fd = false;
      int u=-1;
      int v;
      while (!fd && u!=facePOrbits.size()-1) {
	u++;
	v=-1;
	while (!fd && v!=facePOrbits[u].size()-1) {
	  v++;
	  fd = (fil==facePOrbits[u][v]);
	}
      }
      if (!fd) {
	cerr << "Problem, some face was not found in P-orbits of faces!" << endl;
	exit(1);
      }
      else {
	//	    cout << "Found face #" << fil << " in orbit #" << u << ", in position #" << v << endl;
	for (unsigned m=0; m<no; m++) {
	  //	      cout << "Setting faceindices for vertex #" << vertexPOrbits[k][m] << endl;
	  vertices[vertexPOrbits[k][m]].faceindices.push_back(facePOrbits[u][(v+m)%facePOrbits[u].size()]);
	}
      }
    }
    /*	for (unsigned m=0; m<no; m++) {
	cout << "v#" << vertexPOrbits[k][m] << " is on " << convertword(vertices[vertexPOrbits[k][m]].faceindices) << endl;
	}*/
    
    vector<WVector> mi = findMirrors(vertices[k0]);
    cout << "  and on following mirrors:  " << endl;
    for (unsigned u = 0; u<mi.size(); u++) {
      cout << mi[u].word << endl;
    }
    cout << "[missing program]  Need to find mirrors for P-images of vertices as well..." << endl;

  }

  for (unsigned k=0; k<vertReps.size(); k++) {
    // this is misnamed, for now it only finds pairings of vertices
    computeStabilizer(k);
  }

  for (unsigned k=0; k<vertices.size(); k++) {
    cout << "Vertex #" << k << " is on " << convertword(vertices[k].faceindices) << endl;
  }
}

void Polytope::computeAllEdges() {
  cout << "Will now construct edges..." << endl;
  for (unsigned kk=0; kk<facePOrbits.size(); kk++) {
    unsigned k = facePOrbits[kk][0];
    cout << "Face #" << k << endl;
    //	cout << "Top: ";
    vector<int> tf;
    for (unsigned l=0; l<faces[k].topface.size(); l++) {
      int te = findInList(faces[k].topface[l],vertices);
      if (te<0) {
	cerr << "Problem identifying face vertex" << endl;
	exit(1);
      }
      else {
	tf.push_back(te);
	//	    cout << "v#" << te << "->";
	faces[k].topface[l].reftovlist = te;
      }
    }
    if (tf.size()>1) {
      for (unsigned l=0; l<tf.size()-1; l++) {
	edge e;
	e.origin = tf[l];
	e.end = tf[l+1];	    
	addEdge(e);
      }
      edge e;
      e.origin = tf[tf.size()-1];
      e.end = tf[0];
      addEdge(e);
    }
    
    //	cout << endl << "Mid: ";
    vector<int> mf;
    for (unsigned l=0; l<faces[k].midvertices.size(); l++) {
      int te = findInList(faces[k].midvertices[l],vertices);
      if (te<0) {
	cerr << "Problem identifying face vertex" << endl;
	exit(1);
      }
      else {
	//	    cout << "v#" << te << "  ";
	mf.push_back(te);
	faces[k].midvertices[l].reftovlist = te;
      }
    }
    
    //	cout << endl << "Bottom: ";
    vector<int> bf;
    for (unsigned l=0; l<faces[k].bottomface.size(); l++) {
      int te = findInList(faces[k].bottomface[l],vertices);
      if (te<0) {
	cerr << "Problem identifying face vertex" << endl;
	exit(1);
      }
      else {
	bf.push_back(te);
	//	    cout << "v#" << te << "->";
	faces[k].bottomface[l].reftovlist = te;
      }
    }
    
    for (unsigned l=0; l<bf.size()-1; l++) {
      edge e;
      e.origin = bf[l];
      e.end = bf[l+1];
      addEdge(e);
    }
    edge eb;
    eb.origin = bf[bf.size()-1];
    eb.end = bf[0];
    addEdge(eb);
    
    if (tf.size()>1) {
      for (unsigned l=0; l<bf.size(); l++) {
	if (find(faces[k].midvertindices.begin(),faces[k].midvertindices.end(),l)==faces[k].midvertindices.end()) {
	  edge e;
	  e.origin = bf[l];
	  e.end = tf[l];
	  addEdge(e);
	}
      }
    }
    else {
      for (unsigned l=0; l<bf.size(); l++) {
	if (find(faces[k].midvertindices.begin(),faces[k].midvertindices.end(),l)==faces[k].midvertindices.end()) {
	  edge e;
	  e.origin = bf[l];
	  e.end = tf[0];
	  addEdge(e);
	}
      }
    }
    
    for (unsigned l=0; l<mf.size(); l++) {
      if (tf.size()>1) {
	edge e1, e2;
	e1.origin = bf[faces[k].midvertindices[l]];
	e1.end = mf[l];
	addEdge(e1);
	e2.origin = mf[l];
	e2.end = tf[faces[k].midvertindices[l]];
	addEdge(e2);
      }
      else {
	edge e1, e2;
	e1.origin = bf[faces[k].midvertindices[l]];
	e1.end = mf[l];
	addEdge(e1);
	e2.origin = mf[l];
	e2.end = tf[0];
	addEdge(e2);
      }
    }
  }
  cout << "Created " << edges.size() << " edges..." << endl;
}

void Polytope::computeAllRidges() {
  cout << "Will now construct ridges..." << endl;
  for (unsigned kk=0; kk<facePOrbits.size(); kk++) {
    
    unsigned k = facePOrbits[kk][0];
    cout << "Face #" << k << endl;
    unsigned ns = faces[k].sides.size();
    //	cout << "mid vert indices: " << faces[k].midvertindices << endl;
    for (unsigned l=0; l<ns; l++) {
      ridge r;
      if (faces[k].topface.size()==1) {
	r.verts.push_back(faces[k].topface[0].reftovlist);
      }
      else {
	r.verts.push_back(faces[k].topface[(l+1)%ns].reftovlist);
	r.verts.push_back(faces[k].topface[l].reftovlist);
      }
      bool fd = false;
      int u1=0; 
      while (!fd && u1<faces[k].midvertindices.size()){
	fd = l==faces[k].midvertindices[u1];
	u1++;
      }
      if (fd){
	u1--;
	//	    cout << " mid vertex " << faces[k].midvertices[u1].reftovlist << endl;
	r.verts.push_back(faces[k].midvertices[u1].reftovlist);
      }	  
      r.verts.push_back(faces[k].bottomface[l].reftovlist);
      r.verts.push_back(faces[k].bottomface[(l+1)%ns].reftovlist);
      fd = false;
      int u2=0; 
      while (!fd && u2<faces[k].midvertindices.size()){
	fd = (l+1)%ns==faces[k].midvertindices[u2];
	u2++;
      }
      if (fd){
	u2--;
	//	    cout << " mid vertex " << faces[k].midvertices[u2].reftovlist << endl;
	r.verts.push_back(faces[k].midvertices[u2].reftovlist);
      }
      addRidge(r);
    }
    
    ridge r;
    for (unsigned l=0; l<ns; l++) {
      r.verts.push_back(faces[k].bottomface[l].reftovlist);
    }
    /*    for (unsigned l=ns-1; l>=0; l--) {
      r.verts.push_back(faces[k].bottomface[l].reftovlist);
      }*/
    addRidge(r);
    
    if (faces[k].topface.size()>1) {
      ridge r;
      /*  for (unsigned l=ns-1; l>=0; l--) {
	r.verts.push_back(faces[k].topface[l].reftovlist);
	}*/
      for (unsigned l=0; l<ns; l++) {
	r.verts.push_back(faces[k].topface[l].reftovlist);
      }
      addRidge(r);
    }	
    
  }
  cout << "Created " << ridges.size() << " ridges..." << endl;
}

bool Polytope::isSame(edge e1, edge e2) {
  return (e1.origin==e2.origin && e1.end == e2.end); 
}

bool Polytope::isFlipped(edge e1, edge e2) {
  return (e1.origin==e2.end && e1.end == e2.origin); 
}

bool Polytope::isEdgeNew(edge e) {
  bool res = true;
  unsigned l=0;
  while (res && l<edges.size()){
    res = !isSame(e,edges[l]) && !isFlipped(e,edges[l]);
    l++;
  }
  return res;
}

int Polytope::applyPtoVertex(int k) {
  bool found = false;
  int l=-1;
  int m;
  while (!found && l!=vertexPOrbits.size()-1) {
    l++;
    m=-1;
    while (!found && m!=vertexPOrbits[l].size()-1) {
      m++;
      found = vertexPOrbits[l][m]==k; 
    }
  }
  if (!found) {
    cerr << "Not supposed to happen (trouble locating vertex index in P orbits)" << endl;
    exit(1);
  }
  else {
    return vertexPOrbits[l][(m+1)%vertexPOrbits[l].size()];
  }
}

vector<int> Polytope::getPOrbitOfVertex(int k) {
  bool found = false;
  int l=-1;
  int m;
  //  cout << "Trying to locate v#" << k << endl;
  while (!found && l!=vertexPOrbits.size()-1) {
    l++;
    //    cout << "Searching in following orbit: " << vertexPOrbits[l] << endl;
    m=-1;
    while (!found && m!=vertexPOrbits[l].size()-1) {
      m++;
      found = vertexPOrbits[l][m]==k; 
    }
  }
  if (!found) {
    cerr << "Not supposed to happen (trouble locating vertex index in P orbits)" << endl;
    exit(1);
  }
  else {
    vector<int> res;
    int nmax = vertexPOrbits[l].size();
    for (unsigned n=0; n<nmax; n++) {
      res.push_back(vertexPOrbits[l][(m+n)%nmax]);
    }
    return res;
  }
}

int Polytope::findNextInFacePOrbit(int k) {
  bool found = false;
  int l=-1;
  int m;
  while (!found && l!=facePOrbits.size()-1) {
    l++;
    m=-1;
    while (!found && m!=facePOrbits[l].size()-1) {
      m++;
      found = facePOrbits[l][m]==k; 
    }
  }
  if (!found) {
    cerr << "Not supposed to happen (trouble locating face index in P orbits)" << endl;
    exit(1);
  }
  else {
    return facePOrbits[l][(m+1)%facePOrbits[l].size()];
  }  
}

vector<int> Polytope::getPOrbitOfFace(int k) {
  bool found = false;
  int l=-1;
  int m;
  while (!found && l!=facePOrbits.size()-1) {
    l++;
    m=-1;
    while (!found && m!=facePOrbits[l].size()-1) {
      m++;
      found = facePOrbits[l][m]==k; 
    }
  }
  if (!found) {
    cerr << "Not supposed to happen (trouble locating face index in P orbits)" << endl;
    exit(1);
  }
  else {
    vector<int> res;
    int nmax = facePOrbits[l].size();
    for (unsigned n=0; n<nmax; n++) {
      res.push_back(facePOrbits[l][(m+n)%nmax]);
    }
    return res;
  }
}

void Polytope::addRidge(ridge e) {

  vector<int> orbitinds;
  if (isRidgeNew(e)) {

    unsigned nv = e.verts.size();

    vector<vector<int> > li;
    for (unsigned u=0; u<nv; u++) {
      cout << "vertex #" << e.verts[u] << ": " << vertices[e.verts[u]].faceindices << endl;
      li.push_back(vertices[e.verts[u]].faceindices);
    }
    vector<int> inds = intersect(li);
    if (inds[0]==-1) {
      inds.erase(inds.begin());
    }
    if (inds.size()!=2) {
      cerr << "Problem with ridge, it is on " << inds.size() << " faces, namely " << convertword(inds) << endl;
      exit(1);
    }
    else {
      for (unsigned u=0; u<2; u++) {
	e.faceindices.push_back(inds[u]);
      }
    }

    cout << "Adding ridge w/ vertices " << convertword(e.verts) << "     (which is on faces " << convertword(e.faceindices) << ")" << endl;
    orbitinds.push_back(ridges.size());
    ridges.push_back(e);

    vector<vector<int> > yo;
    unsigned maxorb = 0;
    unsigned maxorbind = 0;
    for (unsigned k=0; k<nv; k++) {
      vector<int> yok = getPOrbitOfVertex(e.verts[k]);
      if (yok.size()>maxorb) {
	maxorb = yok.size();
	maxorbind = k;
      }
      yo.push_back(yok);
    }

    bool test = true;
    for (unsigned l=1; l<yo[maxorbind].size() && test; l++) {
      ridge r;
      for (unsigned m=0; m<nv; m++) {
	r.verts.push_back(yo[m][l%yo[m].size()]);
      }
      for (unsigned u=0; u<2; u++) {
	r.faceindices.push_back(findNextInFacePOrbit(ridges[ridges.size()-1].faceindices[u]));
      }
      test = isRidgeNew(r);
      if (test) {
	cout << "Adding ridge w/ vertices " << convertword(r.verts) << "     (which is on faces " << convertword(r.faceindices) << ")" << endl;
	//	cout << "Adding ridge w/ vertices " << convertword(r.verts) << endl;
	orbitinds.push_back(ridges.size());
	ridges.push_back(r);
      }
    }
    ridgePOrbits.push_back(orbitinds);
  }
}

void Polytope::addEdge(edge e) {

  vector<int> orbitinds;
  int j = e.origin;
  int k= e.end;
  vector<int> commoninds = intersect(vertices[j].faceindices,vertices[k].faceindices);
  if (commoninds[0]==-1) {
    // this may not be appropriate, if we want to have faces entirely contained in the boundary!
    commoninds.erase(commoninds.begin());
  }
  for (unsigned m=0; m<commoninds.size(); m++) {
    if (isVertexSafe(vertices[j],faces[commoninds[m]]) && isVertexSafe(vertices[k],faces[commoninds[m]])){
      e.faceindices.push_back(commoninds[m]);
    } 
  }
  if (isEdgeNew(e)) {
    cout << "Adding edge... v#" << e.origin << "," << e.end << "     (which is on faces " << convertword(e.faceindices) << ")" << endl;
    orbitinds.push_back(edges.size());
    edges.push_back(e);
    //  }// THIS IS VERY STRANGE, SHOULD DO THE REST ONLY WHEN NEW!?!

    vector<int> jo = getPOrbitOfVertex(j);
    vector<int> ko = getPOrbitOfVertex(k);
    //  cout << "Orbit of v#" << j << ": " << jo << endl;
    //  cout << "Orbit of v#" << k << ": " << ko << endl;
    
    bool test = true;
    if (jo.size()<ko.size()) {
      for (unsigned l=1; l<ko.size() && test; l++) {
	edge e;
	e.origin = jo[l%jo.size()];
	e.end = ko[l];
	test = isEdgeNew(e);
	if (test) {
	  unsigned z = edges.size()-1;
	  for (unsigned m=0; m<edges[z].faceindices.size(); m++) {
	    int g = edges[z].faceindices[m];
	    if (g>-1)  {
	      e.faceindices.push_back(findNextInFacePOrbit(g));
	    }
	  }
	  cout << "Adding edge... v#" << e.origin << "," << e.end << "     (which is on faces " << convertword(e.faceindices) << ")" << endl;
	  orbitinds.push_back(edges.size());
	  edges.push_back(e);
	}
      }    
    }
    else {
      if (jo.size()>ko.size()) {
	for (unsigned l=1; l<jo.size() && test; l++) {
	  edge e;
	  e.origin = jo[l];
	  e.end = ko[l%ko.size()];
	  test = isEdgeNew(e);
	  if (test) {
	    unsigned z = edges.size()-1;
	    for (unsigned m=0; m<edges[z].faceindices.size(); m++) {
	      int g = edges[z].faceindices[m];
	      if (g>-1)  {
		e.faceindices.push_back(findNextInFacePOrbit(g));
	      }
	    }
	    cout << "Adding edge... v#" << e.origin << "," << e.end << "     (which is on faces " << convertword(e.faceindices) << ")" << endl;
	    orbitinds.push_back(edges.size());
	    edges.push_back(e);
	  }
	}    
      }
      else {
	for (unsigned l=1; l<jo.size() && test; l++) {
	  edge e;
	  e.origin = jo[l];
	  e.end = ko[l];
	  test = isEdgeNew(e);
	  if (test) {
	    unsigned z = edges.size()-1;
	    for (unsigned m=0; m<edges[z].faceindices.size(); m++) {
	      int g = edges[z].faceindices[m];
	      if (g>-1)  {
		e.faceindices.push_back(findNextInFacePOrbit(g));
	      }
	    }
	    cout << "Adding edge... v#" << e.origin << "," << e.end << "     (which is on faces " << convertword(e.faceindices) << ")" << endl;
	    orbitinds.push_back(edges.size());
	    edges.push_back(e);
	  }
	}
      }
    }
    edgePOrbits.push_back(orbitinds);
  }
  
}

bool Polytope::isSame(ridge e1, ridge e2) {
  if (e1.verts.size()!=e2.verts.size()) {
    return false;
  }
  else {
    bool res = true;
    unsigned l=0;
    while (l<e1.verts.size() && res)  {
      res = e1.verts[l]==e2.verts[l];
      l++;
    }
    return res;
  }
}

bool Polytope::isRotated(ridge e1, ridge e2) {
  int nv = e1.verts.size();
  if (nv!=e2.verts.size()) {
    return false;
  }
  else {
    int k1 = e1.verts[0];
    unsigned l=0;
    bool fd = false;
    while (!fd && l<nv) {
      fd = k1==e2.verts[l];
      l++;
    }
    if (!fd) {
      return false;
    }
    else {
      int k2 = l-1;
      bool res = true;
      unsigned m=1;
      while (res && m<nv) {
	res = e1.verts[m]==e2.verts[(k2+m)%nv];
	m++;
      }
      return res;
    }
  }  
}

bool Polytope::isRidgeNew(ridge r) {
  bool res = true;
  unsigned l=0;
  while (res && l<ridges.size()){
    res = !isRotated(r,ridges[l]);
    l++;
  }
  return res;
}

bool Polytope::testSame(WVector V, WVector W){
  return areDep(V.coords,W.coords);
}

bool Polytope::isInList(WVector V, vector<WVector> li){
  unsigned n = li.size();
  bool res = false;
  unsigned k = 0;
  while (!res && k<n) {
    res = testSame(V,li[k]);
    k++;
  }
  return res;
}

int Polytope::findInList(gen V, vector<WVector> li){
  vector<int> w;
  WVector W = WVector(V,w);
  return findInList(W,li);
}

int Polytope::findInList(WVector V, vector<WVector> li){
  unsigned n = li.size();
  bool res = false;
  int k = 0;
  while (!res && k<n) {
    res = testSame(V,li[k]);
    k++;
  }
  if (res) {
    return (k-1);
  }
  else {
    return -1;
  }
}

bool Polytope::computeVertices(Prism &A) {

  gen te = SqNorm(A.apex.coords);
  int res;
  if (checkSign(te,res)) {
    cout << "Sign of sq nm of apex: " << res << endl;

    unsigned k1 = 0;
    unsigned l1 = A.sides[0].word.size();
    for (unsigned k=1; k<A.sides.size(); k++) {
      unsigned l = A.sides[k].word.size();
      if (l<l1) {
	l1 = l;
	k1 = k;
      }
    }
    unsigned k2;
    if (k1==0) {
      if (A.sides[A.sides.size()-1].word.size()<A.sides[1].word.size()) {
	k2 = A.sides.size()-1;
      }
      else {
	k2 = 1;
      }
    }
    else {
      k2 = (k1+1)%A.sides.size();
    }
    
    gen M1 = reflMat(A.sides[k1].coords,mult)*reflMat(A.sides[k2].coords,mult);
    matred(M1);
    int om1 = order(M1);
    cout << "Product of order " << om1 << endl;
    int bo = braidOrder(reflMat(A.sides[k1].coords,mult),reflMat(A.sides[k2].coords,mult));
    cout << "Braiding order " << bo << endl;

    vector<int> w0 = concat(A.sides[k1].word,A.sides[k2].word);
    vector<int> w = concat(A.sides[k1].word,A.sides[k2].word);

    bool isrefl = false;
    unsigned k=1;
    gen M = M1;
    while (k<om1 && !isrefl) {
      gen dt = discrAdjusted(M);
      isrefl = operator_equal(dt,zeronumber,&ct);
      if (isrefl) {
	int sg;
	if (!checkSign(SqNorm(A.apex.coords),sg)) {
	  cerr << "Problem checking sign" << endl;
	  exit(1);
	}
	else {
	  if (sg>0) {
	    gen X1 = BoxProd(A.cspine,A.apex.coords);
	    gen X2 = M*X1;
	    vectred(X2);
	    isrefl = areDep(X1,X2);
	  }
	  else {
	    if (sg<0) {
	      gen X1 = A.apex.coords;
	      gen X2 = M*X1;
	      vectred(X2);
	      isrefl = areDep(X1,X2);
	    }
	    else {
	      // don't know what to do...
	    }
	  }
	}
      }
      if (isrefl) {
	cout << "Found reflection, power " << k << endl;
      }
      else {
	M = M*M1;
	w = concat(w,w0);
	matred(M);
      }
      k++;
    }
    if (!isrefl) {
      cerr << "Did not find any reflection power..." << endl;
      //      exit(1);
    }

    A.apex.word = w;
    cout << "w apex=" << convertword(w) << endl;

    if (res<1) {
      cout << "Single vertex at top" << endl;
      A.topface.push_back(A.apex);
    }
    else {
      cout << "Top face is not a point.." << endl;

      for (unsigned k=0; k<A.sides.size(); k++) {
	gen yo = BoxProd(A.apex.coords,A.sides[k].coords);
	WVector Vk( yo, concat(A.apex.word,A.sides[k].word) );
	A.topface.push_back(Vk);
        cout << "Vertex for " << convertword(A.sides[k].word) << endl;
	if (checkSign(SqNorm(Vk.coords),res)){
	  if (res>0) {
	    cerr << "This should not happen!" << endl;
	    exit(1);
	  }
	}
	else {
	  cerr << "Problem checking sign!" << endl;
	  exit(1);	  
	}	
      }
    }
  }
  else {
    cout << "Problem checking sign... " << evali(te) << endl;
  }

  cout << "Constructing bottom face..." << endl;
  for (unsigned k=0; k<A.sides.size(); k++) {
    gen yo = BoxProd(A.mirror.coords,A.sides[k].coords);
    if (checkSign(SqNorm(yo),res)){
      if (res>0) {
	cout << "#" << k << "(" << convertword(A.sides[k].word) << "): bottom and mid vertex." << endl;

	gen M1 = reflMat(A.mirror.coords,mult)*reflMat(A.sides[k].coords,mult);
	matred(M1);
	int om1 = order(M1);
	cout << "Product of order " << om1 << endl;
	int bo = braidOrder(reflMat(A.mirror.coords,mult),reflMat(A.sides[k].coords,mult));
	cout << "Braiding order " << bo << endl;

	vector<int> w0 = concat(A.mirror.word,A.sides[k].word);
	vector<int> wk = concat(A.mirror.word,A.sides[k].word);

	bool isrefl = false;
	unsigned kk=1;
	gen M = M1;
	while (kk<om1 && !isrefl) {
	  gen dt = discrAdjusted(M);
	  isrefl = operator_equal(dt,zeronumber,&ct);

	  if (isrefl) {
	    gen X1 = BoxProd(A.sides[k].coords,yo);
	    gen X2 = M*X1;
	    vectred(X2);
	    isrefl = areDep(X1,X2);
	  }
	  
	  // also want to check if vertex is fixed.
	  if (isrefl) {
	    cout << "Found reflection, power " << kk << endl;
	  }
	  M = M*M1;
	  wk = concat(wk,w0);
	  matred(M);
	  kk++;
	}
	if (!isrefl) {
	  cout << "Did not find any reflection power..." << endl;
	  cerr << "Did not find any reflection power..." << endl;
	  //	  exit(1);
	}

	WVector nm(yo,wk);

	gen yok1 = BoxProd(A.mirror.coords,yo);
	gen yok2 = BoxProd(A.sides[k].coords,yo);
	WVector Wk1(yok1,concat(A.mirror.word,wk));
	WVector Wk2(yok2,concat(A.sides[k].word,wk));
	if (checkSign(SqNorm(Wk1.coords),res)){
	  if (res>-1) {
	    cerr << "This should not happen!" << endl;
	    exit(1);
	  }
	}
	else {
	  cerr << "Problem checking sign!" << endl;
	  exit(1);	  
	}	
	if (checkSign(SqNorm(Wk2.coords),res)){
	  if (res>-1) {
	    cerr << "This should not happen!" << endl;
	    exit(1);
	  }
	}
	else {
	  cerr << "Problem checking sign!" << endl;
	  exit(1);	  
	}	
	A.bottomface.push_back(Wk1);
	A.midvertices.push_back(Wk2);
	A.midvertindices.push_back(k);
      }
      else {
	cout << "#" << k << "(" << convertword(A.sides[k].word) << "): Single bottom vertex." << endl;
	WVector Vk( yo, concat(A.mirror.word,A.sides[k].word) );
	A.bottomface.push_back(Vk);
      }
    }
    else {
      cerr << "Problem checking sign!" << endl;
      exit(1);	  
    }	
  }
  

}

bool Polytope::isOnMirror(gen X, Prism A) {
  gen te = Inn(X,A.mirror.coords);
  return operator_equal(te,zeronumber,&ct);  
}

bool Polytope::isOnTopFace(gen X, Prism A) {
  gen te = Inn(X,A.apex.coords);
  return operator_equal(te,zeronumber,&ct);  
}

bool Polytope::isOnPrism(gen X, Prism A){
  gen te = Inn(X,A.apex.coords)*Inn(A.apex.coords,A.apexProjection)*Inn(A.apexProjection,X);
  red(te);
  return isReal(te);
}

gen Polytope::reflectAcrossSpine(gen X, Prism A){
  vecteur co;
  co.push_back(Inn(A.apex.coords,A.apex.coords));
  co.push_back(Inn(A.apex.coords,A.apexProjection));
  co.push_back(Inn(A.apexProjection,A.apex.coords));
  co.push_back(Inn(A.apexProjection,A.apexProjection));
  gen M = createMatrix(2,2,co);

  vecteur cov;
  cov.push_back(Inn(X,A.apex.coords));
  cov.push_back(Inn(X,A.apexProjection));
  gen b = gen(cov);

  gen de = det(M);
  red(de);
  if (operator_equal(de,zeronumber,&ct)) {
    cerr << "Problem reflecting across spine" << endl;
    exit(1);
  }
  else {
    gen dei = inverse(de);
    vecteur coeffs;
    coeffs.push_back(dei*(M[1][1]*b[0]-M[0][1]*b[1]));
    coeffs.push_back(dei*(-M[1][0]*b[0]+M[0][0]*b[1]));
    gen Y = myconj(coeffs[0])*A.apex.coords + myconj(coeffs[1])*A.apexProjection;
    vectred(Y);
    return Y;
  }
}

Prism Polytope::createPimage(Prism A) {
  Prism A1;
 
  gen m = P*A.mirror.coords;
  vectred(m);
  vector<int> w1 = conjugateword(word_p1,shiftIndicesUp(A.mirror.word));
  useRelations(w1);
  A1.mirror = WVector(m,w1);

  A1.apex = WVector(P*A.apex.coords, conjugateword(word_p1,shiftIndicesUp(A.apex.word)));
  vectred(A1.apex.coords);
  useRelations(A1.apex.word);

  A1.apexProjection = P*A.apexProjection;
  vectred(A1.apexProjection);

  A1.cspine = P*A.cspine;
  vectred(A1.cspine);
  
  //  A1.X0 = P*A.X0;
  //  A1.X1 = P*A.X1;
  //  vectred(A1.X0);
  //  vectred(A1.X1);

  for (unsigned k=0; k<A.sides.size(); k++) {
      vector<int> wk = conjugateword(word_p1,shiftIndicesUp(A.sides[k].word));
      //  useRelations(wk);
      gen sk = P*A.sides[k].coords;
      vectred(sk);
      A1.sides.push_back(WVector(sk,wk));
  }

  for (unsigned k=0; k<A.bottomface.size(); k++) {
      vector<int> wk = conjugateword(word_p1,shiftIndicesUp(A.bottomface[k].word));
      //   useRelations(wk);
      gen sk = P*A.bottomface[k].coords;
      vectred(sk);
      A1.bottomface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.topface.size(); k++) {
      vector<int> wk = conjugateword(word_p1,shiftIndicesUp(A.topface[k].word));
      //    useRelations(wk);
      gen sk = P*A.topface[k].coords;
      vectred(sk);
      A1.topface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
      vector<int> wk = conjugateword(word_p1,shiftIndicesUp(A.midvertices[k].word));
      //      useRelations(wk);
      gen sk = P*A.midvertices[k].coords;
      vectred(sk);
      A1.midvertices.push_back(WVector(sk,wk));
      A1.midvertindices.push_back(A.midvertindices[k]);
  }

  A1.isSurrounded = false;
  for (unsigned k=0; k<A1.sides.size(); k++) {
    A1.freeRidges.push_back(k);
  }
  A1.jstart = A.jstart;

  return A1;
  
}

Prism Polytope::createPinvimage(Prism A) {

  Prism A1;
 
  gen m = Pi*A.mirror.coords;
  vectred(m);
  vector<int> w1 = shiftIndicesDown(conjugateword(word_m1,A.mirror.word));
  A1.mirror = WVector(m,w1);
  useRelations(A1.mirror.word);

  A1.apex = WVector(Pi*A.apex.coords,shiftIndicesDown(conjugateword(word_m1,A.apex.word)));
  vectred(A1.apex.coords);
  useRelations(A1.apex.word);

  A1.apexProjection = Pi*A.apexProjection;
  vectred(A1.apexProjection);

  A1.cspine = Pi*A.cspine;
  vectred(A1.cspine);
  
  /*  A1.X0 = P*A.X0;
  A1.X1 = P*A.X1;
  vectred(A1.X0);
  vectred(A1.X1);*/

  for (unsigned k=0; k<A.sides.size(); k++) {
    vector<int> wk = shiftIndicesDown(conjugateword(word_m1,A.sides[k].word));
    gen sk = Pi*A.sides[k].coords;
    vectred(sk);
    A1.sides.push_back(WVector(sk,wk));
  }

  for (unsigned k=0; k<A.bottomface.size(); k++) {
    vector<int> wk = shiftIndicesDown(conjugateword(word_m1,A.bottomface[k].word));
    gen sk = Pi*A.bottomface[k].coords;
    vectred(sk);
    A1.bottomface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.topface.size(); k++) {
    vector<int> wk = shiftIndicesDown(conjugateword(word_m1,A.topface[k].word));
    gen sk = Pi*A.topface[k].coords;
    vectred(sk);
    A1.topface.push_back(WVector(sk,wk));
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
    vector<int> wk = shiftIndicesDown(conjugateword(word_m1,A.midvertices[k].word));
    gen sk = Pi*A.midvertices[k].coords;
    vectred(sk);
    A1.midvertices.push_back(WVector(sk,wk));
    A1.midvertindices.push_back(A.midvertindices[k]);
  }

  A1.isSurrounded = false;
  for (unsigned k=0; k<A1.sides.size(); k++) {
    A1.freeRidges.push_back(k);
  }
  A1.jstart = A.jstart;

  return A1;

}

// this is better when P has infinite order, but here it doesn't make much sense...
void Polytope::addFaceOrbit(Prism A) {
    if (isPrismNew(A)) {

      vector<Prism> facestoadd;

      faces.push_back(A);
      faceReps.push_back(A);
      cout << "Adding " << A << endl;
      updateFreeRidges();
      bool done = false;
      Prism Ak = A.mycopy();
      Prism Aki = A.mycopy();
      unsigned k=0;
      while (!done && k<100) {
	Ak = createPimage(Ak);
	done = comparePrisms(Ak,Aki);
	if (!done) {
	  useRelations(Ak);
	  facestoadd.push_back(Ak);
	  Aki = createPinvimage(Aki);
	  done = comparePrisms(Aki,Ak);
	  if (!done) {
	    useRelations(Aki);
	    facestoadd.push_back(Aki);
	  }
	}
	k++;
      }
      if (!done) {
	cerr << "P does not have finite order? (its order is at least 100) " << endl;
	exit(1);
      }

      vector<int> inds;
      int u = facestoadd.size();
      //      cout << "Nb faces to add: " << u << endl;
      if (u%2!=0) {
	u--;
	inds.push_back(u);
      }
      while (u>1) {
	u--;
	inds.push_back(u);
	u--;
	inds.insert(inds.begin(),u);
      }
      //      cout << "Indices: " << inds << endl;

      vector<int> inds2;
      int nf = faces.size();
      inds2.push_back(nf-1);
      for (int k=0; k<inds.size(); k++) {
	//	cout << "k=" << k << endl;
	inds2.push_back(nf+k);
      }
      //      cout << "New face orbit: " << inds2 << endl;
      facePOrbits.push_back(inds2);

      for (unsigned k=0; k<inds.size(); k++) {
	faces.push_back(facestoadd[inds[k]]);
	cout << "Adding " << facestoadd[inds[k]] << endl;
	updateFreeRidges();
      }


    }
    else {
      cout << "Should not add this prism, it's already in the list!" << endl;
    }
}

vector<WVector> Polytope::findMirrors(gen V) {
  vector<WVector> res;
  for (unsigned k=0; k<faces.size(); k++) {
    if (isOnMirror(V,faces[k])) {
      if (findInList(faces[k].mirror,res)<0) {
	res.push_back(faces[k].mirror);
      }
    }
    if (isOnTopFace(V,faces[k])) {
      if (findInList(faces[k].apex,res)<0) {
	res.push_back(faces[k].apex);
      }
    }
  }
  return res;
}

vector<WVector> Polytope::findMirrors(WVector v) {
  return findMirrors(v.coords);
}

vector<int> Polytope::findFaces(gen V) {
  vector<int> res;
  gen te = SqNorm(V);
  if (operator_equal(te,zeronumber,&ct)) {
    res.push_back(-1);
  }
  for (unsigned k=0; k<faces.size(); k++) {
    if (isOnPrism(V,faces[k])) {
      bool fd = false;
      unsigned u=0;
      while (!fd && u<faces[k].topface.size()) {
	fd = areDep(V,faces[k].topface[u].coords);
	u++;
      }
      if (!fd) {
	u=0;
	while (!fd && u<faces[k].midvertices.size()) {
	  fd = areDep(V,faces[k].midvertices[u].coords);
	  u++;
	}
	if (!fd) {
	  u=0;
	  while (!fd && u<faces[k].bottomface.size()) {
	    fd = areDep(V,faces[k].bottomface[u].coords);
	    u++;
	  }
	  if (fd) {
	    res.push_back(k);
	  }
	}
	else {
	  res.push_back(k);
	}
      }
      else {
	res.push_back(k);
      }
    }
  }
  return res;
}

// beware this finds more faces than "findFaces", in general
vector<int> Polytope::findBisectors(gen V) {
  vector<int> res;
  gen te = SqNorm(V);
  if (operator_equal(te,zeronumber,&ct)) {
    res.push_back(-1);
  }
  for (unsigned k=0; k<faces.size(); k++) {
    if (isOnPrism(V,faces[k])) {
      res.push_back(k);
    }
  }
  return res;
}


vector<int> Polytope::findFaces(WVector v) {
  return findFaces(v.coords);
}

vector<int> Polytope::findBisectors(WVector v) {
  return findBisectors(v.coords);
}

void Polytope::useRelations(vector<int> &w) {
  unsigned n = w.size();
  bool reduced = false;
  while (n>0 && !reduced) {
    unsigned ns = w.size();
    for (unsigned k=0; k<relations.size(); k++)  {
      vector<int> rel = relations[k];
      if (2*w.size()>rel.size()) {
	simplifyword(w,rel);
      }
    }
    reduced = ns==w.size();
  }
}

void Polytope::useRelations(Prism &A) {
  useRelations(A.mirror.word);
  for (unsigned k=0; k<A.sides.size(); k++) {
    useRelations(A.sides[k].word);
  }
  for (unsigned k=0; k<A.bottomface.size(); k++) {
    useRelations(A.bottomface[k].word); 
 }
  for (unsigned k=0; k<A.topface.size(); k++) {
    useRelations(A.topface[k].word);
  }
  for (unsigned k=0; k<A.midvertices.size(); k++) {
    useRelations(A.midvertices[k].word);
  }
}

void Polytope::studyVertexStabilizers() {
  for (unsigned k=0; k<vertReps.size(); k++) {
    orbitsToVisit.push_back(k);
  }
  while (orbitsToVisit.size()>0) {
    int k = orbitsToVisit[0];
    cout << "TEST k=" << k << endl;
    currentVertexOrbit.clear();
    findNextVertexInCycle(k);
  }
  for (unsigned k=0; k<vertexOrbits.size(); k++) {
    cout << "Vertex orbit #" << k << ": " << vertexOrbits[k] << endl;
  }

}

void Polytope::findNextVertexInCycle(unsigned k) {

  currentVertexOrbit.push_back(k);

  cout << "Current vertex orbit " << currentVertexOrbit << endl;
  cout << "Orbits to visit: " << orbitsToVisit << endl;

  bool fd = false;
  int kt = 0;
  while (!fd && kt<orbitsToVisit.size()) {
    fd = orbitsToVisit[kt]==k;
    kt++;
  }
  if (!fd) {
    cerr << "Problem..." << endl;
    exit(1);
  }
  else {
    orbitsToVisit.erase(orbitsToVisit.begin()+(kt-1));
  }

  cout << "Orbits to visit (after): " << orbitsToVisit << endl;

  vector<int> te = findFaces(vertReps[k]);
  cout << endl;
  int yo = findInList(vertReps[k],vertices);
  cout << "Vertex rep #" << k << " (#" << yo << " in list) is on bisectors " << te << endl;
  for (unsigned l=0; l<te.size(); l++) {
    if (te[l]>-1) {
      cout << "Applying pairing for face #" << te[l] << endl;
      cout << "te=" << te << endl;

      /*      cout << "Computing images of neighboring prisms... " << endl;
      vector<Prism> Bli;
      for (unsigned m=0; m<te.size(); m++) {
	if (te[m]>-1) {
	  Bli.push_back(applyMatrix(faces[te[l]].pairing,faces[te[m]]));
	}
	}*/

      gen vimage = faces[te[l]].pairing*vertReps[k].coords;
      vectred(vimage);
      int m = findInList(vimage,vertices);
      if (m>-1) {
	cout << "Image by pairing is vertex #" << m << ", which is on: " << findFaces(vimage) << endl;
	bool foundOrb = false;
	int u = 0;
	while (!foundOrb && u<vertexPOrbits.size()) {
	  foundOrb = find(vertexPOrbits[u].begin(), vertexPOrbits[u].end(), m)!=vertexPOrbits[u].end();
	  u++;
	}
	if (!foundOrb) {
	  cerr << "Image is in none of the orbits?!" << endl;
	  exit(1);
	}
	else {
	  cout << "Image is in orbit #" << (u-1) << endl;
	  bool fd = false;
	  int kt = 0;
	  while (!fd && kt<currentVertexOrbit.size()) {
	    fd = currentVertexOrbit[kt]==(u-1);
	    kt++;
	  }
	  if (!fd) {
	    cout << "  this was not visited yet..." << endl;
	    findNextVertexInCycle(u-1);
	  }
	  else {
	    cout << "  this was already visited!" << endl;	    
	  }
	}
      }
      else {
	cerr << "Problem with vertex cycles... " << endl;
	exit(1);
      }
    }
  }

  cout << "Registering orbit: " << currentVertexOrbit << endl;
  cout << "Clearing current vertex orbit..." << endl;
  vertexOrbits.push_back(currentVertexOrbit);
  currentVertexOrbit.clear();

  cout << "Done with findNextVertex, k=" << k << endl;
  
}

void Polytope::computeStabilizer(int k) {
  // stab of vertex representative #k
  
  vector<int> te = vertices[vertexPOrbits[k][0]].faceindices; 
  // I hope this was computed before!!
  
  //  vector<int> te = findFaces(vertReps[k]);
  cout << endl;
  cout << "vertex #" << vertexPOrbits[k][0] << " is on: " << te << endl;
  for (unsigned l=0; l<te.size(); l++) {
    if (te[l]>-1) {
      cout << "Applying pairing for face #" << te[l] << endl;

      /*      cout << "Computing images of neighboring prisms... " << endl;
      vector<Prism> Bli;
      for (unsigned m=0; m<te.size(); m++) {
	if (te[m]>-1) {
	  Bli.push_back(applyMatrix(faces[te[l]].pairing,faces[te[m]]));
	}
	}*/

      gen vimage = faces[te[l]].pairing*vertReps[k].coords;
      vectred(vimage);
      int m = findInList(vimage,vertices);
      if (m!=-1) {
	cout << "  image by pairing is vertex #" << m << ", which is on: " << vertices[m].faceindices << endl;
	// I hope this was computed before!!

	//	cout << "Image by pairing is vertex #" << m << ", which is on: " << findFaces(vimage) << endl;

	vector<int> vp;
	vp.push_back(vertexPOrbits[k][0]);
	vp.push_back(te[l]);
	vp.push_back(m);
	vertexPairings.push_back(vp);
	cout << "v#" << vp[0] << " --(" << vp[1] << ")--> v#" << vp[2] << endl; 
	vector<int> vpi;
	vpi.push_back(m);
	vpi.push_back(facePairings[te[l]]);
	vpi.push_back(vertexPOrbits[k][0]);
	vertexPairings.push_back(vpi);
	cout << "v#" << vpi[0] << " --(" << vpi[1] << ")--> v#" << vpi[2] << endl; 
	// probably would like to include its P-conjugates?

	bool fdm = false;
	int v = -1;
	int w;
	while (!fdm && v!=vertexPOrbits.size()-1) {
	  v++;
	  w=-1;
	  while (!fdm && w!=vertexPOrbits[v].size()-1) {
	    w++;
	    fdm = m==vertexPOrbits[v][w];
	  }
	}
	if (!fdm) {
	  cerr << "Problem finding vertex in some orbit!" << endl;
	  exit(1);
	}

	bool fdm1 = false;
	int v1 = -1;
	int w1;
	while (!fdm1 && v1!=facePOrbits.size()-1) {
	  v1++;
	  w1=-1;
	  while (!fdm1 && w1!=facePOrbits[v1].size()-1) {
	    w1++;
	    fdm1 = te[l]==facePOrbits[v1][w1];
	  }
	}
	if (!fdm1) {
	  cerr << "Problem finding face index in some orbit!" << endl;
	  exit(1);
	}

	unsigned n = vertexPOrbits[v].size();
	unsigned n1 = facePOrbits[v1].size();
	if (n>n1) {
	  cout << "Interesting, a vertex has smaller P-orbit than the face.." << endl;	  
	}
	for (unsigned u=1; u<n; u++) {
	  vector<int> vpu;
	  vpu.push_back(vertexPOrbits[k][u]);
	  vpu.push_back(facePOrbits[v1][(w1+u)%n1]);
	  vpu.push_back(vertexPOrbits[v][(w+u)%n]);
	  vertexPairings.push_back(vpu);
	  cout << "v#" << vpu[0] << " --(" << vpu[1] << ")--> v#" << vpu[2] << endl; 
	  vector<int> vpui;
	  vpui.push_back(vertexPOrbits[v][(w+u)%n]);
	  vpui.push_back(facePairings[facePOrbits[v1][(w1+u)%n1]]);
	  vpui.push_back(vertexPOrbits[k][u]);
	  vertexPairings.push_back(vpui);
	  cout << "v#" << vpui[0] << " --(" << vpui[1] << ")--> v#" << vpui[2] << endl; 
	}

	bool foundOrb = false;
	int u = 0;
	int vv;
	while (!foundOrb && u<vertexPOrbits.size()) {
	  vv = -1;
	  while (!foundOrb && vv!=vertexPOrbits[u].size()-1) {
	    vv++;
	    foundOrb = vertexPOrbits[u][vv]==m;
	  }
	  //	  foundOrb = find(vertexPOrbits[u].begin(), vertexPOrbits[u].end(), m)!=vertexPOrbits[u].end();
	  u++;
	}
	if (!foundOrb) {
	  cerr << "Image is in none of the orbits?!" << endl;
	  exit(1);
	}
	else {
	  vector<int> pa;
	  pa.push_back(k);
	  pa.push_back(u-1);
	  vertexPOrbitPairings.push_back(pa);
	  vector<int> w;
	  if (faces[te[l]].invertpairing) {
	    w = invertword(faces[te[l]].mirror.word);
	  }
	  else {
	    w = concat(w,faces[te[l]].mirror.word);
	  }

	  if (vv>orderofp/2) {
	    for (unsigned zz=0; zz<orderofp-vv; zz++) { // could do better, if v is more than order of P
	      w.insert(w.begin(),4);
	    }
	  }
	  else {
	    //	  cout << "vv=" << vv << endl;
	    for (unsigned zz=0; zz<vv; zz++) { // could do better, if v is more than order of P
	      w.insert(w.begin(),-4);
	    }
	  }
	  vertexPOrbitPairingMaps.push_back(w);
	  cout << "   (image is in orbit #" << (u-1) << ")" << endl;
	}
      }
      else {
	cerr << "Problem with vertex cycles... " << endl;
	exit(1);
      }
    }
  }
}

gen Polytope::evalWord(vector<int> w) {
  gen res = Id;
  unsigned k=0; 
  while (k<w.size()) {
    int l = w[k];
    if (l==1) {
      res = res*R1;
    }
    else {
      if (l==2) {
	res = res*R2;
      }
      else {
	if (l==3) {
	  res = res*R3;
	}
	else {
	  if (l==-1) {
	    res = res*R1i;
	  }
	  else {
	    if (l==-2) {
	      res = res*R2i;
	    }
	    else {
	      if (l==-3) {
		res = res*R3i;
	      }
	      else {
		if (l==4) {
		  res = res*P;
		}
		else {
		  if (l==-4) {
		    res = res*Pi;
		  }
		  else {
		    cerr << "Wrong index in word?" << endl;
		    exit(1);
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    matred(res);
    k++;
  }
  return res;
}

void Polytope::createBraidRelation(vector<int> w1, vector<int> w2) {
  gen A = evalWord(w1);
  gen B = evalWord(w2);
  //  cout << "Creating braid relation for " << w1 << " and " << w2 << endl;

  int bo = braidOrder(A,B);
  int bo2 = bo/2;

  cout << convertword(w1) << " and " << convertword(w2) << "  braid with order " << bo << endl;

  vector<int> w1i = invertword(w1);
  vector<int> w2i = invertword(w2);
  
  vector<int> r1;
  for (unsigned k=0; k<bo2; k++) {
    r1 = concat(r1,w1);
    r1 = concat(r1,w2);
  }
  if (bo%2!=0) {
    r1 = concat(r1,w1);
    r1 = concat(r1,w2i);
  }
  for (unsigned k=0; k<bo2; k++) {
    r1 = concat(r1,w1i);
    r1 = concat(r1,w2i);
  }

  vector<int> r2;
  for (unsigned k=0; k<bo2; k++) {
    r2 = concat(r2,w2);
    r2 = concat(r2,w1);
  }
  if (bo%2!=0) {
    r2 = concat(r2,w2);
    r2 = concat(r2,w1i);

  }
  for (unsigned k=0; k<bo2; k++) {
    r2 = concat(r2,w2i);
    r2 = concat(r2,w1i);
  }

  //  cout << "Created braid relation: " << r1 << endl;
  //  cout << "Created braid relation: " << r2 << endl;

  relations.push_back(r1);
  relations.push_back(r2);
  relations.push_back(shiftIndicesUp(r1));
  relations.push_back(shiftIndicesUp(r2));
  relations.push_back(shiftIndicesDown(r1));
  relations.push_back(shiftIndicesDown(r2));  

}

bool Polytope::isVertexSafe(WVector V, Prism A) {
  for (unsigned j=0; j<A.topface.size(); j++) {
    if (testSame(V,A.topface[j])) {
      return true;
    }
  }
  for (unsigned j=0; j<A.midvertices.size(); j++) {
    if (testSame(V,A.midvertices[j])) {
      return true;
    }
  }
  for (unsigned j=0; j<A.bottomface.size(); j++) {
    if (testSame(V,A.bottomface[j])) {
      return true;
    }
  }
  return false;  
}

int Polytope::applyPairingToVertex(int jface, int jvertex) {
  bool fd = false;
  int u = -1;
  while (!fd && u!=vertexPairings.size()-1) {
    u++;
    fd = ( vertexPairings[u][0]==jvertex && vertexPairings[u][1]==jface );
  }
  if (fd) {
    return vertexPairings[u][2];
  }
  else {
    return -1;
  }
}

int Polytope::applyPairingToEdge(int jface, int j) {
  int jo = applyPairingToVertex(jface, edges[j].origin);
  int je = applyPairingToVertex(jface, edges[j].end);
  cout << "jo=" << jo << endl;
  cout << "je=" << je << endl;
  if (jo<0 || je<0) {
    cerr << "Trouble computing image of a 1-face (vertices are already a problem)!" << endl;
    exit(1);
  }
  else {
    bool fd = false;
    int u = -1;
    while (!fd && u!=edges.size()) {
      u++;
      fd = ( (edges[u].origin==jo && edges[u].end==je) || (edges[u].origin==je && edges[u].end==jo) ) ;
    } 
    if (!fd) {
      cerr << "Trouble computing image of a 1-face!" << endl;
      exit(1);
    }
    else {
      return u;
    }
  }
}


int Polytope::applyPairingToRidge(int jface, int j) {
  unsigned nv = ridges[j].verts.size();
  vector<int> verteximages;
  for (unsigned u=0; u<nv; u++) {
    int jo = applyPairingToVertex(jface, ridges[j].verts[u]);
    //    cout << "Search for v#" << ridges[j].verts[u] << ", mapped by pairing#" << jface << "    result: " << jo << endl;
    if (jo<0) {
      cerr << "Trouble computing image of a ridge!" << endl;
      exit(1);
    }
    else {
      verteximages.push_back(jo);
    }
  }
  vector<int> commoninds;
  for (unsigned u=0; u<vertices[verteximages[0]].faceindices.size(); u++) {
    commoninds.push_back(vertices[verteximages[0]].faceindices[u]);
  }
  for (unsigned u=1; u<verteximages.size(); u++) {
    commoninds = intersect(commoninds,vertices[verteximages[u]].faceindices);
  }
  if (commoninds[0]==-1) {
    commoninds.erase(commoninds.begin());
  }
  if (commoninds.size()!=2) {
    cerr << "Trouble identifying image of a ridge (nb of indices containing verts: "<<commoninds.size()<<")" << endl;
    exit(1);
  }
  else {
    bool fd = false;
    int v=-1;
    while (!fd && v!=ridges.size()-1) {
      v++;
      fd = ( (ridges[v].faceindices[0]==commoninds[0] && ridges[v].faceindices[1]==commoninds[1]) || (ridges[v].faceindices[1]==commoninds[0] && ridges[v].faceindices[0]==commoninds[1]));
    }
    if (!fd) {
      cerr << "Trouble identifying image of a ridge." << endl;
      exit(1);
    }
    else {
      return v;
    }
  }
}

void Polytope::checkCycles(){

  ofstream file;
  string filename = "pres/pres"+tag+".g";
  file.open(filename.c_str());
  file << "F:=FreeGroup(\"R1\",\"R2\",\"R3\",\"P\");" << endl;
  file << "R1:=F.1; R2:=F.2; R3:=F.3; P:=F.4;" << endl;
  stringstream ops;
  ops << orderofp;
  file << "G:=F/[P*R3*P^-1*R1^-1,P^" << ops.str();

  vector<string> rels;
  vector<int> ridgeorbitstocheck;
  for (unsigned k=0; k<ridgePOrbits.size(); k++) {
    ridgeorbitstocheck.push_back(k);
  }

  while (ridgeorbitstocheck.size()>0) {
    //    cout << "Ridge orbits to check: " << convertword(ridgeorbitstocheck) << endl;
    //    cout << "ridgePOrbits[ridgeorbitstocheck[0]]= " << ridgePOrbits[ridgeorbitstocheck[0]] << endl;
    cout << "Computing cycle for ridge on faces " << ridges[ridgePOrbits[ridgeorbitstocheck[0]][0]] << endl;

    gen C;
    vector<int> cw;

    int init_orbit_index = ridgeorbitstocheck[0];
    cout << "Initial ridge orbit: " << init_orbit_index << endl;

    gen mi0;
    bool iscomplex = isRidgeComplex(ridgePOrbits[init_orbit_index][0],mi0);
    gen Xt, angle_sum;
    angle_sum = zeronumber;
    if (iscomplex) {
      cout << "Should compute angles!" << endl;
      Prism P1 = faces[ridges[ridgePOrbits[init_orbit_index][0]].faceindices[0]];
      Xt = BoxProd(mi0,P1.cspine);
      gen angle_fraction = computeAngle(ridgePOrbits[init_orbit_index][0], Xt);
      //      cout << "angle: 1/" << onenumber/angle_fraction << endl;
      angle_sum = angle_sum + angle_fraction;
    }

    unsigned jstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[0];
    unsigned kstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[1];
    if (faces[jstart].invertpairing) {
      cw = invertword(faces[jstart].mirror.word);
    }
    else  {
      cw = faces[jstart].mirror.word;      
    }
    C = faces[jstart].pairing;
    Xt = C*Xt;
    vectred(Xt);
    
    //    int init_ridge = ridgePOrbits[init_orbit_index][0];
    int current_ridge = ridgePOrbits[init_orbit_index][0];
    int j = jstart;

    bool back = false;
    while (!back) {

      current_ridge = applyPairingToRidge(j,current_ridge);
      bool fd = false;
      int u = -1;
      int v;
      while (!fd && u!=ridgePOrbits.size()-1) {
	u++;
	v = -1;
	while (!fd && v!=ridgePOrbits[u].size()-1) {
	  v++;
	  fd = current_ridge == ridgePOrbits[u][v]; 
	}
      }
      if (!fd) {
	cerr << "Trouble finding image ridge in a ridge P-orbit" << endl;
	exit(1);
      }
      else {
	//	cout << "u=" << u << endl;
	ridgeorbitstocheck.erase(remove(ridgeorbitstocheck.begin(),ridgeorbitstocheck.end(),u),ridgeorbitstocheck.end());
	if (ridges[current_ridge].faceindices[0]==facePairings[j]) {
	  j = ridges[current_ridge].faceindices[1];
	}
	else {
	  if (ridges[current_ridge].faceindices[1]==facePairings[j]) {
	    j = ridges[current_ridge].faceindices[0];
	  }
	  else {
	    cerr << "Trouble finding next pairing" << endl;
	    exit(1);
	  }
	}
	back = (u==init_orbit_index);
	if (!back) {
	  if (faces[j].invertpairing) {
	    cw = concat(invertword(faces[j].mirror.word),cw);
	  }
	  else  {
	    cw = concat(faces[j].mirror.word,cw);      
	  }
	  C = faces[j].pairing*C;
	  matred(C);
	  Xt = faces[j].pairing*Xt;
	  vectred(Xt);
	  if (iscomplex) {
	    Prism P1 = faces[ridges[current_ridge].faceindices[0]];
	    Xt = BoxProd(mi0,P1.cspine);
	    gen angle_fraction = computeAngle(current_ridge, Xt);
	    //	    cout << "angle: 1/" << onenumber/angle_fraction << endl;
	    angle_sum = angle_sum + angle_fraction;
	  }
	}
	else {
	  vector<int> power_of_p;
	  if (2*v<orderofp) {
	    for (int k=0; k<v; k++) {
	      power_of_p.push_back(-4);
	      C = Pi*C;
	      matred(C);
	    }
	  }
	  else {
	    int v2 = orderofp-v;
	    for (int k=0; k<v2; k++) {
	      power_of_p.push_back(4);
	      C = P*C;
	      matred(C);
	    }
	  }
	  cw = concat(power_of_p,cw);	  
	}
      }
    }
    if (!back) {
      cerr << "Trouble with some ridge cycle" << endl;
      exit(1);
    }
    else {
      cout << "Got back to same ridge, word so far is " << convertword(cw) << endl;
      //      useRelations(cw);
      //      cout << "    reduced using relations:  " << convertword(cw) << endl;
      bool isid = false;
      gen tv_start = vertices[ridges[jstart].verts[0]].coords;
      gen tv = vertices[ridges[jstart].verts[0]].coords;
      int pow = 0;
      vector<int> cwp;
      gen Cp = Id;
      //      cout << "Angle sum when back: 1/" << onenumber/angle_sum << endl;
      while (!isid && pow<100) {
	cwp = concat(cwp,cw);
	Cp = Cp*C;
	matred(Cp);
	tv = C*tv;
	vectred(tv);
	isid = areDep(tv,tv_start);
	pow++;
      }
      if (!isid) {
	cerr << "Did not find a small power of cycle tsf that's the identity on the ridge (tried powers up to 100)" << endl;
	exit(1);
      }
      else {
	int o = order(Cp); // I would expect order(C) rather!?!
	int op = o*pow;
	cout << "(" << convertword(cw) << ")^" << pow << "=id on the ridge" << endl;
	cout << "(" << convertword(cw) << ")^" << op << "=id (need to check angles)" << endl;

	if (iscomplex) {

	  if (order(C)==2) {
	    cout << "Order 2, nothing to check for local tiling" << endl;
	  }
	  else {
	    gen XtImage = C*Xt;
	    vectred(XtImage);
	    if (!areDep(Xt,XtImage)) {
	      
	      cout << "First attempt to compute angle failed, will try using isolated fixed point of cycle!" << endl;	    
	      gen Xtt;
	      
	      int p;
	      if (tag.at(1)=='s') {
		p = (int) tag.at(0) - 48;
	      }
	      else {
		if (tag.at(2)=='s') {
		  p = ((int) tag.at(0) - 48)*10 + ((int) tag.at(1) - 48);
		}
		else {
		  cerr << "Could not read tag, is it longer than 2 digits??" << endl;
		  exit(1);
		}
	      }
	      int max = 3*order(C)*p;
	      unsigned ndsave = nd;
	      increasePrecision();
	      bool found = false;
	      gen Hn = evali(H);
	      
	      //	    gen zei = gen("[-0.000000000000000000001..0.00000000000000000001]",&ct);
	      //	    gen oni = gen("[0.999999999999999999999..1.00000000000000000001]",&ct);
	      gen zei = convert_interval(onenumber,3*nd,&ct)-onenumber;
	      gen oni = convert_interval(onenumber,3*nd,&ct);
	      vecteur r1;
	      r1.push_back(oni);
	      r1.push_back(zei);
	      r1.push_back(zei);
	      vecteur r2;
	      r2.push_back(zei);
	      r2.push_back(oni);
	      r2.push_back(zei);
	      vecteur r3;
	      r3.push_back(zei);
	      r3.push_back(zei);
	      r3.push_back(oni);
	      vecteur rr;
	      rr.push_back(r1);
	      rr.push_back(r2);
	      rr.push_back(r3);
	      gen Idn = gen(rr);
	      //	    cout << "Idn: " << Idn << endl;
	      
	      gen cyclo = gen("x^" + print_INT_(max)+"-1",&ct);
	      gen sols = _complexroot(makesequence(cyclo,prec),&ct);
	      
	      //	    cout << "Nb sols: " << (*sols._VECTptr).size() << endl;
	      
	      unsigned uu=0;	    
	      while (!found && uu!=max) {
		
		gen ev = sols[uu][0]; // this is not always an interval, don't know if that's an issue (1 is an exception)
		//	      cout << "ev=" << ev << endl;
		gen evr = re(ev,&ct);
		gen evi = im(ev,&ct);
		if (operator_equal(evr,_left(evr,&ct),&ct) && operator_equal(evi,_left(evi,&ct),&ct)) {
		  //		cout << "This seems not to be an interval, I will convert it to interval!" << endl;
		  if (operator_equal(evr,zeronumber,&ct)) {
		    evr = convert_interval(onenumber,nd,&ct)-onenumber;
		  }
		  else {
		    evr = convert_interval(evr,nd,&ct);
		  }
		  if (operator_equal(evi,zeronumber,&ct)) {
		    evi = convert_interval(onenumber,nd,&ct)-onenumber;
		  }
		  else {
		    evi = convert_interval(evi,nd,&ct);
		  }
		  ev = evr + gen("i",&ct)*evi;
		  //		cout << "new ev=" << ev << endl;
		}
		/*	      else {
			      cout << "   seems to be an interval!" << endl;		  
			      }*/
		
		gen B = C-l__IDNT_e*Id;
		gen db = det(B);
		//	      cout << "db=" << db << endl;
		
		gen de = _subst(makesequence(_subst(makesequence(db,l__IDNT_e,ev),&ct),x__IDNT_e,xvi),&ct);
		//	      cout << "de=" << de << endl;
		
		gen test_det = _abs(de,&ct);
		/*	      cout << "test_det: " << test_det << endl;
	      cout << "right(test_det): " << _right(test_det,&ct) << endl;
	      cout << "greater? " << is_greater(1e-5,_right(test_det,&ct),&ct) << endl;*/

		if (is_greater(1e-5,_right(test_det,&ct),&ct)) { // test_det may not be an interval...
		  
		  //		cout << "Found an eigenvector..." << endl;
		  //		cout << "ev:" << ev << endl;		
		  //		cout << "evali(C): " << evali(C) << endl;
		  
		  //		gen M = matevali(C) - ev*matevali(Id);
		  //		cout << "M: " << M << endl;
		  int ii = 0;
		  bool subfound = false;
		  while (!subfound && ii<3) {
		    //		  cout << "ii=" << ii << endl;
		    gen A = matevali(C)-ev*Idn;
		    gen te1 = _row(makesequence(A,ii),&ct);
		    gen te2 = _row(makesequence(A,(ii+1)%3),&ct);
		    vecteur U;
		    U.push_back(te1[1]*te2[2]-te1[2]*te2[1]);
		    U.push_back(te1[2]*te2[0]-te1[0]*te2[2]);
		    U.push_back(te1[0]*te2[1]-te1[1]*te2[0]);
		    Xtt = gen(U);
		    //		  cout << "Xtt: " << Xtt << endl;
		    gen maxcomp = zeronumber;
		    for (unsigned ll=0; ll<3; ll++) {
		      if (is_greater(_abs(Xtt[ll],&ct),maxcomp,&ct)) {
			maxcomp = _abs(Xtt[ll],&ct);
		      }
		    }
		    //		  cout << "maxcomp: " << maxcomp << endl;
		    subfound = is_greater(maxcomp,1e-12,&ct); 
		    if (subfound) {
		      gen tesn = conj(Xtt,&ct)*Hn*Xtt;
		      //		    cout << "Sq nm: " << tesn << endl;
		      found = is_greater(zeronumber,_right(re(tesn,&ct),&ct),&ct);
		    }
		    ii++;
		  }
		  if (found) {
		    //		  cout << "Xtt: " << Xtt << endl;
		    cout << "Found fixed point of cycle tsf" << endl;
		    cout << "Will now run through cycle again, and compute angles at fixed point" << endl;
		    
		    gen angle_fraction = computeAngleNum(ridgePOrbits[init_orbit_index][0], Xtt);
		    //		  cout << "angle(1): 1/" << onenumber/angle_fraction << endl;
		    angle_sum = angle_fraction;
		    
		    // need to compute the start the cycle over, and compute angle at Xtt
		    jstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[0];
		    kstart = ridges[ridgePOrbits[init_orbit_index][0]].faceindices[1];
		    
		    //		  cout << "matevali(...) " << matevali(faces[jstart].pairing) << endl;
		    
		    Xtt = matevali(faces[jstart].pairing)*Xtt;
		    
		    //		  cout << "Xtt: " << Xtt << endl;
		    
		    current_ridge = ridgePOrbits[init_orbit_index][0];
		    j = jstart;
		    
		    back = false;
		    while (!back) {
		      
		      current_ridge = applyPairingToRidge(j,current_ridge);
		      bool fd = false;
		      int u = -1;
		      int v;
		      while (!fd && u!=ridgePOrbits.size()-1) {
			u++;
			v = -1;
			while (!fd && v!=ridgePOrbits[u].size()-1) {
			  v++;
			  fd = current_ridge == ridgePOrbits[u][v]; 
			}
		      }
		      if (!fd) {
			cerr << "Trouble finding image ridge in a ridge P-orbit" << endl;
			exit(1);
		      }
		      else {
			//   cout << "u=" << u << endl;
			ridgeorbitstocheck.erase(remove(ridgeorbitstocheck.begin(),ridgeorbitstocheck.end(),u),ridgeorbitstocheck.end());
			if (ridges[current_ridge].faceindices[0]==facePairings[j]) {
			  j = ridges[current_ridge].faceindices[1];
			}
			else {
			  if (ridges[current_ridge].faceindices[1]==facePairings[j]) {
			    j = ridges[current_ridge].faceindices[0];
			  }
			  else {
			    cerr << "Trouble finding next pairing" << endl;
			    exit(1);
			  }
			}
			back = (u==init_orbit_index);
			if (!back) {
			  Xtt = matevali(faces[j].pairing)*Xtt;
			  gen angle_fraction = computeAngleNum(current_ridge, Xtt);
			  //			cout << "angle(2): 1/" << onenumber/angle_fraction << endl;
			  angle_sum = angle_sum + angle_fraction;
			}
		      }
		    }
		  }
		}
		uu++;
	      }
	      nd = ndsave;
	      lowerPrecision();
	      if (!found) {
		cerr << "Trouble finding fixed point of cycle tsf :(" << endl;
		// should adjust precision...
		exit(1);
	      }
	    }
	    // should check if Xt is fixed by C, if not the computation below is not valid!
	    cout << "angle_sum=" << angle_sum << endl;
	    gen te_angle = onenumber/(op*(angle_sum));
	    cout << "te_angle=" << te_angle << endl;
	    gen width = _right(te_angle,&ct) - _left(te_angle,&ct);
	    if (is_greater(width,onenumber,&ct)) {
	      cerr << "Ambiguous angle computations, interval contains several integers!" << endl;
	      exit(1);
	    }
	    else {
	      gen i = _floor(_right(te_angle,&ct),&ct);
	      cout << "i=" << i << endl;
	      if (!is_greater(i,_left(te_angle,&ct),&ct)) {
		cerr << "The angle interval contains no inverse of an integer, perhaps the integrality condition fails!" << endl;
		exit(1);
	      }
	      else {
		if (operator_equal(i,onenumber,&ct)) {
		  cout << "Angle is 2*pi, local tiling is OK!" << endl;
		}
		else {
		  cerr << "Problem with integrality condition, angle is 2*pi/" << i << "!!" << endl;
		  exit(1);
		}
	      }
	    }
	  }
	}

	string rel;
	if (op==1) {
	  rel = convertwordgap(cw);
	  bool isnew = true;
	  int uu = 0;
	  while (isnew && uu<rels.size()) {
	    isnew = (rel!=rels[uu]);
	    uu++;
	  }
	  if (isnew) {
	    rels.push_back(rel);
	  }
	  //	  file << "," << convertwordgap(cw);
	}
	else {
	  stringstream powstring;
	  powstring << op;
	  rel = "(" + convertwordgap(cw) + ")^" + powstring.str();	
	  bool isnew = true;
	  int uu = 0;
	  while (isnew && uu<rels.size()) {
	    isnew = (rel!=rels[uu]);
	    uu++;
	  }
	  if (isnew) {
	    rels.push_back(rel);
	  }
	}
	gen D = Cp;
	for (unsigned u=1; u<o; u++) {
	  D = D*Cp;
	  matred(D);
	}
	if (isScalar(D)) {
	  cout << "Check matrix OK" << endl;
	}
	else {
	  cerr << "Problem with cycle tsf, found a non-scalar matrix!" << endl;
	  exit(1);
	}
	//	cout << "Check matrix: " << D << endl;

	// may need to include commutation relation if ridge is stabilized by power of P
	if (tautag=="s5" && iscomplex) {
	  gen Ci = matinverse(C);
	  gen Ppoweri = matinverse(Ppower);
	  gen M = C*Ppower;
	  matred(M);
	  M = M*Ci;
	  matred(M);
	  M = M*Ppoweri;
	  matred(M);
	  if (isScalar(M)) {
	    stringstream ss;
	    ss << orderofppower;
	    vector<int> cwi = invertword(cw);
	    string commrel = "P^" + ss.str() + "*" + convertwordgap(cw) + "*P^-" + ss.str() + "*" + convertwordgap(cwi);	
	    bool isnew = true;
	    int uu = 0;
	    while (isnew && uu<rels.size()) {
	      isnew = (commrel!=rels[uu]);
	      uu++;
	    }
	    if (isnew) {
	      rels.push_back(commrel);
	    }	    
	  }
	}
	
	cout << "" << endl;
      }
    }    
  }
  for (unsigned uu=0; uu<rels.size(); uu++) {
    file << "," << rels[uu];
  }
  file << "];" << endl;
  file.close();
  cout << "Wrote the presentation in GAP style in the file: " << filename << endl;

}

void Polytope::drawFace(unsigned j) {

  ofstream file;
  stringstream ss;
  ss << j;
  string filename = "pics/pic"+tag+"_face"+ss.str()+".asy";
  file.open(filename.c_str());
  file << "import graph3;" << endl;
  file << "import three;" << endl;
  file << "size(7.5cm,0);" << endl;
  file << "pen p=black+1;" << endl;
  file << "draw(unithemisphere,white+opacity(0.1));" << endl;

  gen Hn = matevaln(H);
  //  gen Hn = _subst(makesequence(H,x__IDNT_e,xv),&ct);
  //  cout << "Hn=" << Hn << endl;
  gen V0 = faces[j].apexProjection;
  //  cout << "V0=" << V0 << endl;
  //  cout << "faces[j] apex: " << faces[j].apex.coords << endl;
  gen V1 = Proj(faces[j].apex.coords,faces[j].apexProjection);
  //  cout << "V1=" << V1 << endl;
  gen V2 = faces[j].cspine;
  //  cout << "V2=" << V2 << endl;
  
  gen W0 = vectevaln(V0) / sqrt(re(-evaln(Inn(V0,V0)),&ct),&ct);
  gen W1 = vectevaln(V1) / sqrt(re(evaln(Inn(V1,V1)),&ct),&ct);
  gen W2 = vectevaln(V2) / sqrt(re(evaln(Inn(V2,V2)),&ct),&ct);

  vecteur te;
  te.push_back(W0);
  te.push_back(W1);
  te.push_back(W2);
  gen A = _tran(gen(te),&ct);
  gen Ai = _inverse(A,&ct);
 
  
  int k1 = 0;
  int l1 = faces[j].sides[0].word.size();
  for (unsigned k=1; k<faces[j].sides.size(); k++) {
    unsigned l = faces[j].sides[k].word.size();
    if (l<l1) {
      l1 = l;
      k1 = k;
    }
  }
  int k2;
  if (k1==0) {
    if (faces[j].sides[faces[j].sides.size()-1].word.size()<faces[j].sides[1].word.size()) {
      k2 = faces[j].sides.size()-1;
    }
    else {
      k2 = 1;
    }
  }
  else {
    k2 = (k1+1)%faces[j].sides.size();
  }
  gen zo,to,zt,tt;

  gen U = Ai*vectevaln(faces[j].bottomface[k1].coords);
  gen zu = U[2]/U[0];
  gen tu = re(U[1]/U[0],&ct);
  gen V = Ai*vectevaln(faces[j].bottomface[k2].coords);
  gen zv = V[2]/V[0];
  gen tv = re(V[1]/V[0],&ct);
  zo = (zu+zv)/2;
  to = (tu+tv)/2;
  int n = faces[j].sides.size();
  if (n%2==0) {
    int k3 = (k1 + n/2)%n;
    int k4 = (k2 + n/2)%n;
    U = Ai*vectevaln(faces[j].bottomface[k3].coords);
    zu = U[2]/U[0];
    tu = re(U[1]/U[0],&ct);
    V = Ai*vectevaln(faces[j].bottomface[k4].coords);
    zv = V[2]/V[0];
    tv = re(V[1]/V[0],&ct);
    zt = (zu+zv)/2;
    tt = (tu+tv)/2;
  }
  else {
    int k3;
    if (k2>k1) {
      k3 = (k2 + n/2)%n;
    }
    else {
      k3 = (k1 + n/2)%n;
    }
    gen U = Ai*vectevaln(faces[j].bottomface[k3].coords);
    zt = U[2]/U[0];
    tt = re(U[1]/U[0],&ct);
  }

  gen Xp = Ai*vectevaln(faces[j].topface[0].coords);
  gen top_level = re(Xp[1]/Xp[0],&ct);
  gen zk = zo-zt;
  zk = zk/abs(zk,&ct);  
  file << "currentprojection=orthographic((" << re(zk,&ct) << "," << im(zk,&ct) << "," << 0.3 << "));" << endl;

  string norm = "(0,0,1)";
  string ccw = "false";
  //  cout << "Top" << endl;
  gen z1,z2,den,X;
  if (faces[j].topface.size()>1) {
    vecteur top_verts;
    gen X = Ai*vectevaln(faces[j].topface[0].coords);
    //    gen top_level = re(X[1]/X[0],&ct);
    gen top_rv = onenumber - top_level*top_level;;
    top_verts.push_back(X[2]/X[0]);
    //    cout << X[2]/X[0] << ",   " << top_level << endl;
    for (unsigned k=1; k<faces[j].topface.size(); k++) {
      X = Ai*vectevaln(faces[j].topface[k].coords);
      top_verts.push_back(X[2]/X[0]);
      //     cout << X[2]/X[0] << ",   " << re(X[1]/X[0],&ct) << endl;
    }
    for (unsigned k=0; k<top_verts.size()-1; k++) {
      z1 = top_verts[k];
      z2 = top_verts[k+1];
      den = conj(z1,&ct)*z2 - z1*conj(z2,&ct);
      if (is_greater(1e-4,abs(den,&ct),&ct)) {
	// draw straight line
	//      file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<top_level<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<top_level<<"),p);" << endl;
      }
      else {
	gen ctr = (top_rv*(z2-z1) + abs(z1,&ct)*abs(z1,&ct)*z2 - abs(z2,&ct)*abs(z2,&ct)*z1 )/den;
      //	gen ctr = top_rv*(z2-z1)/den;
	gen xc = re(ctr,&ct);
	gen yc = im(ctr,&ct);
	//      gen rad = sqrt(xc*xc+yc*yc-top_rv,&ct);
	//	cout << "|z1-ctr|=" << abs(z1-ctr,&ct) << endl;
	//	cout << "|z2-ctr|=" << abs(z2-ctr,&ct) << endl;
	file << "draw(Arc(("<<xc<<","<<yc<<","<<top_level<<"),("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<top_level<<"),("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<top_level<<"),"<<norm<<","<<ccw<<"),p);" << endl;      
      }
    }
    z1 = top_verts[top_verts.size()-1];
    z2 = top_verts[0];
    den = conj(z1,&ct)*z2 - z1*conj(z2,&ct);
    if (is_greater(1e-4,abs(den,&ct),&ct) && is_greater(abs(z1-z2,&ct),1e-4,&ct)) {
      // draw straight line
      //    file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<top_level<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<top_level<<"),p);" << endl;
    }
    else {
      gen ctr = (top_rv*(z2-z1) + abs(z1,&ct)*abs(z1,&ct)*z2 - abs(z2,&ct)*abs(z2,&ct)*z1 )/den;
      //     gen ctr = top_rv*(z2-z1)/den;
      //      cout << "|z1-ctr|=" << abs(z1-ctr,&ct) << endl;
      //      cout << "|z2-ctr|=" << abs(z2-ctr,&ct) << endl;
      gen xc = re(ctr,&ct);
      gen yc = im(ctr,&ct);
      //      gen rad = sqrt(xc*xc+yc*yc-top_rv,&ct);
      file << "draw(Arc(("<<xc<<","<<yc<<","<<top_level<<"),("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<top_level<<"),("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<top_level<<"),"<<norm<<","<<ccw<<"),p);" << endl;      
    }
  }

  //  cout << "Mid" << endl;
  for (unsigned k=0; k<faces[j].midvertices.size(); k++) {
    X = Ai*vectevaln(faces[j].midvertices[k].coords);
    //    cout << X[2]/X[0] << ",   " << re(X[1]/X[0],&ct) << endl;
  }

  //  cout << "Bottom" << endl;
  vecteur bottom_verts;
  X = Ai*vectevaln(faces[j].bottomface[0].coords);
  gen bottom_level = re(X[1]/X[0],&ct);
  gen bottom_rv = onenumber - bottom_level*bottom_level;
  bottom_verts.push_back(X[2]/X[0]);
  //  cout << X[2]/X[0] << ",   " << bottom_level << endl;
  for (unsigned k=1; k<faces[j].bottomface.size(); k++) {
    X = Ai*vectevaln(faces[j].bottomface[k].coords);
    bottom_verts.push_back(X[2]/X[0]);
    //   cout << X[2]/X[0] << ",   " << re(X[1]/X[0],&ct) << endl;
  }
  for (unsigned k=0; k<bottom_verts.size()-1; k++) {
    z1 = bottom_verts[k];
    z2 = bottom_verts[k+1];
    den = conj(z1,&ct)*z2 - z1*conj(z2,&ct);
    if (is_greater(1e-4,abs(den,&ct),&ct)) {
      // draw straight line
      //      file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<bottom_level<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<bottom_level<<"),p);" << endl;
    }
    else {
      gen ctr = (bottom_rv*(z2-z1) + abs(z1,&ct)*abs(z1,&ct)*z2 - abs(z2,&ct)*abs(z2,&ct)*z1 )/den;
      //      cout << "|z1-ctr|=" << abs(z1-ctr,&ct) << endl;
      //     cout << "|z2-ctr|=" << abs(z2-ctr,&ct) << endl;
      gen xc = re(ctr,&ct);
      gen yc = im(ctr,&ct);
      //      gen rad = sqrt(xc*xc+yc*yc-bottom_rv,&ct);
      file << "draw(Arc(("<<xc<<","<<yc<<","<<bottom_level<<"),("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<bottom_level<<"),("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<bottom_level<<"),"<<norm<<","<<ccw<<"),p);" << endl;      
      
      vector<int> w = faces[j].mirror.word;
      if (w.size()<10) {
	file << "label(\"$";
	for (unsigned u=0; u<w.size(); u++) {
	  if (w[u]<0) {
	    file << "\\bar" << -w[u];
	  }
	  else {
	    file << w[u];
	  }
	}
	gen zm = (z1+z2)/2;
	gen zma = ctr + (zm-ctr)*abs(z1-ctr,&ct)/abs(zm-ctr,&ct);
	file << "$\",(" << re(zma,&ct) << "," << im(zma,&ct) << "," << bottom_level << "));" << endl;
      }
    }
  }
  z1 = bottom_verts[bottom_verts.size()-1];
  z2 = bottom_verts[0];
  den = conj(z1,&ct)*z2 - z1*conj(z2,&ct);
  if (is_greater(1e-4,abs(den,&ct),&ct) && is_greater(abs(z1-z2,&ct),1e-4,&ct)) {
    // draw straight line
    //    file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<bottom_level<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<bottom_level<<"),p);" << endl;
  }
  else {
    gen ctr = (bottom_rv*(z2-z1) + abs(z1,&ct)*abs(z1,&ct)*z2 - abs(z2,&ct)*abs(z2,&ct)*z1 )/den;
    //   gen ctr = bottom_rv*(z2-z1)/den;
    //    cout << "|z1-ctr|=" << abs(z1-ctr,&ct) << endl;
    //   cout << "|z2-ctr|=" << abs(z2-ctr,&ct) << endl;
    gen xc = re(ctr,&ct);
    gen yc = im(ctr,&ct);
    //      gen rad = sqrt(xc*xc+yc*yc-bottom_rv,&ct);
    file << "draw(Arc(("<<xc<<","<<yc<<","<<bottom_level<<"),("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<bottom_level<<"),("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<bottom_level<<"),"<<norm<<","<<ccw<<"),p);" << endl;      
    vector<int> w = faces[j].mirror.word;
    if (w.size()<10) {
      file << "label(\"$";
      for (unsigned u=0; u<w.size(); u++) {
	if (w[u]<0) {
	  file << "\\bar" << -w[u];
	}
	else {
	  file << w[u];
	}
      }
      gen zm = (z1+z2)/2;
      gen zma = ctr + (zm-ctr)*abs(z1-ctr,&ct)/abs(zm-ctr,&ct);
      file << "$\",(" << re(zma,&ct) << "," << im(zma,&ct) << "," << bottom_level << "));" << endl;
    }
  }

  // then draw vertical edges...  
  if (faces[j].topface.size()>1) {
    for (unsigned l=0; l<faces[j].bottomface.size(); l++) {
      if (find(faces[j].midvertindices.begin(),faces[j].midvertindices.end(),l)==faces[j].midvertindices.end()) {
	// draw straight line from tf[l] to bf[l]
	gen X1 = Ai*vectevaln(faces[j].topface[l].coords);
	gen z1 = X1[2]/X1[0];
	gen t1 = re(X1[1]/X1[0],&ct);
	gen X2 = Ai*vectevaln(faces[j].bottomface[l].coords);
	gen z2 = X2[2]/X2[0];
	gen t2 = re(X2[1]/X2[0],&ct);
	file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<t1<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<t2<<"),p);" << endl;	
	vector<int> w = faces[j].sides[l].word;
	if (w.size()<10) {
	  file << "label(\"$";
	  for (unsigned u=0; u<w.size(); u++) {
	    if (w[u]<0) {
	      file << "\\bar" << -w[u];
	    }
	    else {
	      file << w[u];
	    }
	  }
	  gen zm = (z1+z2)/2;
	  gen tm = (t1+t2)/2;
	  file << "$\",(" << re(zm,&ct) << "," << im(zm,&ct) << "," << tm << "),align=right);" << endl;
	}
      }
    }
  }
  else {
    for (unsigned l=0; l<faces[j].bottomface.size(); l++) {
      if (find(faces[j].midvertindices.begin(),faces[j].midvertindices.end(),l)==faces[j].midvertindices.end()) {
	// draw straight line from bf[l] to tf[0]
	gen X1 = Ai*vectevaln(faces[j].topface[0].coords);
	gen z1 = X1[2]/X1[0];
	gen t1 = re(X1[1]/X1[0],&ct);
	gen X2 = Ai*vectevaln(faces[j].bottomface[l].coords);
	gen z2 = X2[2]/X2[0];
	gen t2 = re(X2[1]/X2[0],&ct);
	file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<t1<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<t2<<"),p);" << endl;	
	vector<int> w = faces[j].sides[l].word;
	if (w.size()<10) {
	  file << "label(\"$";
	  for (unsigned u=0; u<w.size(); u++) {
	    if (w[u]<0) {
	      file << "\\bar" << -w[u];
	    }
	    else {
	      file << w[u];
	  }
	  }
	  gen zm = (z1+z2)/2;
	  gen tm = (t1+t2)/2;
	  file << "$\",(" << re(zm,&ct) << "," << im(zm,&ct) << "," << tm << "),align=right);" << endl;
	}
      }
    }
  }
  
  for (unsigned l=0; l<faces[j].midvertices.size(); l++) {
    if (faces[j].topface.size()>1) {
      
      gen X1 = Ai*vectevaln(faces[j].bottomface[faces[j].midvertindices[l]].coords);
      gen z1 = X1[2]/X1[0];
      gen t1 = re(X1[1]/X1[0],&ct);
      
      gen X2 = Ai*vectevaln(faces[j].midvertices[l].coords);
      gen z2 = X2[2]/X2[0];
      gen t2 = re(X2[1]/X2[0],&ct);
      
      gen X3 = Ai*vectevaln(faces[j].topface[faces[j].midvertindices[l]].coords);
      gen z3 = X3[2]/X3[0];
      gen t3 = re(X3[1]/X3[0],&ct);
      
      file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<t1<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<t2<<"),p);" << endl;	
      file << "draw(("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<t2<<")--("<<re(z3,&ct)<<","<<im(z3,&ct)<<","<<t3<<"),p);" << endl;	

      vector<int> w = faces[j].sides[faces[j].midvertindices[l]].word;
      if (w.size()<10) {
	file << "label(\"$";
	for (unsigned u=0; u<w.size(); u++) {
	  if (w[u]<0) {
	    file << "\\bar" << -w[u];
	  }
	  else {
	    file << w[u];
	  }
	}
	gen zm = (z2+z3)/2;
	gen tm = (t2+t3)/2;
	file << "$\",(" << re(zm,&ct) << "," << im(zm,&ct) << "," << tm << "),align=right);" << endl;
      } 
    }
    else {

      gen X1 = Ai*vectevaln(faces[j].bottomface[faces[j].midvertindices[l]].coords);
      gen z1 = X1[2]/X1[0];
      gen t1 = re(X1[1]/X1[0],&ct);
      
      gen X2 = Ai*vectevaln(faces[j].midvertices[l].coords);
      gen z2 = X2[2]/X2[0];
      gen t2 = re(X2[1]/X2[0],&ct);
      
      gen X3 = Ai*vectevaln(faces[j].topface[0].coords);
      gen z3 = X3[2]/X3[0];
      gen t3 = re(X3[1]/X3[0],&ct);
      
      file << "draw(("<<re(z1,&ct)<<","<<im(z1,&ct)<<","<<t1<<")--("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<t2<<"),p);" << endl;	
      file << "draw(("<<re(z2,&ct)<<","<<im(z2,&ct)<<","<<t2<<")--("<<re(z3,&ct)<<","<<im(z3,&ct)<<","<<t3<<"),p);" << endl;	

      vector<int> w = faces[j].sides[faces[j].midvertindices[l]].word;
      if (w.size()<10) {
	file << "label(\"$";
	for (unsigned u=0; u<w.size(); u++) {
	  if (w[u]<0) {
	    file << "\\bar" << -w[u];
	  }
	  else {
	    file << w[u];
	  }
	}
	gen zm = (z2+z3)/2;
	gen tm = (t2+t3)/2;
	file << "$\",(" << re(zm,&ct) << "," << im(zm,&ct) << "," << tm << "),align=right);" << endl;
      }
    }
    
  }

  file << "pen dp = red+4;"<< endl;
  for (unsigned k=0; k<faces[j].topface.size(); k++) {
    gen X = Ai*vectevaln(faces[j].topface[k].coords);
    gen z = X[2]/X[0];
    gen t = re(X[1]/X[0],&ct);
    file << "dot(("<<re(z,&ct)<<","<<im(z,&ct)<<","<<t<<"),dp);" << endl;
  }
  for (unsigned k=0; k<faces[j].bottomface.size(); k++) {
    gen X = Ai*vectevaln(faces[j].bottomface[k].coords);
    gen z = X[2]/X[0];
    gen t = re(X[1]/X[0],&ct);
    file << "dot(("<<re(z,&ct)<<","<<im(z,&ct)<<","<<t<<"),dp);" << endl;
  }
  for (unsigned k=0; k<faces[j].midvertices.size(); k++) {
    gen X = Ai*vectevaln(faces[j].midvertices[k].coords);
    gen z = X[2]/X[0];
    gen t = re(X[1]/X[0],&ct);
    file << "dot(("<<re(z,&ct)<<","<<im(z,&ct)<<","<<t<<"),dp);" << endl;
  }


  file.close();

}

void Polytope::drawProjection(unsigned axis) {

  if (axis!=1 && axis!=2) {
    axis = 1;
  }

  ofstream file;
  string ax = static_cast<ostringstream*>( &(ostringstream() << axis) )->str();
  //  string filename = "pic"+tag+"_face"+js+".asy";
  string filename = "pics/pic"+tag+"_"+ax+".asy";
  file.open(filename.c_str());
  file << "import graph;" << endl;
  file << "size(7.5cm,0);" << endl;

  //  cout << "Q: " << Q << endl;
  gen Hn = _subst(makesequence(H,x__IDNT_e,xv),&ct);
  gen Qn = evalf(Q,20,&ct);
  gen D = transpose(conj(Qn,&ct))*Hn*Qn;
  vecteur te;
  te.push_back(onenumber/sqrt(-re(D[0][0],&ct),&ct));
  te.push_back(onenumber/sqrt(-re(D[1][1],&ct),&ct));
  te.push_back(onenumber/sqrt(-re(D[2][2],&ct),&ct));
  gen K = _diag(te,&ct);
  Qn = Qn*K;
  //  cout << "Diag? " << transpose(conj(Qn,&ct))*Hn*Qn << endl;

  Qn = inverse(Qn);

  gen fa = gen("180/pi",&ct);
  for (unsigned k=0; k<edges.size(); k++) {
    //    if (edges[k].faceindices[0]<7 || edges[k].faceindices[1]<7) {
    gen X = _subst(makesequence(vertices[edges[k].origin].coords,x__IDNT_e,xv),&ct);
    gen Y = _subst(makesequence(vertices[edges[k].end].coords,x__IDNT_e,xv),&ct);
    gen te = _subst(makesequence(Inn(vertices[edges[k].origin].coords,vertices[edges[k].end].coords),x__IDNT_e,xv),&ct);
    gen Y1 = -Y*te/abs(te);
    //    cout << "<X,Y1>=" << conj(Y1,&ct)*Hn*X << endl;
    gen X2 = Qn*X;
    gen Y2 = Qn*Y1;

    gen a1 = X2[axis]/X2[0];
    gen a3 = Y2[axis]/Y2[0];
    gen a2 = (X2[axis]+Y2[axis])/(X2[0]+Y2[0]);

    gen b1 = conj(a2-a1,&ct);
    gen b2 = conj(a3-a1,&ct);
    gen l1 = abs(a2)*abs(a2)-abs(a1)*abs(a1);
    gen l2 = abs(a3)*abs(a3)-abs(a1)*abs(a1);

    gen c = (conj(b2,&ct)*l1-conj(b1,&ct)*l2)/(b1*conj(b2,&ct)-conj(b1,&ct)*b2);
    gen r = sqrt(abs(a1)*abs(a1)-2*re(a1*conj(c,&ct),&ct)+abs(c)*abs(c),&ct);
    //    cout << "c=" << c << endl;
    //    cout << "r=" << r << endl;
    
    /*    cout << "test center/rad" << endl;
    cout << abs(a1-c)-r << endl;
    cout << abs(a2-c)-r << endl;
    cout << abs(a3-c)-r << endl;*/
    
    gen u1 = a1-c;
    gen u3 = a3-c;
    gen th1 = arg(u1,&ct)*fa;
    gen th2 = arg(u3,&ct)*fa;
    if (is_greater(th2,th1,&ct)) {
      if (is_greater(th2-th1,180,&ct)) {
	th2 = th2-360;
      }
    }
    else {
      if (is_greater(th1-th2,180,&ct)) {
	th1 = th1-360;
      }
    }
    //    cout << "th1=" << th1 << endl;
    //    cout << "th2=" << th2 << endl;

    file << "draw(arc(("<<re(c,&ct)<<","<<im(c,&ct)<<"),"<<r<<","<<th1<<","<<th2<<"),p);" << endl;
    /*    if (k==1) {
    file << "label(\""<<k<<"\",("<<re(a1,&ct)<<","<<im(a1,&ct)<<"));"<<endl;
    file << "label(\""<<k<<"\",("<<re(a2,&ct)<<","<<im(a2,&ct)<<"));"<<endl;
    file << "label(\""<<k<<"\",("<<re(a3,&ct)<<","<<im(a3,&ct)<<"));"<<endl;
    }*/
    // want to write Z = (1-t)X2+tY2 in ball coords, and project onto one axis, i.e. take Z1/Z0 or Z2/Z0
    /*    file << "pair x20_" << k << "=(" << re(X2[0],&ct) << "," << im(X2[0],&ct) << ");" << endl;
    file << "pair x21_" << k << "=(" << re(X2[1],&ct) << "," << im(X2[1],&ct) << ");" << endl;
    file << "pair y20_" << k << "=(" << re(Y2[0],&ct) << "," << im(Y2[0],&ct) << ");" << endl;
    file << "pair y21_" << k << "=(" << re(Y2[1],&ct) << "," << im(Y2[1],&ct) << ");" << endl;
    file << "real x_arc_" << k << "(real t) {" << endl;
    file << "    pair u = (1-t)*x21_" << k << "+ t*y21_" << k << ";" << endl; 
    file << "    pair v = (1-t)*x20_" << k << "+ t*y20_" << k << ";" << endl; 
    file << "    return (u/v).x;" << endl;
    file << "}" << endl;
    file << "real y_arc_" << k << "(real t) {" << endl;
    file << "    pair u = (1-t)*x21_" << k << "+ t*y21_" << k << ";" << endl; 
    file << "    pair v = (1-t)*x20_" << k << "+ t*y20_" << k << ";" << endl; 
    file << "    return (u/v).y;" << endl;
    file << "}" << endl;*/
   
    //    }
  }
  /*  for (unsigned k=0; k<edges.size(); k++) {
    file << "draw(graph(x_arc_"<<k<<",y_arc_"<<k<<",0,1),p);" << endl;
    }*/
  file.close();
}

bool Polytope::checkSide() {
  vector<int> inds;
  for (int k=0; k<faces.size(); k++) {
    inds.push_back(k);
  }
  while (inds.size()>0) {
    //    cout << "inds: " << inds << endl;
    int k = inds[0];
    bool foundbuddy = false;
    int u = 0;
    while (!foundbuddy && u<inds.size()-1) {
      u++;
      gen X = BoxProd(faces[k].cspine,faces[inds[u]].cspine);
      if (!operator_equal(X,zerovector,&ct)) {
	int sx;
	if (checkSign(SqNorm(X),sx)) {
	  foundbuddy = sx<0;
	  if (foundbuddy) {
	    //	    cout << "Face " << k << " is buddy with face " << inds[u] << endl;
	    faces[k].X0 = X;
	    faces[k].X1 = reflectAcrossSpine(X,faces[k]);
	    faces[inds[u]].X0 = X;
	    faces[inds[u]].X1 = reflectAcrossSpine(X,faces[inds[u]]); 
	  }
	}
      }
    }
    if (!foundbuddy) {
      // may have to use other faces as buddies...
      u=-1;
      while (!foundbuddy && u!=faces.size()-1) {
	u++;
	gen X = BoxProd(faces[k].cspine,faces[u].cspine);
	if (!operator_equal(X,zerovector,&ct)) {
	  int sx;
	  if (checkSign(SqNorm(X),sx)) {
	    foundbuddy = sx<0;
	    if (foundbuddy) {
	      //   cout << "Face " << k << " is buddy with face " << u << endl;
	      faces[k].X0 = X;
	      faces[k].X1 = reflectAcrossSpine(X,faces[k]);
	    }
	  }
	}
      }
      if (!foundbuddy) {
	cerr << "Face #" << k << " is not coequidistant from anyone..." << endl;
	cerr << " this may not be a serious issue, but I will stop now." << endl;
	exit(1);
      }
      else {
	inds.erase(inds.begin());
      }
    }
    else {
      inds.erase(inds.begin()+u);
      if (u!=0) {
	inds.erase(inds.begin());
      }
      else {
	cerr << "Trouble in checkSide()" << endl;
	exit(1);
      }
    }
  }
  gen Hn = evali(H);
  gen p0n = evalf(p0,20,&ct);
  bool res = true;
  for (unsigned k=0; k<faces.size() && res; k++) {
    gen u1 = evali(faces[k].X0);
    gen v1 = evali(faces[k].X1);
    gen te = abs( conj(p0n,&ct)*Hn*u1 ) - abs( conj(p0n,&ct)*Hn*v1 );
    res = is_greater(zeronumber,_right(te,&ct),&ct); 
    if (!res) {
      cout << "Test face " << k << ": " << te << endl;
    }
  }
  return res;

}

void Polytope::checkEdge(unsigned j) {
  
  cout << "Studying edge #" << j << ", whose endpoints are vertices of: " << convertword(edges[j].faceindices) << "... ";

  gen X = vertices[edges[j].origin].coords;
  gen Y = vertices[edges[j].end].coords;
  /*  cout << "Origin is on faces: " << vertices[edges[j].origin].faceindices << endl;
  cout << "Origin is on bisectors: " << findBisectors(X) << endl;
  cout << "Endpt is on faces: " << vertices[edges[j].end].faceindices << endl;
  cout << "Endpt is on bisectors: " << findBisectors(Y) << endl;*/

  gen te = Inn(X,Y);
  Y = -te*Y;
  vectred(Y);
  gen Z = (1-t1)*X+t1*Y;
  vectred(Z);
  vector<int> inds;
  for (unsigned k=0; k<faces.size(); k++) {
    if (find(edges[j].faceindices.begin(),edges[j].faceindices.end(),k)==edges[j].faceindices.end()) {
      inds.push_back(k);
    }
  }

  for (unsigned k=0; k<inds.size(); k++) {
    gen eq = simplify(Inn(Z,faces[inds[k]].X0)*Inn(faces[inds[k]].X0,Z) - Inn(Z,faces[inds[k]].X1)*Inn(faces[inds[k]].X1,Z),&ct);
    red(eq);
    eq = convertToReal(eq);
    //    cout << "Equation #" << inds[k] << ": " << eq << endl;
    if (operator_equal(eq,zeronumber,&ct)) {
      cout << "Equation is 0, edge is on a bisector of which it is not a facet (maybe because of reflections of order 2)" << endl;
      //      exit(1);
    }
    else {

      gen eq0 = _coeff(makesequence(eq,t1,0),&ct);
      gen eq1 = simplify(_subst(makesequence(eq,t1,onenumber),&ct),&ct);
      simplify(eq1,&ct);
      red(eq1);

      //    cout << "Values at endpoints: " << evali(eq0) << " ,  " << evali(eq1) << endl;

      int sgn0, sgn1;
      if (!checkSign(eq0,sgn0)){
	cerr << "Problem checking sign of equation at origin, in edge check" << endl;
	exit(1);      
      }
      if (!checkSign(eq1,sgn1)){
	cerr << "Problem checking sign of equation at endpt, in edge check" << endl;
	exit(1);      
      }
      if (sgn0>0 || sgn1>0) {
	cerr << "Edge is on WRONG side! (found this at an endpoint)" << endl;
	exit(1);      
      }
      
      if (sgn0==0){
	//      cout << "t=0 is a root" << endl;
	if (sgn1==0){
	  cerr << "Both t=0 and t=1 are roots, this should mean edge is in the bisector!" << endl;
	  exit(1);
	}
	else {
	  gen a = _coeff(makesequence(eq,t1,2),&ct);
	  if (operator_equal(a,zeronumber,&ct)) {
	    //	  cout << "Only t=0 gives a root" << endl;
	    //	  cout << "Edge is on negative side" << endl;
	    //	  return true;
	  }
	  else {
	    // check if te=-b/a is in [0,1]
	    gen te = -_coeff(makesequence(eq,t1,1),&ct)*inverse(a);
	    red(te);
	    //	  cout << "te=" << te << endl;
	    //	  cout << "Need to check if " << evali(te) << " is in [0,1]..." << endl;
	    int sc;
	    if (!checkSign(te,sc)) {
	      cerr << "Problem checking sign of t-value in edge parameter" << endl;
	      exit(1);
	    }
	    else {
	      if (sc>0) {
		if (!checkSign(te-1,sc)) {
		  cerr << "Problem checking whether t<=1 in edge parameter" << endl;
		  exit(1);
		}
		else {
		  if (sc<0) {
		    cerr << "The 1-skeleton is not embedded!" << endl;
		    exit(1);
		  }
		  /*	else {
			return true;
			}*/
		}
	      }
	      else {
		//  cout << "Only t=0 gives a root in [0,1]" << endl;	      
		//  cout << "Edge is on negative side" << endl;
		//	      return true;
	      }
	    }
	  }	
	}
      }
      else {
	// sgn0 != 0
	if (sgn1==0){
	  //	cout << "t=1 is a root" << endl;
	  gen a = _coeff(makesequence(eq,t1,2),&ct);
	  if (operator_equal(a,zeronumber,&ct)) {
	    //	  cout << "Only t=1 gives a root" << endl;
	    //	  return true;
	  }
	  else {
	    // check if te=-1-b/a is in [0,1]
	    gen te = -1-_coeff(makesequence(eq,t1,1),&ct)*inverse(a);
	    red(te);
	    //	  cout << "Need to check if " << evali(te) << " is in [0,1]..." << endl;
	    int sc;
	    if (!checkSign(te,sc)) {
	      cerr << "Problem checking sign of t-value in edge parameter" << endl;
	      exit(1);
	    }
	    else {
	      if (sc>0) {
		if (!checkSign(te-1,sc)) {
		  cerr << "Problem checking whether t<=1 in edge parameter" << endl;
		  exit(1);
		}
		else {
		  if (sc<0) {
		    cerr << "The 1-skeleton is not embedded!" << endl;
		    exit(1);
		  }
		  else {
		    //		  cout << "Only t=1 gives a root in [0,1]" << endl;	      
		    //		  cout << "Edge is on negative side" << endl;
		    //		  return true;
		  }
		}
	      }
	      else {
		//	      cout << "Only t=1 gives a root in [0,1]" << endl;	      
		//	      cout << "Edge is on negative side" << endl;
		//	      return true;
	      }
	    }
	  }	
	}
	else {
	  // no endpoint is on bisector, need to work harder?
	  // no, will simply evaluate value at maximum?  //2at+b -> -b/2a
	  gen a = _coeff(makesequence(eq,t1,2),&ct);
	  int sga;
	  if (!checkSign(a,sga)) {
	    cerr << "Problem checking sign of a, in edge check" << endl;
	    exit(1);      
	  }
	  if (sga==0) {
	    //	  cout << "Linear equation, enough to check endpoints!" << endl;	      
	    //	  cout << "Edge is on negative side" << endl;
	    //	  return true;
	  }
	  else {
	    if (sga>0) {
	      //	    cout << "Concavity is up, checking endpoints is enough..." << endl;	    
	      //	    cout << "Edge is on negative side" << endl;
	      //	    return true;
	    }
	    else {
	      gen tm = -_coeff(makesequence(eq,t1,1),&ct)*inverse(2*a);
	      red(tm);
	      
	      //	    cout << "t-max: " << evali(tm) << endl;
	      
	      // check if max is in [0,1], and if so, check that the max value is negative
	      int sg;
	      if (!checkSign(tm,sg)) {
		cerr << "Problem checking sign of t-value of max of parabola, in edge check" << endl;
		exit(1);      	      
	      }
	      else {
		if (sg>0) {
		  // tmax is > 0
		  if (!checkSign(tm-1,sg)) {
		    cerr << "Problem checking t-value of max of parabola, in edge check" << endl;
		    exit(1);      	      
		  }
		  else {
		    if (sg<0) {
		      // t-max is between 0 and 1...
		      gen yo = _subst(makesequence(eq,t1,tm),&ct);
		      if (!checkSign(yo,sg)) {
			cerr << "Problem checking sign of max value, in edge check" << endl;
			exit(1);      	      		      
		      }
		      else {
			if (sg<=0) {
			  //		cout << "Max of parabola occurs in [0,1], but value is negative." << endl;	    
			  //		cout << "Edge is on negative side" << endl;
			  //	    return true;
			}
			else {
			  cerr << "Edge is on WRONG side! (found this at an endpoint)" << endl;
			  exit(1);      
			}
		      }
		    }
		    else {
		      // tmax is >1
		      //		    cout << "t-max is > 1" << endl;	    
		      //		    cout << "Edge is on negative side" << endl;
		    }
		  }
		}
		else {
		  // tmax is < 0
		  //		cout << "t-max is < 0" << endl;	    
		  //		cout << "Edge is on negative side" << endl;
		}
	      }	    
	    }	  
	  }
	}
      }
    }
  }
  cout << " check!" << endl;
}

void Polytope::checkRidge(unsigned i) {

  int j =  ridges[i].faceindices[0];
  int k =  ridges[i].faceindices[1];
  cout << endl;
  cout << "Checking ridge on faces #" << j << " and " << k << endl;

  bool genparam = true;
  gen X0 = BoxProd(faces[j].cspine,faces[k].cspine);
  gen V0, W0, Z0; // Orthogonal (but not orthonormal) basis, with V0, W0 spanning common slice
  if (operator_equal(X0,zerovector,&ct)) {
    // cospinal
    genparam = false;
    gen Xte = BoxProd(faces[j].mirror.coords,faces[k].mirror.coords);
    if (operator_equal(Xte,zerovector,&ct)) {
      cout << "Same mirror, this gives the common slice" << endl;
      V0 = faces[j].apexProjection;
      W0 = Proj(faces[j].bottomface[0].coords,V0);	
      Z0 = BoxProd(V0,W0); // not clear if we need this
    }
    else {
      Xte = BoxProd(faces[j].mirror.coords,faces[k].apex.coords);
      if (operator_equal(Xte,zerovector,&ct)) {
	cout << "Mirror of face#" << j << " is same as top face of #" << k << endl;
	cout << "     this gives the common slice" << endl;	
	V0 = faces[j].apexProjection;
	W0 = Proj(faces[j].bottomface[0].coords,V0);
	Z0 = BoxProd(V0,W0); // not clear if we need this
      }
      else {
	Xte = BoxProd(faces[k].mirror.coords,faces[j].apex.coords);
	if (operator_equal(Xte,zerovector,&ct)) {
	  cout << "Mirror of face#" << k << " is same as top face of #" << j << endl;
	  cout << "     this gives the common slice" << endl;	
	  V0 = faces[k].apexProjection;
	  W0 = Proj(faces[k].bottomface[0].coords,V0);
	  Z0 = BoxProd(V0,W0); // not clear if we need this
	}
	else {
	  cerr << "There is probably no common slice, will need to check this.." << endl;	  
	  //	  exit(1);
	}
      }
    }
  }
  else {
    // complex spines intersect in projective space, need to check if they intersect on the real spines
    if (isOnPrism(X0,faces[j])) {
      cout << "Extended real spine of face #" << j << " intersects extor #" << k << endl;
      if (isOnPrism(X0,faces[k])) {
	cout << "Extended real spine of face #" << k << " intersects extor #" << j << endl;
	cout << "Real spines intersect... "; //either cotranchal, or linear?
	int sx;
	if (!checkSign(SqNorm(X0),sx)) {
	  cerr << "Trouble computing sign of square norm (in checkRidge)" << endl;
	  exit(1);
	}
	else {
	  if (sx>0) {
	    cout << "Cotranchal pair" << endl;

	    gen Y1 = faces[j].apexProjection;
	    Y1 = Inn(X0,Y1)*Y1;
	    red(Y1);
	    Y1 = Proj(Y1,X0);
	    gen Y2 = faces[k].apexProjection;
	    Y2 = Inn(X0,Y2)*Y2;
	    red(Y2);
	    Y2 = Proj(Y2,X0);
	    gen tee = Inn(Y1,faces[j].X0)*Inn(faces[j].X0,Y1)-Inn(Y1,faces[j].X1)*Inn(faces[j].X1,Y1);
	    red(tee);
	    tee = Inn(Y2,faces[k].X0)*Inn(faces[k].X0,Y2)-Inn(Y2,faces[k].X1)*Inn(faces[k].X1,Y2);
	    red(tee);
	    gen tt = (Inn(Y1,Y2)+Inn(Y2,Y1))/2;
	    gen te = SqNorm(Y1)*SqNorm(Y2) - tt*tt;
	    red(te);
	    te = convertToReal(te);
	    int sgn;
	    if (!checkSign(te,sgn)) {
	      cerr << "Trouble checking sign (is intersection is slice only)" << endl;
	      exit(1);
	    }
	    else {
	      if (sgn<0) {
		cerr << "Intersection is not just the common slice!?" << endl;
		exit(1);
	      }
	      else {
		cout << "intersection is slice only!" << endl;
		genparam = false;
	      }
	    }
	      
	    V0 = X0;
	    if (areDep(X0,faces[j].apex.coords)) {
	      cout << "Common slice is top face of face #" << j << endl;
	      W0 = Proj(faces[j].topface[0].coords,V0);
	    }
	    else {
	      if (areDep(X0,faces[j].mirror.coords)) {
		cout << "Common slice is mirror of face #" << j << endl;
		W0 = Proj(faces[j].bottomface[0].coords,V0);
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	    Z0 = BoxProd(V0,W0); // not clear if we need this
	    if (areDep(X0,faces[k].apex.coords)) {
	      cout << "Common slice is top face of face #" << k << endl;
	    }
	    else {
	      if (areDep(X0,faces[k].mirror.coords)) {
		cout << "Common slice is mirror of face #" << k << endl;
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	  }
	  else {
	    if (sx==0) {
	      cerr << "Strange, real spines intersect at infinity..." << endl;
	      exit(1);
	    }
	    else {
	      cerr << "Real spines intersect inside, linear intersection (not implemented yet)" << endl;
	      exit(1);
	    }
	  }
	}
      }	
      else {
	cout << "Real spines don't intersect" << endl;
	genparam = true;
      }
    }
    else {
      cout << "Real spines don't intersect" << endl;
      genparam = true;
    }
  }

  if (genparam) {
    gen vj, wj, vk, wk, vl, wl;

    wj = faces[j].apexProjection;
    vj = faces[j].apex.coords;
    vj = Inn(wj,vj)*vj;
    vectred(vj);
    vj = Proj(vj,wj);

    wk = faces[k].apexProjection;
    vk = faces[k].apex.coords;
    vk = Inn(wk,vk)*vk;
    vectred(vk);
    vk = Proj(vk,wk);

    // need vl and wl, for third bisector in Giraud's theorem!

    cout << "Will construct third bisector (Giraud)" << endl;

    bool fd = false;
    int u1=-1;
    while (!fd && u1!=faces[j].sides.size()-1) {
      u1++;
      fd = areDep(faces[j].sides[u1].coords,faces[k].mirror.coords);
    }
    if (!fd) {
      cerr << "Trouble findind third bisector (mirror)" << endl;
      exit(1);
    }

    int u2=-1;
    int v;
    fd = false;
    while (!fd && u2!=faces[j].sides.size()-1) {
      u2++;
      v=-1;
      while (!fd && v!=faces[k].sides.size()-1) {
	v++;
	fd = areDep(faces[j].sides[u2].coords,faces[k].sides[v].coords);
      }
    }
    if (!fd) {
      cerr << "Trouble findind third bisector (side)" << endl;
      exit(1);
    }

    // mirror and apex of third bisector
    gen mi = faces[j].sides[u2].coords;
    gen ap = BoxProd(faces[j].sides[u1].coords,faces[j].mirror.coords);
    gen app = Proj(ap,mi);
    
    wl = app;
    vl = Inn(app,ap)*ap;
    vectred(vl);
    vl = Proj(vl,wl);

    gen t1min, t1max, t2min, t2max, t3min, t3max;
    for (unsigned z=0; z<ridges[i].verts.size(); z++) {
      //      cout << "Testing vertex #" << ridges[i].verts[z] << endl;
      gen t1v = -Inn(vj,vertices[ridges[i].verts[z]].coords)*inverse(Inn(wj,vertices[ridges[i].verts[z]].coords));
      gen t2v = -Inn(vk,vertices[ridges[i].verts[z]].coords)*inverse(Inn(wk,vertices[ridges[i].verts[z]].coords));
      gen t3v = -Inn(vl,vertices[ridges[i].verts[z]].coords)*inverse(Inn(wl,vertices[ridges[i].verts[z]].coords));
      red(t1v);
      red(t2v);
      red(t3v);
      t1v = convertToReal(t1v);
      t2v = convertToReal(t2v);
      t3v = convertToReal(t3v);
      if (z==0) {
	t1min = t1v;
	t1max = t1v;
	t2min = t2v;
	t2max = t2v;
	t3min = t3v;
	t3max = t3v;
      }
      else {
	int sg;
	if (!checkSign(t1v-t1min,sg)) {
	  cerr << "Problem checking sign (t1min)" << endl;
	}
	else {
	  if (sg<0) {
	    t1min = t1v;
	  }
	}
	if (!checkSign(t1v-t1max,sg)) {
	  cerr << "Problem checking sign (t1max)" << endl;
	}
	else {
	  if (sg>0) {
	    t1max = t1v;
	  }
	}

	if (!checkSign(t2v-t2min,sg)) {
	  cerr << "Problem checking sign (t2min)" << endl;
	}
	else {
	  if (sg<0) {
	    t2min = t2v;
	  }
	}
	if (!checkSign(t2v-t2max,sg)) {
	  cerr << "Problem checking sign (t2max)" << endl;
	}
	else {
	  if (sg>0) {
	    t2max = t2v;
	  }
	}

	if (!checkSign(t3v-t3min,sg)) {
	  cerr << "Problem checking sign (t3min)" << endl;
	}
	else {
	  if (sg<0) {
	    t3min = t3v;
	  }
	}
	if (!checkSign(t3v-t3max,sg)) {
	  cerr << "Problem checking sign (t3max)" << endl;
	}
	else {
	  if (sg>0) {
	    t3max = t3v;
	  }
	}
      }
    }
    /*    gen t1mini = evali(t1min);
    gen t1maxi = evali(t1max);
    gen t2mini = evali(t2min);
    gen t2maxi = evali(t2max);
    gen t3mini = evali(t3min);
    gen t3maxi = evali(t3max);
    // verbose?
    cout << "t1 range: " << t1mini << " , " << t1maxi << endl;
    cout << "t2 range: " << t2mini << " , " << t2maxi << endl;
    cout << "t3 range: " << t3mini << " , " << t3maxi << endl;*/
    
    // take (vj+t1*wj)x(vk+t2*wk)
    gen v0, v1, v2, v3; 
    v0 = BoxProd(vj,vk);
    v1 = BoxProd(wj,vk);
    v2 = BoxProd(vj,wk);
    v3 = BoxProd(wj,wk);
    gen vt = v0 + t1*v1 + t2*v2 + t1*t2*v3;
    //    gen te = SqNorm(vt);
    //    cout << "Eq bdy: " << te << endl;

    gen t3n = -Inn(vt,vl);
    gen t3d = Inn(vt,wl);

    for (unsigned l=0; l<faces.size(); l++) {
      gen eq = Inn(vt,faces[l].X0)*Inn(faces[l].X0,vt) - Inn(vt,faces[l].X1)*Inn(faces[l].X1,vt); 
      red(eq);
      eq = convertToReal(eq);
      
      if (operator_equal(eq,zeronumber,&ct)) {
	cout << "Eq #" << l << " is 0..." << endl;
      }
      else {
	//	cout << "Eq #" << l << ": " << eq << endl; 
	gen eq1 = _derive(makesequence(eq,t1),&ct);
	gen eq2 = _derive(makesequence(eq,t2),&ct);
	
	vecteur eqs_rur;
	eqs_rur.push_back(eq1);
	eqs_rur.push_back(eq2);
	eqs_rur.push_back(realminpol);
	
	//	cout << "Starting computation of RUR..." << endl;

	gen gb1 = _gbasis(makesequence(eqs_rur,vars2d,change_subtype(_RUR_REVLEX,_INT_GROEBNER)),&ct);

	//	cout << "gb1=" << gb1 << endl;

	if (gb1[0]==-4) {

	  vecteur valpars;
	  if (!safeSift(gb1,valpars)) {
	    cerr << "Problem sifting RUR values... (crit pts)" << endl;
	    exit(1);
	  }
	  else {
	    cout << "For eq#" << l << ", found " << valpars.size() << " critical point(s).. " << endl; 
	    
	    for (unsigned u = 0; u<valpars.size(); u++) {
	      
	      cout << "Checking crit pt #" << u << "..  ";
	      
	      gen t1vi = _horner(makesequence(gb1[5],valpars[u],s__IDNT_e),&ct)/_horner(makesequence(gb1[3],valpars[u],s__IDNT_e),&ct);
              gen t2vi = _horner(makesequence(gb1[6],valpars[u],s__IDNT_e),&ct)/_horner(makesequence(gb1[3],valpars[u],s__IDNT_e),&ct);
	      // verbose??
	      //	      cout << t1vi << " , " << t2vi << endl;
	      	      

	      int stest;
	      if (!checkt1val(gb1,valpars[u],t1min,stest)) {
		cerr << "Trouble comparing with t1min" << endl;
		exit(1);
	      }
	      if (stest<0) {
		cout << "outside (t1<t1min)" << endl;
	      }
	      else {
		if (!checkt1val(gb1,valpars[u],t1max,stest)) {
		  cerr << "Trouble comparing with t1max" << endl;
		  exit(1);
		}
		if (stest>0) {
		  cout << "outside (t1>t1max)" << endl;
		}
		else {
		  if (!checkt2val(gb1,valpars[u],t2min,stest)) {
		    cerr << "Trouble comparing with t2min" << endl;
		    exit(1);
		  }
		  if (stest<0) {
		    cout << "outside (t2<t2min)" << endl;
		  }
		  else {
		    if (!checkt2val(gb1,valpars[u],t2max,stest)) {
		      cerr << "Trouble comparing with t2max" << endl;
		      exit(1);
		    }
		    if (stest>0) {
		      cout << "outside (t2>t2max)" << endl;
		    }
		    else {
		      if (!checkt3val(gb1,valpars[u],t3min,vl,wl,vt,stest)) {
			cerr << "Trouble comparing with t3min" << endl;
			exit(1);
		      }
		      if (stest<0) {
			cout << "outside (t3<t3min)" << endl;
		      }
		      else {
			if (!checkt3val(gb1,valpars[u],t3max,vl,wl,vt,stest)) {
			  cerr << "Trouble comparing with t3max" << endl;
			  exit(1);
			}
			if (stest>0) {
			  cout << "outside (t3>t3max)" << endl;
			}
			else {
			  if (!checkEqSign(gb1,valpars[u],eq,stest)) {
			    cerr << "Trouble checking sign of equation" << endl;
			    exit(1);
			  }
			  if (eq>0) {
			    cerr << "Equation is positive inside the G-polygon..." << endl;
			    cerr << "There seems to be an isolated component!!!!!!!" << endl;
			    exit(1);
			  }
			  else {
			    cout << "CRITICAL POINT IS INSIDE, but the equation is negative there!" << endl;
			  }
			}
		      }
		    }
		  }
		}
	      }
	    }	      
	  }
	}
	else {
	  cout << "Non isolated critical points? the system is not 0-dim" << endl;
	  // could either check if eqn for geodesics give (multiple) factors, or
	  //    use factorization over a number field, or factorization by hand... 
	  // For now, I choose the last option.

	  bool needmore = false;

	  gen co = _coeff(makesequence(eq,t2),&ct);

	  gen a = _coeff(makesequence(co[0],t1,2),&ct);
	  gen b = _coeff(makesequence(co[1],t1,2),&ct);
	  gen c = _coeff(makesequence(co[2],t1,2),&ct);

	  gen te1,te2;
	  te1 = _rem(makesequence(a*co[1]-b*co[0],realminpol,s__IDNT_e),&ct);
	  if (operator_equal(te1,zeronumber,&ct)) {
	    te2 = _rem(makesequence(a*co[2]-c*co[0],realminpol,s__IDNT_e),&ct);
	    if (operator_equal(te2,zeronumber,&ct)) {
	      gen discr = _rem(makesequence(b*b-4*a*c,realminpol,s__IDNT_e),&ct);
	      if (operator_equal(discr,zeronumber,&ct)) {
		cout << "There is a square factor of the form (t2-r)^2" << endl;
		gen r = -b*inverse(2*a);
		red(r);
		cout << "  r = " << r << endl;
		int sg;
		gen zz = r-t2min;
		red(zz);
		//		cout << "r-t2min=" << zz << endl;
		if (checkSign(zz,sg)){
		  if (sg>0) {
		    cout << "r>t2min" << endl;
		    zz = r-t2max;
		    red(zz);
		    //		    cout << "r-t2max=" << zz << endl;
		    if (checkSign(zz,sg)){
		      if (sg<0) {
			cout << "r<t2max" << endl;
			cout << "Horizontal line crosses the interior of the ridge!" << endl;
			exit(1);
		      }
		    }
		    else {
		      cerr << "Trouble comparing r with t2max" << endl;
		      exit(1);
		    }
		  }
		}
		else {
		  cerr << "Trouble comparing r with t2min" << endl;
		  exit(1);
		}

		// need to check sign of co[0], check it is positive in [t1min,t1max]!
		gen yo1 = _subst(makesequence(co[0],t1,t1min),&ct);
		red(yo1);
		//		cout << "yo1=" << yo1 << endl;
		if (checkSign(yo1,sg)) {
		  if (sg>0) {
		    cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
		    exit(1);
		  }
		  else {
		    cout << "g(t1)==eq/(t2-r)^2 is <=0 at t1min" << endl;
		    gen yo2 = _subst(makesequence(co[0],t1,t1max),&ct);
		    red(yo2);
		    //		    cout << "yo2=" << yo2 << endl;
		    if (checkSign(yo2,sg)) {
		      if (sg>0) {
			cout << "g(t1)==eq/(t2-r)^2 is <=0 at t1max" << endl;
			cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
			exit(1);
		      }
		      else {
			// co[0] negative at enpoints, if degree two then still something to do!
			cout << "co[0] is neg at endpoints, but need to check more" << endl;
			cout << "co[0]=" << co[0] << endl;
			gen de = _degree(makesequence(co[0],t1),&ct);
			cout << "degree: " << de << endl;
			if (operator_equal(de,twonumber,&ct)) {
			  gen aa = _coeff(makesequence(co[0],t1,2),&ct);
			  red(aa);
			  //			  cout << "aa=" << aa << endl;
			  if (checkSign(aa,sg)) {
			    if (sg<0) {
			      cout << "coeff of t1^2 is <0, will check tv = t1 of vertex of parabola" << endl;
			      gen tmax = _coeff(makesequence(co[0],t1,1),&ct)*inverse(2*_coeff(makesequence(co[0],t1,2),&ct));
			      red(tmax);
			      gen zzz = tmax-t1min;
			      red(zzz);
			      //			      cout << "tver-t1min=" << zz << endl;
			      if (checkSign(zzz,sg)) {
				if (sg>0) {
				  cout << "tv>t1min" << endl;
				  zzz = tmax-t1max;
				  red(zzz);
				  //				  cout << "tver-t1max=" << zz << endl;
				  if (checkSign(zzz,sg)) {
				    if (sg<0) {
				      cout << "tv<t1max" << endl;
				      gen val = _subst(makesequence(co[0],t1,tmax),&ct);
				      red(val);
				      cout << "val at vertex of parabola: " << val << endl;
				      if (checkSign(val,sg)) {
					if (sg>0) {
					  cout << "Value at tv is > 0!!!" << endl;
					  cout << "Polytope does not seem embedded, sorry!" << endl;
					  exit(1);
					}
					else {
					  cout << "Value at tv is <= 0, OK." << endl;
					}
				      }
				      else {
					cerr << "Trouble checking sign at vertex of parabola." << endl;
					exit(1);
				      }
				    }
				  }
				  else {
				    cerr << "Trouble comparing t-vertex with t1max." << endl;
				    exit(1);
				  }
				}
			      }
			      else {
				cerr << "Trouble comparing t-vertex with t1min." << endl;
				exit(1);
			      }
			    }
			  }
			  else {
			    cerr << "Trouble checking sign of coeff." << endl;
			    exit(1);
			  }
			}
		      }
		    }
		    else {
		      cerr << "Trouble checking sign of coeff at t1min" << endl;
		      exit(1);		  
		    }
		  }
		}
		else {
		  cerr << "Trouble checking sign of coeff at t1min" << endl;
		  exit(1);		  
		}
	      }
	      else {
		needmore = true;
	      }
	    }
	    else {
	      needmore = true;
	    }
	  }
	  else {
	    needmore = true;
	  }

	  if (needmore) {
	    // start over with other variable!

	    bool needevenmore = false;

	    gen co = _coeff(makesequence(eq,t1),&ct);

	    gen a = _coeff(makesequence(co[0],t2,2),&ct);
	    gen b = _coeff(makesequence(co[1],t2,2),&ct);
	    gen c = _coeff(makesequence(co[2],t2,2),&ct);

	    gen te1,te2;
	    te1 = _rem(makesequence(a*co[1]-b*co[0],realminpol,s__IDNT_e),&ct);
	    if (operator_equal(te1,zeronumber,&ct)) {
	      te2 = _rem(makesequence(a*co[2]-c*co[0],realminpol,s__IDNT_e),&ct);
	      if (operator_equal(te2,zeronumber,&ct)) {
		gen discr = _rem(makesequence(b*b-4*a*c,realminpol,s__IDNT_e),&ct);
		if (operator_equal(discr,zeronumber,&ct)) {
		  cout << "There is a square factor of the form (t1-r)^2" << endl;
		  gen r = -b*inverse(2*a);
		  red(r);
		  cout << "  r = " << r << endl;
		  int sg;
		  gen zzzz = r-t1min;
		  red(zzzz);
		  //		  cout << "r-t1min=" << zzzz << endl;
		  if (checkSign(zzzz,sg)){
		    if (sg>0) {
		      zzzz = r-t1max;
		      red(zzzz);
		      //		      cout << "r-t1max=" << zzzz << endl;
		      if (checkSign(zzzz,sg)){
			if (sg<0) {
			  cout << "Vertical line crosses the interior of the ridge!" << endl;
			  exit(1);
			}
		      }
		      else {
			cerr << "Trouble comparing r with t1max" << endl;
			exit(1);
		      }
		    }
		  }
		  else {
		    cerr << "Trouble comparing r with t1min" << endl;
		    exit(1);
		  }

		  // need to check sign of co[0], check it is positive in [t1min,t1max]!
		  gen yo1 = _subst(makesequence(co[0],t2,t2min),&ct);
		  red(yo1);
		  //	  cout << "yo1=" << yo1 << endl;
		  if (checkSign(yo1,sg)) {
		    if (sg>0) {
		      cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
		      exit(1);
		    }
		    else {
		      gen yo2 = _subst(makesequence(co[0],t2,t2max),&ct);
		      red(yo2);
		      //     cout << "yo2=" << yo2 << endl;
		      if (checkSign(yo2,sg)) {
			if (sg>0) {
			  cerr << "Equation seems to cut into the ridge, have you checked edges?" << endl;
			  exit(1);
			}
			else {
			  // co[0] negative at enpoints, if degree two then still something to do!
			  //	  cout << "co[0]=" << co[0] << endl;
			  gen de = _degree(makesequence(co[0],t2),&ct);
			  //	  cout << "degree: " << de <<  endl;
			  if (operator_equal(de,twonumber,&ct)) {
			    gen aa = _coeff(makesequence(co[0],t2,2),&ct);
			    red(aa);
			    //    cout << "aa=" << aa << endl;
			    if (checkSign(aa,sg)) {
			      if (sg<0) {
				gen tmax = _coeff(makesequence(co[0],t2,1),&ct)*inverse(2*_coeff(makesequence(co[0],t2,2),&ct));
				red(tmax);
				//		cout << "t-vert=" << tmax << endl;
				gen zzzzz = tmax-t2min;
				red(zzzzz);
				//				cout << "tmax-t2min=" << zzzzz << endl;
				if (checkSign(tmax-t2min,sg)) {
				  if (sg>0) {
				    zzzzz = tmax-t2max;
				    red(zzzzz);
				    //	    cout << "tmax-t2max=" << zzzzz << endl;				    
				    if (checkSign(tmax-t2max,sg)) {
				      if (sg<0) {
					gen val = _subst(makesequence(co[0],t2,tmax),&ct);
					red(val);
					cout << "Val at vert: " << val << endl;
					if (checkSign(val,sg)) {
					  if (sg>0) {
					    cout << "Polytope does not seem embedded, sorry!" << endl;
					    exit(1);
					  }
					}
					else {
					  cerr << "Trouble checking sign at vertex of parabola." << endl;
					  exit(1);
					}
				      }
				    }
				    else {
				      cerr << "Trouble comparing t-vertex with t1max." << endl;
				      exit(1);
				    }
				  }
				}
				else {
				  cerr << "Trouble comparing t-vertex with t1min." << endl;
				  exit(1);
				}
			      }
			    }
			    else {
			      cerr << "Trouble checking sign of coeff." << endl;
			      exit(1);
			    }
			  }
			}
		      }
		      else {
			cerr << "Trouble checking sign of coeff at t1min" << endl;
			exit(1);		  
		      }
		    }
		  }
		  else {
		    cerr << "Trouble checking sign of coeff at t1min" << endl;
		    exit(1);		  
		  }
		}
		else {
		  needevenmore = true;
		}
	      }
	      else {
		needevenmore = true;
	      }
	    }
	    else {
	      needevenmore = true;
	    }

	    if (needevenmore) {
	      cerr << "Factors (t1-r)^2 or (t2-r)^2 did not explain not 0-dim" << endl;
	      cerr << "It may help to factor over number field, or use t3?" << endl;
	      exit(1);
	    }
	  }
	}
      }
    }
    
  }
  else {
    // use V0, W0, Z0
    cout << "For this ridge there is nothing to do, it is enough to check the 1-skeleton!" << endl;
  }
}

void Polytope::drawRidgePic(unsigned i) {
  int j =  ridges[i].faceindices[0];
  int k =  ridges[i].faceindices[1];
  cout << "Trying to draw pic of ridge on faces #" << j << " and " << k << endl;
  cout << " (I will only do it provided this ridge is a complex disk)" << endl;

  bool genparam;
  gen X0 = BoxProd(faces[j].cspine,faces[k].cspine);
  gen V0, W0, Z0; // Orthogonal (but not orthonormal) basis, with V0, W0 spanning common slice
  if (operator_equal(X0,zerovector,&ct)) {
    // cospinal
    genparam = false;
    gen Xte = BoxProd(faces[j].mirror.coords,faces[k].mirror.coords);
    if (operator_equal(Xte,zerovector,&ct)) {
      cout << "Same mirror, this gives the common slice" << endl;
      V0 = faces[j].apexProjection;
      W0 = Proj(faces[j].bottomface[0].coords,V0);	
      Z0 = BoxProd(V0,W0); // not clear if we need this
    }
    else {
      Xte = BoxProd(faces[j].mirror.coords,faces[k].apex.coords);
      if (operator_equal(Xte,zerovector,&ct)) {
	cout << "Mirror of face#" << j << " is same as top face of #" << k << endl;
	cout << "     this gives the common slice" << endl;	
	V0 = faces[j].apexProjection;
	W0 = Proj(faces[j].bottomface[0].coords,V0);
	Z0 = BoxProd(V0,W0); // not clear if we need this
      }
      else {
	Xte = BoxProd(faces[k].mirror.coords,faces[j].apex.coords);
	if (operator_equal(Xte,zerovector,&ct)) {
	  cout << "Mirror of face#" << k << " is same as top face of #" << j << endl;
	  cout << "     this gives the common slice" << endl;	
	  V0 = faces[k].apexProjection;
	  W0 = Proj(faces[k].bottomface[0].coords,V0);
	  Z0 = BoxProd(V0,W0); // not clear if we need this
	}
	else {
	  cerr << "There is probably no common slice, will need to check this.." << endl;
	  exit(1);
	}
      }
    }
  }
  else {
    // complex spines intersect in projective space, need to check if they intersect on the real spines

    if (isOnPrism(X0,faces[j])) {
      cout << "Extended real spine of face #" << j << " intersects extor #" << k << endl;
      if (isOnPrism(X0,faces[k])) {
	cout << "Extended real spine of face #" << k << " intersects extor #" << j << endl;
	cout << "Real spines intersect... "; //either cotranchal, or linear?
	int sx;
	if (!checkSign(SqNorm(X0),sx)) {
	  cerr << "Trouble computing sign of square norm (in checkRidge)" << endl;
	  exit(1);
	}
	else {
	  if (sx>0) {
	    cout << "Cotranchal pair" << endl;

	    gen Y1 = faces[j].apexProjection;
	    Y1 = Inn(X0,Y1)*Y1;
	    red(Y1);
	    Y1 = Proj(Y1,X0);
	    gen Y2 = faces[k].apexProjection;
	    Y2 = Inn(X0,Y2)*Y2;
	    red(Y2);
	    Y2 = Proj(Y2,X0);
	    gen tee = Inn(Y1,faces[j].X0)*Inn(faces[j].X0,Y1)-Inn(Y1,faces[j].X1)*Inn(faces[j].X1,Y1);
	    red(tee);
	    tee = Inn(Y2,faces[k].X0)*Inn(faces[k].X0,Y2)-Inn(Y2,faces[k].X1)*Inn(faces[k].X1,Y2);
	    red(tee);
	    gen tt = (Inn(Y1,Y2)+Inn(Y2,Y1))/2;
	    gen te = SqNorm(Y1)*SqNorm(Y2) - tt*tt;
	    red(te);
	    te = convertToReal(te);
	    int sgn;
	    if (!checkSign(te,sgn)) {
	      cerr << "Trouble checking sign (is intersection is slice only)" << endl;
	      exit(1);
	    }
	    else {
	      if (sgn<0) {
		cerr << "Intersection is not just the common slice!?" << endl;
		exit(1);
	      }
	      else {
		cout << "intersection is slice only!" << endl;
		genparam = false;
	      }
	    }
	      
	    V0 = X0;
	    if (areDep(X0,faces[j].apex.coords)) {
	      cout << "Common slice is top face of face #" << j << endl;
	      W0 = Proj(faces[j].topface[0].coords,V0);
	    }
	    else {
	      if (areDep(X0,faces[j].mirror.coords)) {
		cout << "Common slice is mirror of face #" << j << endl;
		W0 = Proj(faces[j].bottomface[0].coords,V0);
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	    Z0 = BoxProd(V0,W0); // not clear if we need this
	    if (areDep(X0,faces[k].apex.coords)) {
	      cout << "Common slice is top face of face #" << k << endl;
	    }
	    else {
	      if (areDep(X0,faces[k].mirror.coords)) {
		cout << "Common slice is mirror of face #" << k << endl;
	      }
	      else {
		cerr << "Unidentified common slice, this is odd.." << endl;
		exit(1);
	      }
	    }
	  }
	  else {
	    if (sx==0) {
	      cerr << "Strange, real spines intersect at infinity..." << endl;
	      exit(1);
	    }
	    else {
	      cerr << "Real spines intersect inside, linear intersection (not implemented yet)" << endl;
	      exit(1);
	    }
	  }
	}
      }	
      else {
	cerr << "Real spines don't intersect" << endl;
	genparam = true;
      }
    }
    else {
      cerr << "Real spines don't intersect" << endl;
      genparam = true;
    }
  }

  if (genparam) {
    cout << "This is a generic ridge, picture not implemented yet" << endl;
  }
  else {
    ofstream file;
    string jc = static_cast<ostringstream*>( &(ostringstream() << j) )->str();
    string kc = static_cast<ostringstream*>( &(ostringstream() << k) )->str();
    string filename = "pics/pic-ridge"+tag+"_"+jc+"-"+kc+".asy";
    file.open(filename.c_str());
    file << "import graph;" << endl;
    file << "size(7.5cm,0);" << endl;
    file << "draw(circle((0,0),1),red);" << endl;

    vector<int> inds = findFaces(V0);
    gen V0n = evali(V0);
    gen W0n = evali(W0);
    gen Hn = evali(H);

    gen X1 = evali(vertices[ridges[i].verts[0]].coords);
    gen X2 = evali(vertices[ridges[i].verts[3]].coords);
    X1 = X1/sqrt(-re(conj(X1,&ct)*Hn*X1,&ct),&ct);
    X2 = X2/sqrt(-re(conj(X2,&ct)*Hn*X2,&ct),&ct);
    gen yo = conj(X2,&ct)*Hn*X1;
    X2 = (-yo/abs(yo,&ct))*X2;
    gen MP = X1+X2;
    gen mpz = (-conj(MP,&ct)*Hn*W0n)/(conj(MP,&ct)*Hn*V0n);

    W0n = (conj(mpz,&ct)/abs(mpz,&ct))*W0n;
    V0n = V0n/sqrt(-re(conj(V0n,&ct)*Hn*V0n,&ct),&ct);
    W0n = W0n/sqrt(re(conj(W0n,&ct)*Hn*W0n,&ct),&ct);

    gen mpz2 = (-conj(MP,&ct)*Hn*W0n)/(conj(MP,&ct)*Hn*V0n);
    //    cout << "This should be real: " << mpz2 << endl;
    //    file << "label(\"mp\",(" << re(mpz2,&ct) << ",0));" << endl;

    //    cout << "V0n = " << V0n << endl;
    //    cout << "W0n = " << W0n << endl;
    //   cout << "Hn = " << Hn << endl;

    for (int u=0; u<faces.size(); u++) {
      if (u!=j && u!=k) {
	cout << "Working on face #" << u << endl;
	gen X0i = evali(faces[u].X0);
	gen X1i = evali(faces[u].X1);
	
	//	cout << "X0i = " << X0i << endl;
	//	cout << "X1i = " << X1i << endl;

	gen a = conj(X0i,&ct)*Hn*W0n;
	gen b = conj(X0i,&ct)*Hn*V0n;
	gen c = conj(X1i,&ct)*Hn*W0n;
	gen d = conj(X1i,&ct)*Hn*V0n;
	gen den = a*conj(a,&ct)-c*conj(c,&ct);
	
	/*	cout << "a=" << a << endl;
	cout << "b=" << b << endl;
	cout << "c=" << c << endl;
	cout << "d=" << d << endl;*/
	den = re(den,&ct);

	if (is_greater(zeronumber,_left(den,&ct),&ct) && is_greater(_right(den,&ct),zeronumber,&ct)) {
	  cerr << "Interval for denominator contains 0, maybe a line?" << endl;	  
	}
	else {
	  gen ctr = (conj(c,&ct)*d-conj(a,&ct)*b)/den;
	  cout << "ctr=" << ctr << endl;
	  gen radsq = re( ctr*conj(ctr,&ct) , &ct ) - re( b*conj(b,&ct)-d*conj(d,&ct) , &ct )/den;
	  if (is_greater(_left(radsq,&ct),zeronumber,&ct)) {
	    gen rad = sqrt(radsq,&ct);
	    cout << "rad = " << rad << endl;
	    if (is_greater(1000,rad,&ct)) {
	      file << "draw(circle(("<<_left(re(ctr,&ct),&ct)<<","<<_left(im(ctr,&ct),&ct)<<"),"<<_left(rad,&ct)<<"),black);" << endl;
	      gen tee = ctr*(1-rad/abs(ctr,&ct));
	      /*	      if (u==1) {
		file << "label(\"$1,14$\",(" << re(tee,&ct) << "," << im(tee,&ct) << "),right);" << endl;
	      }
	      if (u==2) {
		file << "label(\"$2,7$\",(" << re(tee,&ct) << "," << im(tee,&ct)+0.1 << "));" << endl;
		}*/
	      file << "label(\"$" << u << "$\",(" << re(tee,&ct) << "," << im(tee,&ct) << "),black+fontsize(8));" << endl;	      
	    }
	  }
	  else {
	    if (is_greater(zeronumber,_right(radsq,&ct),&ct)) {
	      cout << "No intersection, radius of circle is negative" << endl;
	    }
	    else {
	      cout << "Doubtful intersection, the interval for radius contains 0" << endl;
	      cout << "Will simply draw nothing..." << endl;
	    }
	  }
	}
      } 
    }
    // then draw the geodesic polygon through vertices...
    cout << "Drawing polygon" << endl;
    vector<gen> verts_num;
    gen ii = gen("exp(pi*i/2)",&ct);
    gen fa = gen("180/pi",&ct);
    for (int u=0; u<ridges[i].verts.size(); u++) {
      gen X = evali(vertices[ridges[i].verts[u]].coords);
      gen rat = -(conj(X,&ct)*Hn*W0n)/(conj(X,&ct)*Hn*V0n);
      verts_num.push_back(_left(re(rat,&ct),&ct)+ii*_left(im(rat,&ct),&ct));
    }
    int si = verts_num.size();
    /*    for (int u = 0; u<si; u++) {
      file << "label(\"$" << u << "$\",(" << re(verts_num[u],&ct) << "," << im(verts_num[u],&ct) << "));" << endl;
      }*/

    file << "path pol = ";
    for (int u=0; u<si; u++) {
      gen alp = verts_num[u];
      gen bet = verts_num[(u+1)%si];

      gen den = conj(alp,&ct)*bet - alp*conj(bet,&ct);
      gen ctr = (alp*(alp*conj(alp,&ct)-bet*conj(bet,&ct)) + (bet-alp)*(1+alp*conj(alp,&ct)))/den;

      gen rad = sqrt(ctr*conj(ctr,&ct)-1,&ct);
      //      cout << "ctr: " << ctr << ",   ns = " << re(ctr*conj(ctr,&ct),&ct) << endl;
      //      cout << "rad: " << rad  << endl;
      gen u1 = alp-ctr;
      gen u3 = bet-ctr;
      gen th1 = arg(u1,&ct)*fa;
      gen th2 = arg(u3,&ct)*fa;
      //     cout << "th1=" << th1 << endl;
      //     cout << "th2=" << th2 << endl;
      if (is_greater(th2,th1,&ct)) {
	if (is_greater(th2-th1,180,&ct)) {
	  th2 = th2-360;
	}
      }
      else {
	if (is_greater(th1-th2,180,&ct)) {
	  th1 = th1-360;
	}
      }
      file << "arc(("<<re(ctr,&ct)<<","<<im(ctr,&ct)<<"),"<<rad<<","<<th1<<","<<th2<<")--" << endl;      
    } 
    file << "cycle;" << endl;
    file << "filldraw(pol,gray);" << endl;
    file << "limits((-1.1,-1.1),(1.1,1.1),Crop);";
    file.close();
  }
  
}

bool Polytope::checkt1val(gen rur, gen par, gen t1v, int &res) {
  gen par_refined;
  if (checkt1valAttempt(rur,par,t1v,res)) {
    return true;
  }
  else {
    if (checkEqualt1val(rur,par,t1v)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	// compute par_refined?
	//	cout << "rur[2]=" << rur[2] << endl;
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  //	  cout << "sols[k][0]=" << sols[k][0] << endl;
	  gen Jt = _intersect(makesequence(sols[k][0],par),&ct);
	  //	  cout << "J=" << J << endl;
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkt1valAttempt(rur, par_refined, t1v, res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool Polytope::checkt1valAttempt(gen rur, gen par, gen t1v, int &res) {
  gen t1vi = evali(t1v);
  gen t1_from_rur = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  if (is_greater(_left(t1vi,&ct),_right(t1_from_rur,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(t1_from_rur,&ct),_right(t1vi,&ct),&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool Polytope::checkEqualt1val(gen rur, gen par, gen t1v) {
  gen te = _subst(makesequence(t1v,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _simplify(te - rur[5]/rur[3],&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);
  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}

bool Polytope::checkt2val(gen rur, gen par, gen t2v, int &res) {
  gen par_refined;
  if (checkt2valAttempt(rur,par,t2v,res)) {
    return true;
  }
  else {
    if (checkEqualt2val(rur,par,t2v)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	// compute par_refined?
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  gen rt;
	  if (sols[k][0].type==8) {
	    rt = convert_interval(sols[k][0],nd,&ct);      
	  }
	  else {
	    rt = sols[k][0];
	  }
	  gen Jt = _intersect(makesequence(rt,par),&ct);
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkt2valAttempt(rur, par_refined, t2v, res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool Polytope::checkt2valAttempt(gen rur, gen par, gen t2v, int &res) {
  gen t2vi = evali(t2v);
  gen t2_from_rur = _horner(makesequence(rur[6],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  if (is_greater(_left(t2vi,&ct),_right(t2_from_rur,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(t2_from_rur,&ct),_right(t2vi,&ct),&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool Polytope::checkEqualt2val(gen rur, gen par, gen t2v) {
  gen te = _subst(makesequence(t2v,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _simplify(te - rur[6]/rur[3],&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);
  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    //    cout << "g=" << g << endl;
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}

bool Polytope::checkt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res) {
  gen par_refined;
  if (checkt3valAttempt(rur,par,t3v,vl,wl,vt,res)) {
    return true;
  }
  else {
    if (checkEqualt3val(rur,par,t3v,vl,wl,vt)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  gen rt;
	  if (sols[k][0].type==8) {
	    rt = convert_interval(sols[k][0],nd,&ct);      
	  }
	  else {
	    rt = sols[k][0];
	  }
	  gen Jt = _intersect(makesequence(rt,par),&ct);
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkt3valAttempt(rur, par_refined, t3v,vl,wl,vt,res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool Polytope::checkt3valAttempt(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt, int &res) {

  gen svvi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t1vi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t2vi = _horner(makesequence(rur[6],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);

  gen alp = -Inn(vt,vl);
  gen bet = Inn(vt,wl);
  alp = (alp*myconj(bet)+myconj(alp)*bet)/2;
  bet = bet*myconj(bet);
  red(alp);
  red(bet);
  alp = convertToReal(alp);
  bet = convertToReal(bet);

  alp = _subst(makesequence(alp,s__IDNT_e,svvi),&ct);
  alp = _subst(makesequence(alp,t1,t1vi),&ct);
  alp = _subst(makesequence(alp,t2,t2vi),&ct);
  bet = _subst(makesequence(bet,s__IDNT_e,svvi),&ct);
  bet = _subst(makesequence(bet,t1,t1vi),&ct);
  bet = _subst(makesequence(bet,t2,t2vi),&ct);
  
  gen t3_from_rur = alp/bet;

  gen t3vi = evali(t3v);

  if (is_greater(_left(t3vi,&ct),_right(t3_from_rur,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(t3_from_rur,&ct),_right(t3vi,&ct),&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool Polytope::checkEqualt3val(gen rur, gen par, gen t3v, gen vl, gen wl, gen vt) {

  gen alp = -Inn(vt,vl);
  gen bet = Inn(vt,wl);
  alp = (alp*myconj(bet)+myconj(alp)*bet)/2;
  bet = bet*myconj(bet);
  red(alp);
  red(bet);
  alp = convertToReal(alp);
  bet = convertToReal(bet);

  alp = _subst(makesequence(alp,s__IDNT_e,rur[4]/rur[3]),&ct);
  alp = _subst(makesequence(alp,t1,rur[5]/rur[3]),&ct);
  alp = _subst(makesequence(alp,t2,rur[6]/rur[3]),&ct);

  bet = _subst(makesequence(bet,s__IDNT_e,rur[4]/rur[3]),&ct);
  bet = _subst(makesequence(bet,t1,rur[5]/rur[3]),&ct);
  bet = _subst(makesequence(bet,t2,rur[6]/rur[3]),&ct);

  gen te = alp/bet - _subst(makesequence(t3v,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);

  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}


bool Polytope::checkEqSign(gen rur, gen par, gen eq, int &res) {
  gen par_refined;
  if (checkEqSignAttempt(rur,par,eq,res)) {
    return true;
  }
  else {
    if (checkEqZero(rur,par,eq)) {
      res = 0;
      return true;
    }
    else {
      unsigned savend = nd;
      bool done = false;
      while (nd<ndmax && !done) {
	gen sols =_realroot(makesequence(rur[2],prec),&ct);
	int count = 0;
	int ksave;
	for (unsigned k=0; k<(*sols._VECTptr).size(); k++) {
	  gen rt;
	  if (sols[k][0].type==8) {
	    rt = convert_interval(sols[k][0],nd,&ct);      
	  }
	  else {
	    rt = sols[k][0];
	  }
	  gen Jt = _intersect(makesequence(rt,par),&ct);
	  if (Jt.type==3) {
	    count++;
	    ksave = k;
	  }
	}
	if (count!=1) {
	  cout << "Could not identify par from rur (count=" << count << ")" << endl;
	  return false;
	}
	else {
	  par_refined = sols[ksave][0];
	}
	done = checkEqSignAttempt(rur,par_refined,eq,res);
	if (!done) {
	  increasePrecision();
	}
      }
      nd = savend;
      lowerPrecision();
      return done;
    }
  }
}

bool Polytope::checkEqSignAttempt(gen rur, gen par, gen eq, int &res) {
  gen svvi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t1vi = _horner(makesequence(rur[5],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);
  gen t2vi = _horner(makesequence(rur[6],par,s__IDNT_e),&ct)/_horner(makesequence(rur[3],par,s__IDNT_e),&ct);

  gen eq_eval = _subst(makesequence(eq,s__IDNT_e,svvi),&ct);
  eq_eval = _subst(makesequence(eq_eval,t1,t1vi),&ct);
  eq_eval = _subst(makesequence(eq_eval,t2,t2vi),&ct);

  if (is_greater(zeronumber,_right(eq_eval,&ct),&ct)) {
    res = -1;
    return true;
  }
  else {
    if (is_greater(_left(eq_eval,&ct),zeronumber,&ct)) {
      res = 1;
      return true;
    }
    else {
      return false;
    }
  }
}

bool Polytope::checkEqZero(gen rur, gen par, gen eq) {
  gen te = _subst(makesequence(eq,s__IDNT_e,rur[4]/rur[3]),&ct);
  te = _subst(makesequence(te,t1,rur[5]/rur[3]),&ct);
  te = _subst(makesequence(te,t2,rur[6]/rur[3]),&ct);
  te = _numer(te,&ct);
  te = _rem(makesequence(te,rur[2],s__IDNT_e),&ct);

  if (operator_equal(te,zeronumber,&ct)) {
    return true;
  }
  else {
    // need to check if par is a root of te!
    gen g = _gcd(makesequence(te,rur[2]),&ct);
    gen sols = _realroot(makesequence(g,prec),&ct);
    int count = 0;
    bool fd = false;
    int k = 0;
    while (!fd && k<(*sols._VECTptr).size()) {
      gen rt;
      if (sols[k][0].type==8) {
	rt = convert_interval(sols[k][0],nd,&ct);      
      }
      else {
	rt = sols[k][0];
      }
      gen Jt = _intersect(makesequence(rt,par),&ct);
      fd = Jt.type==3;
      k++;
    }
    return fd;
  }
}

void Polytope::exportFaceData() {

  ofstream file;
  string filename = "faceData/faceData_"+tag+".chg";
  file.open(filename.c_str());
  for (unsigned k=0; k<faces.size(); k++) {
    file << "[" << faces[k].sides.size() << "]" << endl;
    file << convertword(faces[k].mirror.word) << endl;
    for (unsigned l=0; l<faces[k].sides.size(); l++) {
      file << convertword(faces[k].sides[l].word) << endl;
    }
  }
  /*  for (unsigned k=0; k<facePOrbits.size(); k++) {
    file << "[" << faces[facePOrbits[k][0]].sides.size() << "]" << endl;
    file << convertword(faces[facePOrbits[k][0]].mirror.word) << endl;
    for (unsigned l=0; l<faces[facePOrbits[k][0]].sides.size(); l++) {
      file << convertword(faces[facePOrbits[k][0]].sides[l].word) << endl;
    }
    }*/
  file.close();

}

void Polytope::checkLinks() {

  for (unsigned k=0; k<vertexPOrbits.size(); k++) {
    int kk = vertexPOrbits[k][0];
    WVector V = vertices[kk];
    int e0 = 0;
    int e1 = 0;
    int e2 = 0;
    vector<int> incident_edges;
    for (unsigned l=0; l<edges.size(); l++) {
      if (edges[l].origin==kk || edges[l].end==kk) {
	incident_edges.push_back(l);
      }
    }
    e0 = incident_edges.size();
    for (unsigned l=0; l<ridges.size(); l++) {
      int m = -1;
      bool isvert = false;
      while (!isvert && m!=ridges[l].verts.size()-1) {
	m++;
	isvert = ridges[l].verts[m]==kk;
      }
      if (isvert) {
	e1++;
      }
    }
    vector<vector<int> > twofaces_in_link; // Note these could be DIGONS!
                                           // I think this is not a problem, digons do not affect orientation?
    for (unsigned u=0; u<vertices[kk].faceindices.size(); u++) {
      if (vertices[kk].faceindices[u]!=-1) {
	vector<int> twoface, ordered_twoface;
	for (unsigned v=0; v<incident_edges.size(); v++) {
	  vector<int> fi = edges[incident_edges[v]].faceindices;
	  if (find(fi.begin(),fi.end(),vertices[kk].faceindices[u])!=fi.end()) {
	    twoface.push_back(incident_edges[v]);
	  }
	}
	//	cout << "Unordered 2-face: " << convertword(twoface) << endl;
	ordered_twoface.push_back(twoface[0]);
	twoface.erase(twoface.begin());
	bool foundneighb = false;
	int st1 = -1;
	while (!foundneighb && st1!=twoface.size()-1) {
	  st1++;
	  foundneighb = intersect(edges[twoface[0]].faceindices,edges[twoface[st1]].faceindices).size()!=0;
	}
	if (!foundneighb) {
	  cerr << "Trouble finding neighbor in 2-face of link" << endl;
	  exit(1);
	}
	else {
	  ordered_twoface.push_back(twoface[st1]);
	  twoface.erase(twoface.begin()+st1);
	  while (twoface.size()>0) {
	    foundneighb = false;
	    int t = -1;
	    while (!foundneighb && t!=twoface.size()-1) {
	      t++;
	      foundneighb = ( intersect(edges[ordered_twoface[ordered_twoface.size()-1]].faceindices,edges[twoface[t]].faceindices).size()!=0 );
	    }
	    if (!foundneighb) {
	      cerr << "Trouble finding neighbor in 2-face of link" << endl;
	      exit(1);
	    }
	    else {
	      ordered_twoface.push_back(twoface[t]);
	      twoface.erase(twoface.begin()+t);
	      //   cout << "twoface: " << convertword(twoface) << endl;	  
	    }
	  }
	}
	if (ordered_twoface.size()>2) {
	  twofaces_in_link.push_back(ordered_twoface);
	}
	/*	else {
	  cout << "Ignoring digon, irrelevant for orientation check" << endl;
	  }*/
	//	cout << "Ordered 2-face: " << convertword(ordered_twoface) << endl;
      }
    }
    /*    cout << "Ordered 2-faces:"  << endl;
    for (unsigned uu=0; uu<twofaces_in_link.size(); uu++) {
      cout << convertword(twofaces_in_link[uu]) << endl;
      }*/
    // then need to check if we can orient them coherently!
    
    vector<int> list;
    for (unsigned uu=0; uu<twofaces_in_link.size(); uu++) {
      list.push_back(uu);
    }
    vector<vector<int> > well_oriented_twofaces;
    
    bool is_orientable = true;
    while (is_orientable && list.size()>0) {
      
      // find a 2-face that neighbors a well oriented one
      bool isneighbor = false;
      int zz=-1;
      vector<int> common;
      while (!isneighbor && zz!=well_oriented_twofaces.size()-1) {
	zz++;
	vector<int> common = intersect(twofaces_in_link[0],well_oriented_twofaces[zz]); 
	isneighbor = common.size()==2;
      }
      if (!isneighbor) {
	list.erase(list.begin());
      }
      else {
	// found a neighbor, will orient it well with that neighbor, then check compatibility

	bool or1, or2;
	int uuu=-1;
	bool fdcommon = false;
	while (!fdcommon && uuu!=twofaces_in_link[0].size()-1) {
	  uuu++;
	  fdcommon = common[0]==twofaces_in_link[0][uuu];
	}
	if (!fdcommon) {
	  cerr << "Trouble finding common" << endl;
	  exit(1);
	}
	else {
	  or1 = (common[1]==twofaces_in_link[0][(uuu+1)%twofaces_in_link[0].size()]);
	}
	int uuuu=-1;
	fdcommon = false;
	while (!fdcommon && uuuu!=well_oriented_twofaces[zz].size()-1) {
	    uuuu++;
	    fdcommon = common[0]==well_oriented_twofaces[zz][uuuu];
	}
	if (!fdcommon) {
	  cerr << "Trouble finding common" << endl;
	  exit(1);
	}
	else {
	  or2 = (common[1]==well_oriented_twofaces[zz][(uuuu+1)%well_oriented_twofaces[zz].size()]);
	}
	if (or1==or2) {
	  // switch orientation!
	  twofaces_in_link[0] = flipword(twofaces_in_link[0]);
	}
       
	bool is_coherent;
	for (unsigned y=0; y<well_oriented_twofaces.size(); y++) {	  
	  vector<int> common = intersect(twofaces_in_link[0],well_oriented_twofaces[y]); 
	  if (common.size()==2) {
	    // locate common in both 2faces?
	    bool or1, or2;
	    int uuu=-1;
	    bool fdcommon = false;
	    while (!fdcommon && uuu!=twofaces_in_link[0].size()-1) {
	      uuu++;
	      fdcommon = common[0]==twofaces_in_link[0][uuu];
	    }
	    if (!fdcommon) {
	      cerr << "Trouble finding common" << endl;
	      exit(1);
	    }
	    else {
	      or1 = (common[1]==twofaces_in_link[0][(uuu+1)%twofaces_in_link[0].size()]);
	    }
	    int uuuu=-1;
	    fdcommon = false;
	    while (!fdcommon && uuuu!=well_oriented_twofaces[y].size()-1) {
	      uuuu++;
	      fdcommon = common[0]==well_oriented_twofaces[y][uuuu];
	    }
	    if (!fdcommon) {
	      cerr << "Trouble finding common" << endl;
	      exit(1);
	    }
	    else {
	      or2 = (common[1]==well_oriented_twofaces[y][(uuuu+1)%well_oriented_twofaces[y].size()]);
	    }
	    if (or1==or2) {
	      cerr << "This vertex link is not orientable!!" << endl;
	      exit(1);
	    }
	  }
	}
	twofaces_in_link.erase(twofaces_in_link.begin());
      }
    }
    //    cout << "If you read this, this vertex links is an orientable surface!" << endl;

    if (vertices[kk].faceindices[0]==-1) {
      e2 = vertices[kk].faceindices.size()-1;
    }
    else {
      e2 = vertices[kk].faceindices.size();
    }
    gen ee = (e0-e1+e2);
    if (!operator_equal(ee,twonumber,&ct)) {
      cerr << "Euler characteristic of vertex: " << ee << ", this is not a sphere!" << endl;
      exit(1);
    }
  }
  
  cout << endl;
  cout << "Great! Vertex links are spheres!" << endl;
  cout << "The boundary of the polytope is a manifold." << endl;
  cout << endl;

}

void Polytope::checkSphere() {
  
  vector<vector<int> > graph_edges;
  for (unsigned uu=0; uu<edges.size(); uu++) {
    vector<int> e;
    e.push_back(edges[uu].origin);
    e.push_back(edges[uu].end);
    graph_edges.push_back(e);
  }
  sgraph G(graph_edges);
      
  vector<vector<int> > stabgens;
  vector<vector<bool> > orientations;
  G.fundamental_group(0,stabgens,orientations);

  cout << "1-skeleton has " << vertices.size() << " vertices, " << edges.size() << " edges" << endl;
  cout << "Found " << G.generators.size() << " gens for pi_1 of 1-skeleton" << endl;

  vector<vector<int> > ridgerels;
  vector<vector<bool> > ridgeorients;
  for (unsigned u=0; u<ridges.size(); u++) {
    vector<int> rel;
    vector<bool> orient;
    int nv = ridges[u].verts.size();
    for (unsigned v=0; v<nv; v++) {
      int jo = ridges[u].verts[v%nv];
      int ko = ridges[u].verts[(v+1)%nv];
      bool fd = false;
      bool flipped;
      int w = -1;
      while (!fd && w!=edges.size()-1) {
	w++;
	fd = (jo==edges[w].origin && ko==edges[w].end);
	if (fd) {
	  flipped = false;
	}
	else {
	  fd = (jo==edges[w].end && ko == edges[w].origin);
	  if (fd) {
	    flipped = true;
	  }
	}
      }
      if (fd) {
	rel.push_back(w);
	orient.push_back(flipped);
      }
      else {
	cerr << "Problem identifying edge in ridge" << endl;
	exit(1);
      }
    }
    ridgerels.push_back(rel);
    ridgeorients.push_back(orient);
  }
  /*  cout << "Ridge relations: " << endl;
  for (unsigned u=0; u<ridgerels.size(); u++) {
    for (unsigned v=0; v<ridgerels[u].size(); v++) {
      if (ridgeorients[u][v]) {
	cout << "-";
      }
      cout << ridgerels[u][v] << ",";
    }
    cout << endl;
    }*/

  G.export_fg_with_relations(0,tag,ridgerels,ridgeorients);
  cout << "  (you should check that that group is trivial by simplifying it with GAP)" << endl;
  cout << endl;

}

string Polytope::checkSignature(gen A) {
  // this only makes sense provided A is a Hermitian matrix!
 
  gen de = det(A);
  red(de);
  de = convertToReal(de);
  int sgn_de;
  if (!checkSign(de,sgn_de)) {
    cerr << "Trouble checking signature of the Hermitian form" << endl;
    exit(1);
  }
  else {
    if (sgn_de<0) {
      return "-++";
      // this only works for sporadic/triangle group Hermitian forms!
      //   (because by construction there are some positive vectors, so signature cannot be ---)
    }
    else {
      gen X = Inn(e1,e1)*e2 - Inn(e2,e1)*e1;
      red(X);
      gen sn = SqNorm(X);
      int sgn_sn;
      if (!checkSign(sn,sgn_sn)) {
	cerr << "Trouble checking signature of the Hermitian form (doing Gramm-Schmidt)" << endl;
	exit(1);
      }
      else {
	if (sgn_sn>0) {
	  return "+++";
	}
	else {
	  return "--+";
	}
      }
    }
  }
}

bool Polytope::isArithmetic() {

  vecteur other_roots;

  gen fa = _factors(makesequence(_subst(makesequence(minpol,x__IDNT_e,a__IDNT_e),&ct),genasrootof),&ct);
  
  //  cout << "fa=" << fa << endl;

  int nf = (*fa._VECTptr).size()/2;
    //    cout << "Nb factors: " << nf << endl;  

  int k=0;
  bool found = false;
  while (k<nf && !found) {
    gen df = _degree(makesequence(fa[2*k],a__IDNT_e),&ct);
    if (df==1) {
      //      cout << "Factor: " << fa[2*k] << endl;
      gen te = _simplify(-_coeff(makesequence(fa[2*k],a__IDNT_e,0),&ct)/_coeff(makesequence(fa[2*k],a__IDNT_e,1),&ct),&ct);
      //      cout << _evalf(te,&ct) << endl;
      if (te.is_symb_of_sommet(at_prod)) {
	gen subfa = te._SYMBptr->feuille;
	gen yo = _simplify(te[1][1]*te[2],&ct);
	int umax = dc.val-(*yo._VECTptr).size();
	for (int u=0; u<umax; u++) {
	  (*yo._VECTptr).insert((*yo._VECTptr).begin(),zeronumber);
	}
	gen ls = _inverse(genasrootofpowers,&ct)*yo;
	gen yoo = _poly2symb(makesequence(ls,x__IDNT_e),&ct);
	red(yoo);
	other_roots.push_back(yoo);
      }
      else {
	if (te.is_symb_of_sommet(at_rootof)) {
	  gen yo = _simplify(te[1],&ct);
	  int umax = dc.val-(*yo._VECTptr).size();
	  for (int u=0; u<umax; u++) {
	    (*yo._VECTptr).insert((*yo._VECTptr).begin(),zeronumber);
	  }
	  //  gen ls = _linsolve(makesequence(genasrootofpowers,yo),&ct);
	  gen ls = _inverse(genasrootofpowers,&ct)*yo;
	  gen yoo = _poly2symb(makesequence(ls,x__IDNT_e),&ct);
	  red(yoo);
	  other_roots.push_back(yoo);
	}
	else {
	  cout << "Problem reading factors!" << endl;
	}
      }
    }
    k++;
  }

  vecteur gensSmallField;
  if (tag=="3s4c") {
    gensSmallField.push_back(findInField(gen("u^2-21",&ct),_evalf(gen("sqrt(21)",&ct),&ct)));
  }
  if (tag=="4s4c") {
    gensSmallField.push_back(findInField(gen("u^2-7",&ct),_evalf(gen("sqrt(7)",&ct),&ct)));
  }
  if (tag=="5s4c") {
    gensSmallField.push_back(findInField(gen("u^4-140*u^2+3920",&ct),_evalf(gen("sqrt(14)*sqrt(5+sqrt(5))",&ct),&ct)));
  }
  if (tag=="6s4c") {
    gensSmallField.push_back(findInField(gen("u^2-21",&ct),_evalf(gen("sqrt(21)",&ct),&ct)));
  }
  if (tag=="8s4c") {
    gensSmallField.push_back(findInField(gen("u^2-2",&ct),_evalf(gen("sqrt(2)",&ct),&ct)));
    gensSmallField.push_back(findInField(gen("u^2-7",&ct),_evalf(gen("sqrt(7)",&ct),&ct)));
  }
  if (tag=="12s4c") {
    gensSmallField.push_back(findInField(gen("u^2-3",&ct),_evalf(gen("sqrt(3)",&ct),&ct)));
    gensSmallField.push_back(findInField(gen("u^2-7",&ct),_evalf(gen("sqrt(7)",&ct),&ct)));
  }



  if (tag=="3s1") {
    gensSmallField.push_back(findInField(gen("u^2-6",&ct),_evalf(gen("sqrt(6)",&ct),&ct)));
  }
  if (tag=="4s1") {
    gensSmallField.push_back(findInField(gen("u^2-2",&ct),_evalf(gen("sqrt(2)",&ct),&ct)));
  }
  if (tag=="6s1") {
    gensSmallField.push_back(findInField(gen("u^2-6",&ct),_evalf(gen("sqrt(6)",&ct),&ct)));
  }



  if (tag=="2s5") {
    gensSmallField.push_back(findInField(gen("u^2-5",&ct),_evalf(gen("sqrt(5)",&ct),&ct)));
  }
  if (tag=="3s5") {
    gensSmallField.push_back(findInField(gen("u^2-5",&ct),_evalf(gen("sqrt(5)",&ct),&ct)));
  }
  if (tag=="4s5") {
    gensSmallField.push_back(findInField(gen("u^2-3",&ct),_evalf(gen("sqrt(3)",&ct),&ct)));
    gensSmallField.push_back(findInField(gen("u^2-5",&ct),_evalf(gen("sqrt(5)",&ct),&ct)));
  }

  for (unsigned u=0; u<gensSmallField.size(); u++){
    red(gensSmallField[u]);
  }

  vecteur or_short;
  for (unsigned u=0; u<other_roots.size(); u++) {
    gen teu = other_roots[u];
    gen teuc = myconj(teu);
    bool fdd = false;
    int yy = 0;
    while (!fdd && yy<or_short.size()) {
      fdd = operator_equal(teu,or_short[yy],&ct) || operator_equal(teuc,or_short[yy],&ct); 
      yy++;
    }
    if (!fdd) {
      or_short.push_back(teu);
    }    
  }
  
  cout << "Signature of (non trivial) Galois conjugates?" << endl;
  bool is_arithmetic = true;
  vector<vecteur> vals;
  vecteur ty;
  vals.push_back(gensSmallField);
  for (unsigned u=0; u<or_short.size(); u++) {
    ty.clear();
    for (int v=0; v<gensSmallField.size(); v++) {
      gen te = _subst(makesequence(gensSmallField[v],x__IDNT_e,or_short[u]),&ct);
      red(te);
      ty.push_back(te);
    }
    bool isnew = true;
    int w=0;
    while (isnew && w<vals.size()) {
      bool issame = true;
      for (unsigned z=0; z<gensSmallField.size(); z++) {
	issame = issame && operator_equal(vals[w][z],ty[z],&ct);
      }
      if (issame) {
	isnew = false;
      }
      w++;
    }
    if (isnew) {
      gen Hu = _subst(makesequence(H,x__IDNT_e,or_short[u]),&ct);
      matred(Hu);
      string si = checkSignature(Hu);
      cout << si << endl;
      if (si=="-++" || si=="--+") {
	is_arithmetic=false;
      }
      vals.push_back(ty);
    }
  }
  
  return is_arithmetic;

}
